#$Id$
# VT50-2000-2016-84-28-4
# $Copyright (c) 2017 Veritas Technologies LLC.  All rights reserved. Veritas
# and the Veritas Logo are trademarks or registered trademarks of Veritas
# Technologies LLC or its affiliates in the U.S. and other countries. Other 
# names may be trademarks of their respective owners.
#
# The Licensed Software and Documentation are deemed to be commercial computer
# software as defined in FAR 12.212 and subject to restricted rights as defined
# in FAR Section 52.227-19 "Commercial Computer Software - Restricted Rights" and
# DFARS 227.7202, et seq. "Commercial Computer Software and Commercial Computer
# Software Documentation," as applicable, and any successor regulations, whether
# delivered by Veritas as on premises or hosted services. Any use, modification,
# reproduction release, performance, display or disclosure of the Licensed Software
# and Documentation by the U.S. Government shall be solely in accordance with the
# terms of this Agreement. $

package VRTS::Enclosure::Correlation;
use strict;
use Carp;
require Exporter;
use File::Spec;
use Data::Dumper;
use File::Basename;
use Fcntl ':flock';
use FileHandle;

require VRTS::Paths;
require JSON;
require VRTS::Util;
require VRTS::Summary;
require VRTS::Enclosure::Utils;

require VRTS::Constants::P_ARRAY_NAS_SHARE;
require VRTS::Constants::P_ARRAY_DEVICE_ASSIGNMENT;
require VRTS::Constants::P_ARRAY_SUBSYSTEM;

use vars qw($VERSION @ISA @EXPORT);
@ISA = qw(Exporter);
@EXPORT = qw(correlate_nas_share_esx_sp pull_esx_storagepool_data);
$VERSION = '3.0';

use constant UPDATE_SP_CACHE => 7200; # once every 2 hrs.
my (@hid, $context);
sub correlate_nas_share_esx_sp
{
    my ($agent, $ip_list, $subsys_id, $subsys_name, $family) = @_;

    my $ret = 0;
    my $pool_hash = undef;
    my $method = $agent->{'NAME'};

    ($ret, $pool_hash) = pull_esx_storagepool_data($agent);

    return 1 if ($ret != 0 or not defined $pool_hash);

    # Seems like we have all the data is required to perform corelation.

    my $nfs_export_tb = $agent->gettable("P_ARRAY_NAS_SHARE");
    my $nfs_rows; 
    #Netapp CDOT: Get the rows only for the specified vserver
    if ((defined($family)) && ($family eq 'VRP_NETAPP_CDOT')) {
        my @map_cols =
            (&VRTS::Constants::P_ARRAY_NAS_SHARE::SUBSYS_ID,
            &VRTS::Constants::P_ARRAY_NAS_SHARE::MOUNT_POINT_TYPE);
        my @map_vals = ($subsys_id, 'NFS');
        $nfs_rows = $nfs_export_tb->getallrowswithvaluez(\@map_cols, \@map_vals);
    } else {
        $nfs_rows = $nfs_export_tb->getallrowswithvalue(&VRTS::Constants::P_ARRAY_NAS_SHARE::MOUNT_POINT_TYPE, 'NFS');
    }
    return if (not defined $nfs_rows);

    my $dev_assg_tb = $agent->gettable("P_ARRAY_DEVICE_ASSIGNMENT");
    my $enc_tb = $agent->gettable("P_ARRAY_SUBSYSTEM");

    foreach my $nfs_r (@{$nfs_rows})
    {
        my $pool_id = undef;

        #Look for Correlatable SP mounted with IP of filer for the given share.
        foreach my $tmp_ip (@{$ip_list})
        {
            if ((defined($family)) && ($family eq 'VRP_NETAPP_CDOT')) {
                $pool_id = correlated_share($pool_hash, $nfs_r->[&VRTS::Constants::P_ARRAY_NAS_SHARE::PATH], 
                                            $tmp_ip, $context->{$subsys_name}, 'VRP_NETAPP_CDOT');
            } else {
                $pool_id = correlated_share($pool_hash, $nfs_r->[&VRTS::Constants::P_ARRAY_NAS_SHARE::PATH], $tmp_ip);
            }
            last if (defined($pool_id));
        }

        #Look for Correlatable SP mounted with FQDN of filer for the given share.
        if ((not defined($pool_id)) or ($pool_id eq '')) {
            if ((defined($family)) && ($family eq 'VRP_NETAPP_CDOT')) {
                $pool_id = correlated_share($pool_hash, $nfs_r->[&VRTS::Constants::P_ARRAY_NAS_SHARE::PATH], 
                                            undef, $context->{$subsys_name}, 'VRP_NETAPP_CDOT');
            } else {
                $pool_id = correlated_share($pool_hash, $nfs_r->[&VRTS::Constants::P_ARRAY_NAS_SHARE::PATH]);
            }
        }

        if (defined($pool_id)) {
            VRTS::Logger::log('debug', "SP ID = [$pool_hash->{$pool_id}->{id}] SP Name = [$pool_hash->{$pool_id}->{name}]");

            #Skip the correlation if valid correlation already exists
            my $exist = check_existing_correlation('NETAPP', $dev_assg_tb,
                            $nfs_r->[&VRTS::Constants::P_ARRAY_NAS_SHARE::PATH], $pool_hash->{$pool_id}->{'id'}); 
            next if($exist);

            # For the specific nas share, get parent/container information as well.
            my $enc_row = $enc_tb->getrowwithvalue(&VRTS::Constants::P_ARRAY_SUBSYSTEM::ID,
                $nfs_r->[&VRTS::Constants::P_ARRAY_NAS_SHARE::SUBSYS_ID]);
            _report_device_assignment($agent, $dev_assg_tb,
                $nfs_r->[&VRTS::Constants::P_ARRAY_NAS_SHARE::ID],
                'NFS_SHARE',
                $nfs_r->[&VRTS::Constants::P_ARRAY_NAS_SHARE::NAME],
                $pool_hash->{$pool_id}->{id},
                'STORAGE_POOL',
                $pool_hash->{$pool_id}->{name},
                $enc_row->[&VRTS::Constants::P_ARRAY_SUBSYSTEM::ID],
                'ENCLOSURE',
                $enc_row->[&VRTS::Constants::P_ARRAY_SUBSYSTEM::NAME],
                undef, undef, undef,
                $nfs_r->[&VRTS::Constants::P_ARRAY_NAS_SHARE::SOURCE_CONTAINER_ID]);
        }
    }
}

sub pull_esx_storagepool_data
{
    my $agent = shift;

    $context = $agent->get_context();

    my $app_dir = VRTS::Paths::get_path("AppDir");
    $app_dir = File::Spec->catdir($app_dir, "SIData");

    eval { File::Path::mkpath($app_dir, 0, 755); };

    return (1, undef) if (! -d $app_dir);

    my $last_time = undef;
    my $time_f = File::Spec->catfile($app_dir, ".cache_status");
    my $cache_file = File::Spec->catfile($app_dir, "storage_pool.cache");
    my $cache_lock = File::Spec->catfile($app_dir, "storage_pool.lock");

    if (-f $time_f)
    {
        open (HAN, $time_f);
        $last_time = <HAN>;
        close(HAN);

        chomp $last_time;
    }
    else
    {
        $last_time = 0; # this will force the cache to be built up for the very first time.
    }

    my $lc_time = time;
    # Compare this with what is stampped in the file.

    # Acquire a lock and a blocking lock in this case.
    open(CACHE_LOCK, "> $cache_lock");
    flock(CACHE_LOCK, LOCK_EX);

    #Set the update period to value of context varaible 'cache_update_period'
    #else keep to default value of UPDATE_SP_CACHE
    my $update_period;
    my $cache_update_period = $context->{'cache_update_period'};
    if ((defined($cache_update_period)) && ($cache_update_period =~ /^\d+$/)) {
        $update_period = $cache_update_period;
    } else {
        $update_period = &UPDATE_SP_CACHE;
    }

    my ($ret, $op);
    if ($lc_time > ($last_time + $update_period) or $context->{update_cache} == 1)
    {
        VRTS::Logger::log('debug', "Need to refresh the cache under [$cache_file]");

        # Lets get the NFS type storage pools from the database.
        my $recs = VRTS::Enclosure::Utils::pull_nfs_storage_pools();
        return (1, undef) if (not defined $recs);

        my $pool_hash = _massage_pool_data($recs);
        my $ret = _sync_cache($pool_hash, $cache_file);

        if ($ret != 0)
        {
            $ret = 1;
            $op = undef;
            VRTS::Logger::log('debug', "Nothing obtained to generate cache, skip correlation for now");
            goto SKIP_MARKER_CREATION;
        }
        else
        {
            $ret = 0;
            $op = $pool_hash;
        }

        open(HAN, "> $time_f");
        print HAN $lc_time;
        close(HAN);
    }
    else
    {
        VRTS::Logger::log('debug', "No need to refresh the cache, reading cache from [$cache_file]");
        ($ret, $op) = _read_cache($cache_file);

        if ($ret == 2)
        {
            VRTS::Logger::log('debug', "Invalidating the cache and its contents, along with timestamp file");
            unlink ("$time_f");
            unlink ("$cache_file");

            my @ignore_tbs = ("P_ARRAY_DEVICE_ASSIGNMENT");
            $agent->ignoretables(\@ignore_tbs);
        }
    }

SKIP_MARKER_CREATION:

    # Release the lock in this case.
    close(CACHE_LOCK);

    return $ret, $op;
}

# Some helper routines to manage a cache.

sub _report_device_assignment
{
    my $agent = shift;
    my $map_tb = shift;
    my $source_id = shift;
    my $source_type = shift;
    my $source_name = shift;
    my $dest_id = shift;
    my $dest_type = shift;
    my $dest_name = shift;
    my $source_con_id = shift;
    my $source_con_type = shift;
    my $source_con_name = shift;
    my $dest_con_id = shift;
    my $dest_con_type = shift;
    my $dest_con_name = shift;
    my $src_parent_id = shift;


    my @map_row;

    $map_row[&VRTS::Constants::P_ARRAY_DEVICE_ASSIGNMENT::SOURCE_ID] = $source_id;
    $map_row[&VRTS::Constants::P_ARRAY_DEVICE_ASSIGNMENT::SOURCE_TYPE] = $source_type;
    $map_row[&VRTS::Constants::P_ARRAY_DEVICE_ASSIGNMENT::SOURCE_NAME] = $source_name;
    $map_row[&VRTS::Constants::P_ARRAY_DEVICE_ASSIGNMENT::DEST_ID] = $dest_id;
    $map_row[&VRTS::Constants::P_ARRAY_DEVICE_ASSIGNMENT::DEST_TYPE] = $dest_type;
    $map_row[&VRTS::Constants::P_ARRAY_DEVICE_ASSIGNMENT::DEST_NAME] = $dest_name;

    $map_row[&VRTS::Constants::P_ARRAY_DEVICE_ASSIGNMENT::SOURCE_CONTAINER_ID] = $source_con_id;
    $map_row[&VRTS::Constants::P_ARRAY_DEVICE_ASSIGNMENT::SOURCE_CONTAINER_TYPE] = $source_con_type;
    $map_row[&VRTS::Constants::P_ARRAY_DEVICE_ASSIGNMENT::SOURCE_CONTAINER_NAME] = $source_con_name;

    $map_row[&VRTS::Constants::P_ARRAY_DEVICE_ASSIGNMENT::DEST_CONTAINER_ID] = $dest_con_id;
    $map_row[&VRTS::Constants::P_ARRAY_DEVICE_ASSIGNMENT::DEST_CONTAINER_TYPE] = $dest_con_type;
    $map_row[&VRTS::Constants::P_ARRAY_DEVICE_ASSIGNMENT::DEST_CONTAINER_NAME] = $dest_con_name;
    $map_row[&VRTS::Constants::P_ARRAY_DEVICE_ASSIGNMENT::SOURCE_PARENT_ID] = $src_parent_id;

    $map_tb->addrow(\@map_row);
    #$map_tb->print_table();
}

sub _report_host_device_assignment
{
    my $agent = shift;
    my $map_tb = shift;
    my $source_id = shift;
    my $source_type = shift;
    my $source_name = shift;
	my $pool_hash = shift;
	my $share_id = shift;
	my $e_vdid = shift;
	my $enc_name = shift;
	my $enc_type = shift;
    my $source_parent_id = shift;

    my @map_row;

    $map_row[&VRTS::Constants::P_ARRAY_DEVICE_ASSIGNMENT::SOURCE_ID] = $source_id;
    $map_row[&VRTS::Constants::P_ARRAY_DEVICE_ASSIGNMENT::SOURCE_TYPE] = $source_type;
    $map_row[&VRTS::Constants::P_ARRAY_DEVICE_ASSIGNMENT::SOURCE_NAME] = $source_name;
    $map_row[&VRTS::Constants::P_ARRAY_DEVICE_ASSIGNMENT::DEST_ID] =
	    $pool_hash->{$share_id}->{dest_id};
    $map_row[&VRTS::Constants::P_ARRAY_DEVICE_ASSIGNMENT::DEST_TYPE] = 
	    $pool_hash->{$share_id}->{dest_type};
    $map_row[&VRTS::Constants::P_ARRAY_DEVICE_ASSIGNMENT::DEST_NAME] = 
	    $pool_hash->{$share_id}->{dest_name};
	$map_row[&VRTS::Constants::P_ARRAY_DEVICE_ASSIGNMENT::SOURCE_CONTAINER_ID] =
	    $e_vdid;
    $map_row[&VRTS::Constants::P_ARRAY_DEVICE_ASSIGNMENT::SOURCE_CONTAINER_TYPE] =
	    $enc_type;
	$map_row[&VRTS::Constants::P_ARRAY_DEVICE_ASSIGNMENT::SOURCE_CONTAINER_NAME]
	    = $enc_name;
	$map_row[&VRTS::Constants::P_ARRAY_DEVICE_ASSIGNMENT::DEST_CONTAINER_ID] = 
	    $pool_hash->{$share_id}->{dest_container_id};
	$map_row[&VRTS::Constants::P_ARRAY_DEVICE_ASSIGNMENT::DEST_CONTAINER_NAME] =
	    $pool_hash->{$share_id}->{dest_container_name};
	$map_row[&VRTS::Constants::P_ARRAY_DEVICE_ASSIGNMENT::DEST_CONTAINER_TYPE] =
	    $pool_hash->{$share_id}->{dest_container_type};
	$map_row[&VRTS::Constants::P_ARRAY_DEVICE_ASSIGNMENT::SOURCE_PARENT_ID] = $source_parent_id;

    $map_tb->addrow(\@map_row);
    my $share_tb = $agent->gettable("P_ARRAY_SHARE_CONSUMER");
    my $nfs_rows =
         $share_tb->getallrowswithvalue(
                  &VRTS::Constants::P_ARRAY_SHARE_CONSUMER::HOST_NAME, $pool_hash->{$share_id}->{dest_container_name});
    foreach my $nfs_r (@{$nfs_rows})
    {
        $nfs_r->[&VRTS::Constants::P_ARRAY_SHARE_CONSUMER::HOST_ID] = $pool_hash->{$share_id}->{dest_container_id};
    }
    # Check if mapped using IP
    $nfs_rows = $share_tb->getallrowswithvalue(
                   &VRTS::Constants::P_ARRAY_SHARE_CONSUMER::HOST_NAME, $pool_hash->{$share_id}->{dest_container_ip});
    foreach my $nfs_r (@{$nfs_rows})
    {
        $nfs_r->[&VRTS::Constants::P_ARRAY_SHARE_CONSUMER::HOST_ID] = $pool_hash->{$share_id}->{dest_container_id};
    }
    # For CFS shares
    $share_id =~ /^\\(\d+\.\d+\.\d+\.\d+)\\/;
    $nfs_rows = $share_tb->getallrowswithvalue(
                &VRTS::Constants::P_ARRAY_SHARE_CONSUMER::HOST_NAME, $1);
    foreach my $nfs_r (@{$nfs_rows})
    {
        $nfs_r->[&VRTS::Constants::P_ARRAY_SHARE_CONSUMER::HOST_ID] = $pool_hash->{$share_id}->{dest_container_id};
    }
}

sub _read_cache
{
    my $cache_file = shift;

    return 1, undef if (! -f $cache_file);

    open FILE, "< $cache_file";
    my $buf = do { local $/; <FILE> };
    close(FILE);

    my $jsonobject = new JSON ();
    my $data = undef;

    eval {
        $data = $jsonobject->decode($buf);
    };

    if ($@)
    {
        VRTS::Logger::log('error', "Cache seems invalid, need to re-generate");
        return 2, undef;
    }

    return 0, $data;
}

sub _sync_cache
{
    my $hash = shift;
    my $cache_file = shift;

    my @key = (keys %{$hash});

    return 1 if (!@key or scalar(@key) == 0);

    my $final_buf = JSON::to_json($hash, {pretty => 1, indent => 1});

    my $fd = new FileHandle $cache_file, O_WRONLY|O_CREAT;
    print $fd "$final_buf\n";

    close($fd);

    return 0;
}

sub _massage_pool_data
{
    my $recs = shift;

    my %pool_hash = ();

    foreach my $r (@{$recs})
    {
        $pool_hash{$r->{POOL_UUID}}->{id} = $r->{ID};
        $pool_hash{$r->{POOL_UUID}}->{name} = $r->{NAME};
    }

    return \%pool_hash;
}


sub _massage_share_data
{
    my $recs = shift;

    my %pool_hash = ();

    foreach my $r (@{$recs})
    {
        $r->{ID} =~ /(.*)@(.*)/;
        push(@hid, $2);
        $pool_hash{$r->{ID}}->{dest_id} = $r->{ID};
        $pool_hash{$r->{ID}}->{dest_name} = $r->{SHARE_ID};
        $pool_hash{$r->{ID}}->{dest_type} = 'FILESYSTEM';
        $pool_hash{$r->{ID}}->{dest_container_id} = $r->{HOST_ID};
        $pool_hash{$r->{ID}}->{dest_container_name} = $r->{NAME};
        $pool_hash{$r->{ID}}->{dest_container_type} = 'HOST';
        $pool_hash{$r->{ID}}->{dest_container_ip} = $r->{IP};
    }

    return \%pool_hash;
}

sub correlate_nas_share_host
{
    my ($agent, $ip_list, $e_vdid, $enc_name, $enc_type, $family) = @_;

    my $ret = 0;
    my $pool_hash = undef;

    ($ret, $pool_hash) = pull_host_share_data($agent);

    return 1 if ($ret != 0 or not defined $pool_hash);

    # Seems like we have all the data is required to perform corelation.

    my $nfs_export_tb = $agent->gettable("P_ARRAY_NAS_SHARE");
    my $nfs_rows;
    #Netapp CDOT: Get the rows only for the specified vserver
    if ((defined($family)) && ($family eq 'VRP_NETAPP_CDOT')) {
        my @map_cols =
            (&VRTS::Constants::P_ARRAY_NAS_SHARE::SUBSYS_ID,
            &VRTS::Constants::P_ARRAY_NAS_SHARE::MOUNT_POINT_TYPE);
        my @map_vals = ($e_vdid, 'NFS');
        $nfs_rows = $nfs_export_tb->getallrowswithvaluez(\@map_cols, \@map_vals);
    } else {
        $nfs_rows = $nfs_export_tb->getallrowswithvalue(&VRTS::Constants::P_ARRAY_NAS_SHARE::MOUNT_POINT_TYPE, 'NFS');
    }
    return if (not defined $nfs_rows);

    my $dev_assg_tb = $agent->gettable("P_ARRAY_DEVICE_ASSIGNMENT");
    foreach my $nfs_r (@{$nfs_rows})
    {
      # Try and construct the possible url that is obtained from the Host discovery side.
      foreach my $host (@hid) {
        foreach my $tmp_ip (@{$ip_list})
        {
            my $share_id;
            if ((defined($family)) && ($family eq 'VRP_NETAPP_CDOT')) {
                $share_id = correlated_share_host($pool_hash, $nfs_r->[&VRTS::Constants::P_ARRAY_NAS_SHARE::PATH], $host, 
                                        $tmp_ip, $context->{$enc_name}, 'VRP_NETAPP_CDOT');
            } else {
                $share_id = correlated_share_host($pool_hash, $nfs_r->[&VRTS::Constants::P_ARRAY_NAS_SHARE::PATH], $host, $tmp_ip);
            }
            if (defined($share_id))
            {
                _report_host_device_assignment($agent, $dev_assg_tb,
                      $nfs_r->[&VRTS::Constants::P_ARRAY_NAS_SHARE::ID],
                      "NFS_SHARE",
                      $nfs_r->[&VRTS::Constants::P_ARRAY_NAS_SHARE::NAME],
					  $pool_hash, $share_id, $e_vdid, $enc_name, $enc_type,
                      $nfs_r->[&VRTS::Constants::P_ARRAY_NAS_SHARE::SOURCE_CONTAINER_ID]);
                last;
            }
        }
        # This is to handle AIX case or any other case where the share_id
        # matches format /share:@hostid. Ex:
        # "/APPS_POOL_NAS_FS@{00090011-25c5-7758-0000-00000ad159a8}"
        my $share_id = $nfs_r->[&VRTS::Constants::P_ARRAY_NAS_SHARE::PATH] .
                       "@".$host;
        if (defined $pool_hash->{$share_id})
        {
            _report_host_device_assignment($agent, $dev_assg_tb,
                    $nfs_r->[&VRTS::Constants::P_ARRAY_NAS_SHARE::ID],
                    "NFS_SHARE",
                    $nfs_r->[&VRTS::Constants::P_ARRAY_NAS_SHARE::NAME],
                    $pool_hash, $share_id, $e_vdid, $enc_name, $enc_type,
                    $nfs_r->[&VRTS::Constants::P_ARRAY_NAS_SHARE::SOURCE_CONTAINER_ID]);
        }
      }
    }
    my $cifs_rows = $nfs_export_tb->getallrowswithvalue(
		    &VRTS::Constants::P_ARRAY_NAS_SHARE::MOUNT_POINT_TYPE, "CIFS");
    return if (not defined $cifs_rows);
    foreach my $cifs_r (@{$cifs_rows})
    {
        # Try and construct the possible url that is obtained from the Host discovery side.
	foreach my $host (@hid) {
	   foreach my $tmp_ip (@{$ip_list})
	   {
	       my $share_id = "\\$tmp_ip\\$cifs_r->[&VRTS::Constants::P_ARRAY_NAS_SHARE::NAME]" . "@".$host;
	       if (defined $pool_hash->{$share_id})
	       {
		   _report_host_device_assignment($agent, $dev_assg_tb,
			    $cifs_r->[&VRTS::Constants::P_ARRAY_NAS_SHARE::ID],
			    "NFS_SHARE",
			    $cifs_r->[&VRTS::Constants::P_ARRAY_NAS_SHARE::NAME],
			    $pool_hash, $share_id, $e_vdid, $enc_name, $enc_type);
			    last;
               }
            }
        }
    }
}

sub pull_host_share_data
{
    my $agent = shift;

    $context = $agent->get_context();

    my $app_dir = VRTS::Paths::get_path("AppDir");
    $app_dir = File::Spec->catdir($app_dir, "SIData");

    eval { File::Path::mkpath($app_dir, 0, 755); };

    return (1, undef) if (! -d $app_dir);

    my $last_time = undef;
    my $time_f = File::Spec->catfile($app_dir, ".cache_status_host");
    my $cache_file = File::Spec->catfile($app_dir, "host_share.cache");
    my $cache_lock = File::Spec->catfile($app_dir, "host_share.lock");

    if (-f $time_f)
    {
        open (HAN, $time_f);
        $last_time = <HAN>;
        close(HAN);

        chomp $last_time;
    }
    else
    {
        $last_time = 0; # this will force the cache to be built up for the very first time.
    }

    my $lc_time = time;
    # Compare this with what is stampped in the file.

    # Acquire a lock and a blocking lock in this case.
    open(CACHE_LOCK, "> $cache_lock");
    flock(CACHE_LOCK, LOCK_EX);

    #Set the update period to value of context varaible 'cache_update_period'
    #else keep to default value of UPDATE_SP_CACHE
    my $update_period;
    my $cache_update_period = $context->{'cache_update_period'};
    if ((defined($cache_update_period)) && ($cache_update_period =~ /^\d+$/)) {
        $update_period = $cache_update_period;
    } else {
        $update_period = &UPDATE_SP_CACHE;
    }

    my ($ret, $op);
    if ($lc_time > ($last_time + $update_period) or $context->{update_cache} == 1)
    {
        VRTS::Logger::log('debug', "Need to refresh the cache under [$cache_file]");

        # Lets get the NFS from the database.
        my $recs = VRTS::Enclosure::Utils::pull_nfs_shares();
        return (1, undef) if (not defined $recs);

        my $pool_hash = _massage_share_data($recs);
        my $ret = _sync_cache($pool_hash, $cache_file);

        if ($ret != 0)
        {
            $ret = 1;
            $op = undef;
            VRTS::Logger::log('debug', "Nothing obtained to generate cache, skip correlation for now");
            goto SKIP_MARKER_CREATION;
        }
        else
        {
            $ret = 0;
            $op = $pool_hash;
        }

        open(HAN, "> $time_f");
        print HAN $lc_time;
        close(HAN);
    }
    else
    {
        # Lets get the NFS from the database.
        my $recs = VRTS::Enclosure::Utils::pull_nfs_shares();
        return (1, undef) if (not defined $recs);

        my $pool_hash = _massage_share_data($recs);
        VRTS::Logger::log('debug', "No need to refresh the cache, reading cache from [$cache_file]");
        ($ret, $op) = _read_cache($cache_file);

        if ($ret == 2)
        {
            VRTS::Logger::log('debug', "Invalidating the cache and its contents, along with timestamp file");
            unlink ("$time_f");
            unlink ("$cache_file");

            my @ignore_tbs = ("P_ARRAY_DEVICE_ASSIGNMENT");
            $agent->ignoretables(\@ignore_tbs);
        }
    }

SKIP_MARKER_CREATION:
    # Release the lock in this case.
    close(CACHE_LOCK);

    return $ret, $op;
}

sub correlated_share {
    my ($pool_hash, $share_path, $ip, $fqhn_to_ip_map, $family) = @_;

    if (not defined($fqhn_to_ip_map)) {
        $fqhn_to_ip_map = $context->{'fqhn_to_ip_map'};
    }

    my $saved_pool_id = undef;
    foreach my $pool_id (keys %{$pool_hash}) {
        my $fqhn;
        if ((defined($family)) && ($family eq 'VRP_NETAPP_CDOT')) {
            if ($pool_id =~ m#^netfs://([^/]*)//?$share_path/#) {
                $fqhn = $1;
            }
        } elsif ($pool_id =~ m#^netfs://(.*)/$share_path/#) {
            $fqhn = $1;
        }

        if ((not defined($fqhn)) || ($fqhn eq '')) {
            if ((defined($family)) && ($family eq 'VRP_NETAPP_CDOT')) {
                if ($pool_id =~ m#^netfs://([^/]*)/?$share_path/#) {
                    $fqhn = $1;
                }
            } elsif ($pool_id =~ m#^netfs://(.*)$share_path/#) {
                $fqhn = $1;
            }
        }
        if ((defined($ip)) && ($ip eq $fqhn)) {
            VRTS::Logger::log('debug', "IP = [$ip]. Share Path = [$share_path]");
            VRTS::Logger::log('debug', "Found correlatable share with pool id = [$pool_id]");
            $saved_pool_id = $pool_id;
            last;
        } elsif ((exists($fqhn_to_ip_map->{$fqhn})) && ($fqhn_to_ip_map->{$fqhn} ne '')) {
            VRTS::Logger::log('debug', "FQHN = [$fqhn]. Share Path = [$share_path]");
            VRTS::Logger::log('debug', "Found correlatable share with pool id = [$pool_id]");
            $saved_pool_id = $pool_id;
            last;
        }
    }
    return $saved_pool_id;
}

sub check_existing_correlation {
    my ($method, $dev_assg_tb, $curr_mnt_path, $pool_id) = @_;
    #For NETAPP, if both qtree and volume are exported then preference should
    #be given to qtree over volume to correlate and report only
    #one entry. 
    if ($method eq 'NETAPP') {
        my @tmp_cols = (&VRTS::Constants::P_ARRAY_DEVICE_ASSIGNMENT::SOURCE_TYPE,
            &VRTS::Constants::P_ARRAY_DEVICE_ASSIGNMENT::DEST_TYPE,
            &VRTS::Constants::P_ARRAY_DEVICE_ASSIGNMENT::DEST_ID);
        my @tmp_vals = ('NFS_SHARE', 'STORAGE_POOL', $pool_id);
        my $rep_row = $dev_assg_tb->getrowwithvaluez(\@tmp_cols, \@tmp_vals);
        if (defined($rep_row)) {
            #If volume is correlated and now correlating qtree then
            #remove volume correlation entry and report qtree
            #correlation entry.
            #If qtree is correlated and now correlating volume then skip
            #the correlation since preferred correlation is already
            #reported.
            my $rep_mnt_path = $rep_row->[&VRTS::Constants::P_ARRAY_NAS_SHARE::PATH];
            if ($curr_mnt_path =~ m#^$rep_mnt_path#) {
                VRTS::Logger::log('debug', 'Deleting reported corrrelation for volume and going ahead with correlation for qtree');
                $dev_assg_tb->deleteallrowswithvaluez(\@tmp_cols, \@tmp_vals);
                return 0;
            } elsif ($rep_mnt_path =~ m#^$curr_mnt_path#) {
                VRTS::Logger::log('debug', 'Skipping correlation for volume as we have already reported correlation for qtree');
                return 1;
            }
        }
    }
    return 0;
}

sub correlated_share_host {
    my ($pool_hash, $share_path, $host, $ip, $fqhn_to_ip_map, $family) = @_;
    if (not defined($fqhn_to_ip_map)) {
        $fqhn_to_ip_map = $context->{'fqhn_to_ip_map'};
    }
    my $share_id = $ip . ':' . $share_path . '@' . $host;
    if ((!exists($pool_hash->{$share_id})) && (defined($family)) && ($family eq 'VRP_NETAPP_CDOT')) {
        $share_id = $ip . ':/' . $share_path . '@' . $host;
    }
    if (exists($pool_hash->{$share_id})) {
        VRTS::Logger::log('debug', "IP = [$ip]. Share Path = [$share_path]");
        VRTS::Logger::log('debug', "Found correlatable share with share id = [$share_id]");
        return $share_id; 
    }

    foreach my $fqhn (keys %{$fqhn_to_ip_map}) {
        if ($fqhn_to_ip_map->{$fqhn} eq $ip) {
            $share_id = $fqhn . ':' . $share_path . '@' . $host;
            if ((!exists($pool_hash->{$share_id})) && (defined($family)) && ($family eq 'VRP_NETAPP_CDOT')) {
                $share_id = $fqhn . ':/' . $share_path . '@' . $host;
            }
            if (exists($pool_hash->{$share_id})) {
                VRTS::Logger::log('debug', "FQHN = [$fqhn]. Share Path = [$share_path]");
                VRTS::Logger::log('debug', "Found correlatable share with share id = [$share_id]");
                return $share_id; 
            }
        }
    }
    return undef;
}

1;
