#$Id$
# VT50-2000-2016-84-28-4
# $Copyright (c) 2017 Veritas Technologies LLC.  All rights reserved. Veritas
# and the Veritas Logo are trademarks or registered trademarks of Veritas
# Technologies LLC or its affiliates in the U.S. and other countries. Other 
# names may be trademarks of their respective owners.
#
# The Licensed Software and Documentation are deemed to be commercial computer
# software as defined in FAR 12.212 and subject to restricted rights as defined
# in FAR Section 52.227-19 "Commercial Computer Software - Restricted Rights" and
# DFARS 227.7202, et seq. "Commercial Computer Software and Commercial Computer
# Software Documentation," as applicable, and any successor regulations, whether
# delivered by Veritas as on premises or hosted services. Any use, modification,
# reproduction release, performance, display or disclosure of the Licensed Software
# and Documentation by the U.S. Government shall be solely in accordance with the
# terms of this Agreement. $

package VRTS::Enclosure::Discover::3PAR;
use strict;
use Carp;
require Exporter;
require XML::Simple;
require VRTS::Enclosure::VDID;
require VRTS::Enclosure::Utils;
require VRTS::Paths;
require VRTS::Util;

use File::Spec;
use Data::Dumper;
use JSON;

my ($xml, $vdid_obj);

my %capabilities = ("VOL_CAPABLE" => 0, "SHOW_RG_LDEV" => 1, "SHOW_RG_PDEV" => 1, "SHOW_PDEV" => 1, "SHOW_LDEV" => 1, "CONSIDER_SEGMENT_LUN" => 1,
   "SHOW_REPLICATION" => 1, "SHOW_TP_SOURCELDEV" => 0, "SHOW_TP_THINLDEV" => 1, "SHOW_LDEV_PDEV" => 1, "LDEV_PDEV_MAP" => 1,
   "PRIMARY_REMOTE_REP_CAPABLE" => 1, "SHOW_VSET" => 0, "SHOW_TP" => 1, "SHOW_RG" => 0, "SECONDARY_REMOTE_REP_CAPABLE" => 1,
   "SHOW_PDEV_FOR_THIN_LDEV" => 1, "SHOW_PDEV_FOR_TP_SOURCE_THINLDEV" => 0, "SHOW_DEVICE_GROUP_REPLICATION" => 0);



sub new 
{
    my $class = shift;
    my $self = {};

    bless ($self, $class);
    return $self;
}

sub initialize
{
    my $top = shift;

    my $module = new VRTS::Enclosure::Discover::3PAR;
    return $module;

}

sub get_enclosure_listing
{
    my $self = shift;
    my $params = shift;

    my $ch = $params->{"control-host"};
    my $id = $params->{ID};
    my $server = $params->{"server"};
    my $port = $params->{"port"};
    my $user = $params->{"user"};
    my $pass = $params->{"pass"};

    $pass = quotemeta($params->{'pass'});

    my $probe_timeout = $self->{"probe_timeout"};

    VRTS::Logger::log('trace', "control-host = [$ch]"); 
    VRTS::Logger::log('trace', "ID = [$id]"); 
    VRTS::Logger::log('trace', "server = [$server]"); 
    VRTS::Logger::log('trace', "user = [$user]"); 

    # Due to change to halssh to work in a non-interactive mode, 3PAR discovery will only work
    #  with managed host version 5.0 and above, we need to check this as part of probe and provide the correct error code.
    my $xprtlc = File::Spec->catfile(VRTS::Paths::get_path("InstallDir"), "bin", "xprtlc");

    my $my_grep = "grep";
    $my_grep = "findstr" if ($^O =~ /Win32/i);
    my $op = undef;
    my $xprtlc_cmd = VRTS::Util::make_command($xprtlc, "-l", "https://$ch/world/getvitals");
    my $xprtlout = `$xprtlc_cmd`;
    chomp $xprtlout;

    VRTS::Logger::log('trace', "COMMAND = [$xprtlc_cmd]"); 
    VRTS::Logger::log('trace', "OUTPUT = [$xprtlout]"); 

    my $jsonobject = JSON->new->allow_nonref;
    eval {
      $op = $jsonobject->decode($xprtlout);
    };

    if ($@)
    {
        return undef, &VRTS::Enclosure::Discover::MH_NOT_REACHABLE;
    }

    if (not defined $op->{XPRTLD_VERSION})
    {
        return undef, &VRTS::Enclosure::Discover::MH_NOT_REACHABLE;
    }

    if ($op->{XPRTLD_VERSION} =~ /^[5-9]\./)
    {
        VRTS::Logger::log('debug', "Managed host version is 5.0 or above, supported version");
    }
    else
    {
        VRTS::Logger::log('error', "Managed host version is less than 5.x, not supported");
        return undef, &VRTS::Enclosure::Discover::PLATFORM_UNSUPPORTED;
    }

    # First check if we support the firmware version which should be greater than 2.2.3
    # This is required as the cli options and printing for this version and must be for versions below is very different.

    my $actual_cmd = 'showversion';
    my $vardir= VRTS::Paths::get_path("AppDir");
    my $script = "$vardir/3par.pl";

    _create_probe_script($user, $server, $actual_cmd, $id, $pass, $script);

    my $out = VRTS::Enclosure::Utils::push_exec($ch, $script, "\$VARDIR\/3par.pl", $probe_timeout);
    
    VRTS::Logger::log('trace', "COMMAND = [$actual_cmd]"); 
    VRTS::Logger::log('trace', "OUTPUT = [$out->[1]]"); 

   if ($out->[0] != 0)
    {
        my $err = $out->[0] >> 8;
        if ($err == 192)
        {
            print "Request timed out due to no activity for timeout period\n";
            return undef, &VRTS::Enclosure::Discover::SESSION_TIMEDOUT;
        }
         print "Error in push_exec\n";
         return undef, &VRTS::Enclosure::Discover::MH_NOT_REACHABLE;
    }

    unlink $script;

    my $blob = $out->[1];
    my $not_supported_fmware = 1;

    my @arr_list = split /\n/, $blob;
    foreach my $line (@arr_list)
    {
        if ($line =~ /Release version/)
        {
            my $version = $line;
            VRTS::Logger::log('debug', "Firmware version detected as [$line]");
            $line =~ s/Release version\s+([0-9])\.([0-9])\.([0-9])//;
            if ($1 > 2)
            {
                $not_supported_fmware = 0;
            }
            elsif ($1 == 2 and $2 > 2)
            {
                $not_supported_fmware = 0;
            } 
            elsif ($1 == 2 and $2 <= 2 and $3 > 3)
            {
                $not_supported_fmware = 0;
            }
            else
            {
                $not_supported_fmware = 1;
            }

            last;
        }
    }

    if ($blob =~ /Connection timed out/)
    {
        return undef, &VRTS::Enclosure::Discover::ARRAY_NOT_REACHABLE;
    }
    elsif ($blob =~ /not found/ or $blob =~ /No such file/)
    {
        # SYMCLI path not found.
        return undef, &VRTS::Enclosure::Discover::VENDOR_COMP_NOT_FOUND;
    }
    elsif ($blob =~ /invalid credentials/ or $blob =~ /Authentication failur/)
    {
        return undef, &VRTS::Enclosure::Discover::AUTHENTICATION_FAILED;
    }
    elsif ($blob =~ /Too many CLI connections/ or $blob =~ /No route to host/ or $blob =~ /Host does not exist/)
    {
        return undef, &VRTS::Enclosure::Discover::ARRAY_NOT_REACHABLE;
    }
    elsif ($not_supported_fmware == 1)
    {
        return undef, &VRTS::Enclosure::Discover::ARRAY_DISCOVERY_FM_LOWER_VERSION_ERROR;
    }

    VRTS::Logger::log('debug', 
        "Firmware version on the array side is higher than the minimum requirement, proceeding with probe");

    $vdid_obj = new VRTS::Enclosure::VDID ();
    $vdid_obj->initialize("3PAR", 0);

    $actual_cmd = 'showsys';
    $vardir= VRTS::Paths::get_path("AppDir");
    $script = "$vardir/3par.pl";

    _create_probe_script($user, $server, $actual_cmd, $id, $pass, $script);

    my $out = VRTS::Enclosure::Utils::push_exec($ch, $script, "\$VARDIR\/3par.pl", $probe_timeout);

    VRTS::Logger::log('trace', "COMMAND = [$actual_cmd]"); 
    VRTS::Logger::log('trace', "OUTPUT = [$out->[1]]"); 

    if ($out->[0] != 0)
    {
        my $err = $out->[0] >> 8;
        if ($err == 192)
        {
            print "Request timed out due to no activity for timeout period\n";
            return undef, &VRTS::Enclosure::Discover::SESSION_TIMEDOUT;
        }
         print "Error in push_exec\n";
         return undef, &VRTS::Enclosure::Discover::MH_NOT_REACHABLE;
    }

    unlink $script;

    my $blob = $out->[1];

    if ($blob =~ /not found/ or $blob =~ /No such file/)
    {
        # SYMCLI path not found.
        return undef, &VRTS::Enclosure::Discover::VENDOR_COMP_NOT_FOUND;
    }

    my $enc_hash = get_enclosure_details($blob);

    unlink $script;
    return $enc_hash;
}

sub refresh_device_manager
{
    my $self = shift;
    my $params = shift;

    my $out;
    my $ch = $params->{"control-host"};
    my $ch1 = $params->{"secondary-ch"};
    my $install_path = $params->{"install-path"};
    my $array_id = $params->{"array-id"};

    return 0;
}

sub get_enclosure_details
{
    my $data = shift;

    my ($system_id, $serial_no, $model, $array_name);
    my @data_arr = split/\n/, $data;

    return undef if (not defined $data_arr[2] or $data_arr[2] eq "");

    my @enc_data = split / /, $data_arr[2];
    $system_id = $enc_data[0];

    # Converting the system id into hex for vdid formulation.
    $system_id = sprintf("%x", $system_id);

    $array_name = $enc_data[1];
    $model = $enc_data[2]. " ". $enc_data[3];
    $serial_no = $enc_data[4];

    my $vdid = $vdid_obj->get_enclosure_vdid("3PARdata", "VV", $system_id);

    my %val_hash;
    $val_hash{VDID} = $vdid;
    $val_hash{serial} = $serial_no;
    $val_hash{model} = $model;
    $val_hash{displayname} = $array_name;
    $val_hash{product} = "3PARdata";
    $val_hash{vendor} = "VV";

    my $op_hash;
    $op_hash->{$vdid} = \%val_hash;
    return $op_hash;
}

sub _create_probe_script
{
    my ($user, $server, $actual_cmd, $id, $pass, $script) = @_;

    my $r = open(my $fh, '>', $script);
    if (not defined($fh)) {
        VRTS::Logger::log('error', "Unable to open file $script for writing");
        exit $r;
    }

    binmode $fh;
    print $fh <<EOF;
#!/opt/VRTSsfmh/bin/perl

BEGIN { \@INC = ('/opt/VRTSsfmh/lib/modules') if ( \$^O !~ /Win32/i ) }

require VRTS::Paths;
require VRTS::Util;
require File::Spec;

my \$cmd = VRTS::Paths::get_path('MHBin');
\$cmd = File::Spec->catfile(\$cmd, 'halssh');
my (\$rc, \$pass) = VRTS::Util::decrypt_encoded_passwd('$pass', '$id');
if (\$rc) {
    print "decrypt_encoded_passwd failed with error code = [\$rc]";
    exit \$rc;
}

if ( \$^O !~ /Win32/i ) {
    \$ENV{'LD_LIBRARY_PATH'} = '/opt/VRTSsfmh/lib';
} else {
    \$pass = VRTS::Util::esc_special_chars(\$pass);
    \$cmd = \$cmd . '.exe';
}
if (! -x \$cmd) {
    print 'not found'; 
    return;
}

system(\$cmd, '-ssh', '-l', '$user', '-4', '$server', '-pw', \$pass, '$actual_cmd');
EOF
    close($fh);
}

sub provide_capabilities
{
    my $self = shift;

    return \%capabilities;
}

1;
