#$Id$
# VT50-2000-2016-84-28-4
# $Copyright: Copyright (c) 2016 Veritas Technologies LLC.
# All rights reserved.
#
# THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF
# VERITAS TECHNOLOGIES LLC.  USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED
# WITHOUT THE PRIOR EXPRESS WRITTEN PERMISSION OF VERITAS TECHNOLOGIES LLC.
#
# The  Licensed  Software  and  Documentation  are  deemed  to be commercial
# computer  software  as  defined  in  FAR  12.212 and subject to restricted
# rights  as  defined in FAR Section 52.227-19 "Commercial Computer Software
# - Restricted  Rights"  and  DFARS 227.7202,  et seq.  "Commercial Computer
# Software  and  Commercial Computer Software Documentation," as applicable,
# and any successor regulations, whether delivered by Veritas as on premises
# or  hosted  services.  Any  use,  modification,  reproduction  release,
# performance,  display  or  disclosure  of  the  Licensed  Software  and
# Documentation by the U.S. Government shall be solely in accordance with
# the terms of this Agreement. $
#$Source$

package VRTS::Enclosure::Discover::HPEVA;
use strict;
require XML::SAX::ExpatXS;
require VRTS::Util;
require VRTS::XMLHandler;
require VRTS::Enclosure::VDID;
require VRTS::Enclosure::Utils;

my ($xml, $enc_map);

use constant HPEVA_COMPAQ_NAME => "COMPAQ";
use constant HPEVA_CONTROLLER_NAME => "HSV";

use constant HPEVA_SSSU_DEF_PATH_WIN => "C:\Program Files\Compaq\SSSU";
use constant HPEVA_SSSU_DEF_PATH_SOL => "/opt/CPQhsv/bin/";

my $vardir= VRTS::Paths::get_path("AppDir");
my $sssu_path = undef;
my $model = undef;
my $vdid_obj;
my ($id, $pass); 

sub new
{
    my $class = shift;
    my $self = {};

    bless ($self, $class);
    return $self;
}

sub initialize
{
    my $top = shift;

    my $module = new VRTS::Enclosure::Discover::HPEVA;
    return $module;
}

# get_enclosure_listing - Uses the sssu binary
# to connect to the HP EVA CommandView and get
# the list of HP EVA enclosures seen by the
# CommandView.
# This module has proper error checking to verify
# the HP EVA CV and connectivity
# Returns the hash containing the details of
# HP EVA enclosures
sub get_enclosure_listing
{
    my $self = shift;
    my $params = shift;

    my $control_host = $params->{"control-host"};
    my $secnd_control_host = $params->{"secondary-ch"};
    $id = $params->{ID};
    my $eva_server = $params->{"server"};
    my $user = $params->{"user"};
    $pass = quotemeta($params->{"pass"});
    my $passwdfile_dir = $params->{"ARRAY_HPEVA_PASSWDFILE_DIR_PATH"};
    my $probe_timeout = $self->{"probe_timeout"};

# Default paths (as obtained from CCS)
# Windows: "C:\Program Files\Compaq\SSSU"
# Solaris: "/opt/CPQhsv/bin/"
# Linux: Need to check for Linux
# In case of CCS, if sssu cli is not found on
# Solaris, then explorer checks for package
# (pkginfo -r CPQhsv) basedir and then appends
# /CPQhsv/bin
    $sssu_path = $params->{"install-path"};

    VRTS::Logger::log('trace', "control-host = [$control_host]"); 
    VRTS::Logger::log('trace', "ID = [$id]"); 
    VRTS::Logger::log('trace', "server = [$eva_server]"); 
    VRTS::Logger::log('trace', "user = [$user]"); 
    VRTS::Logger::log('trace', "install-path = [$sssu_path]"); 

    if ((not defined $sssu_path) or ($sssu_path eq ""))
    {
        if ($^O !~ /Win32/i)
        {
            $sssu_path = HPEVA_SSSU_DEF_PATH_SOL . "/sssu";
        }
        else
        {
            $sssu_path = HPEVA_SSSU_DEF_PATH_WIN . "\\sssu.exe";
        }
    }
    $sssu_path =~ s/^\s*//;
    $sssu_path =~ s/\s*$//;
    $sssu_path =~ s/\\/\\\\/g;

    if ((defined $pass) and ($pass ne ""))
    {
        #If password is specified, set the password_file_dir to empty
        $passwdfile_dir = ""; 
    }
    elsif ((defined $passwdfile_dir) and ($passwdfile_dir ne ""))
    {
        VRTS::Logger::log('debug', "SSSU Password is not specified. Check if " .
                "CV Manager entry exists in password-file");
        VRTS::Logger::log('debug', "Password file dir is $passwdfile_dir");

        my $script = "$vardir/vom_sssu.pl";

        _create_probe_script("-l", undef, $passwdfile_dir, $script);
        my $out = VRTS::Enclosure::Utils::push_exec($control_host, $script, "\$VARDIR\/vom_sssu_probe.pl", $probe_timeout);
        unlink($script);

        VRTS::Logger::log('trace', 'COMMAND = [Manager Entry]'); 
        VRTS::Logger::log('trace', "OUTPUT = [$out->[1]]"); 

        if ($out->[0] != 0)
        {
            my $err = $out->[0] >> 8;
            if ($err == 192)
            {
                VRTS::Logger::log('error', "Request timed out due to no activity for timeout period");
                return undef, &VRTS::Enclosure::Discover::SESSION_TIMEDOUT;
            }
            VRTS::Logger::log('error', "Error in push_exec. Error $out->[0]");
            return undef, &VRTS::Enclosure::Discover::MH_NOT_REACHABLE;
        }

        if ($out->[1] =~ /Vendor CLI not found/ or $out->[1] =~ /No such file/)
        {
            # sssu path not found.
            VRTS::Logger::log('error', "Could not find SSSU CLI at the specified" .
                    " path. Exit");
            return undef, &VRTS::Enclosure::Discover::VENDOR_COMP_NOT_FOUND;
        }
        elsif(($out->[1] =~ /Failed to open password file/))
        {
            VRTS::Logger::log('error', "SSSU CLI could not find the Password-File " .
                    "in the directory Path provided");
            VRTS::Logger::log('error', "$sssu_path -l output is $out->[1]");
            return undef,
                   &VRTS::Enclosure::Discover::HPEVA_SSSU_PASSWDFILE_NOT_FOUND;
        }
        elsif (($out->[1] =~ /Password file empty/) or ($out->[1] !~ /$eva_server/i)
                or ($out->[1] !~ /$user/i))
        {
            VRTS::Logger::log('error', "Password not specified and the CV " .
                    "Manager entry does not exists in the SSSU password-file");
            VRTS::Logger::log('error', "$sssu_path -l output is $out->[1]");
            return undef,
                   &VRTS::Enclosure::Discover::HPEVA_SSSU_CV_MANAGER_NOT_DEFINED_IN_PASSWDFILE;
        }
        VRTS::Logger::log('debug', "Password not specified and the CV " .
                "Manager entry exists in the SSSU password-file");
    }
    else
    {
        VRTS::Logger::log('error', "Password not specified and also the " .
                "Directory Path to the Password-File is not specified");
        return undef,
               &VRTS::Enclosure::Discover::HPEVA_SSSU_PASSWD_PASSWDFILE_NOT_DEFINED;
    }

    VRTS::Logger::log('debug', "SSSU cli path is $sssu_path");
    my $eva_args = "select manager $eva_server username=$user";
    my $version_args = 'exit';

    my $script = "$vardir/vom_sssu.pl";
    _create_probe_script($eva_args, $version_args, $passwdfile_dir, $script);

    my $out = VRTS::Enclosure::Utils::push_exec($control_host, $script, "\$VARDIR\/vom_sssu_probe.pl", $probe_timeout);
    
    VRTS::Logger::log('trace', "COMMAND = [$eva_args]"); 
    VRTS::Logger::log('trace', "OUTPUT = [$out->[1]]"); 

    if ($out->[0] != 0)
    {
        my $err = $out->[0] >> 8;
        if ($err == 192)
        {
            VRTS::Logger::log('error', "Request timed out due to no activity for timeout period");
            return undef, &VRTS::Enclosure::Discover::SESSION_TIMEDOUT;
        }
        VRTS::Logger::log('error', "Error in push_exec. Error $out->[0]");
         return undef, &VRTS::Enclosure::Discover::MH_NOT_REACHABLE;
    }

    unlink($script);

    if ($out->[1] =~ /Vendor CLI not found/ or $out->[1] =~ /No such file/)
    {
        # sssu path not found.
        return undef, &VRTS::Enclosure::Discover::VENDOR_COMP_NOT_FOUND;
    }
    elsif ($out->[1] =~ /Error opening https connection/i)
    {
        # Under the below cases, this error is obtained
        # 1. When HP EVA CV IP is incorrect
        # 2. When HP EVA CV is down
        # 3. When HP EVA CV Username/password is incorrect
        return undef, &VRTS::Enclosure::Discover::HPEVA_HTTPS_ERROR_CONN;
    }

# In case of older versions of sssu, 'ls system'
# does not work (the command is 'show system' in such case)
# But in case of VOM, we plan to support sssu version 8.0
# and above and hence below verify if the version is less
# than 8.0, send an error message

    my @lines = split ("\n", $out->[1]);
    foreach (@lines) {
        chomp($_);
        if ($_ =~ /Version:/) {
            my @ver = split (":", $_);
            if ((defined $ver[0]) and
                (defined $ver[1])) {
                $ver[1] =~ s/^\s*//;
                $ver[1] =~ s/\s*$//;
                VRTS::Logger::log('debug', " Got SSSU version: $ver[1]");
                if ($ver[1] < 8.0) {
                    VRTS::Logger::log("SSSU version less than 8.0 not supported. Exit");
                    return undef, &VRTS::Enclosure::Discover::VENDOR_COMP_VERSION_INCOMPAT;
                }
                last;
            }
            else {
                VRTS::Logger::log("Got SSSU version [$_] not supported. Exit");
                return undef, &VRTS::Enclosure::Discover::VENDOR_COMP_VERSION_INCOMPAT;
            }
            last;
        }
    }

    my $ls_system_args = 'ls system full xml';
    _create_probe_script($eva_args, $ls_system_args, $passwdfile_dir, $script);

    $out = VRTS::Enclosure::Utils::push_exec($control_host, $script, "\$VARDIR\/vom_sssu_probe.pl", $probe_timeout);

    VRTS::Logger::log('trace', "COMMAND = [$eva_args $ls_system_args]"); 
    VRTS::Logger::log('trace', "OUTPUT = [$out->[1]]"); 

    if ($out->[0] != 0)
    { 
        my $err = $out->[0] >> 8;
        if ($err == 192)
        {
            VRTS::Logger::log('error', "Request timed out due to no activity for timeout period");
            return undef, &VRTS::Enclosure::Discover::SESSION_TIMEDOUT;
        }
        VRTS::Logger::log("Error in push_exec. Error $out->[0]");
         return undef, &VRTS::Enclosure::Discover::MH_NOT_REACHABLE;
    }

    unlink($script);

    if ($out->[1] =~ /Vendor CLI not found/ or $out->[1] =~ /No such file/)
    {
        # sssu path not found.
        return undef, &VRTS::Enclosure::Discover::VENDOR_COMP_NOT_FOUND;
    }
    elsif ($out->[1] =~ /Error opening https connection/i)
    {
        # Under the below cases, this error is obtained
        # 1. When HP EVA CV IP is incorrect
        # 2. When HP EVA CV is down
        # 3. When HP EVA CV Username/password is incorrect
        return undef, &VRTS::Enclosure::Discover::HPEVA_HTTPS_ERROR_CONN;
    }

    $vdid_obj = new VRTS::Enclosure::VDID ();
    $vdid_obj->initialize("HPEVA", 0);

    my $handler = VRTS::XMLHandler->new();
    $xml = XML::SAX::ExpatXS->new(Handler => $handler);

    $out->[1] = "<vom>\n" . $out->[1] . "\n</vom>\n";
    my %markers = ("object" => "");
    $handler->initialize(undef, \%markers, "VRTS::Enclosure::Discover::HPEVA::fill_enc_details");

    my %parser_args = (Source => {String => $out->[1]});
    eval {$xml->parse(%parser_args)};
    if ($@)
    {
        VRTS::Logger::log("Error in parsing XML");
        VRTS::Logger::log("Error is - $@");
        return undef;
    }

    if ((defined $enc_map) and (%{$enc_map}))
    {
        return $enc_map;
    }
    else
    {
        return undef;
    }
}

sub fill_enc_details
{
    my $handler = shift;
    my $token = shift;
    my $depth = shift;

    return if (not defined $handler->{DATA}->{object}->{objectwwn});

    my $hpsys = $handler->{DATA}->{object};

# Use this Object wwwn to create the vdid
# Later while discovery, the array system name (objectname)
# will be obtained using this obectwwwn

    my $vdid = $vdid_obj->get_enclosure_vdid(HPEVA_COMPAQ_NAME,
        HPEVA_CONTROLLER_NAME, $hpsys->{objectwwn});

    my %enc_hash;
    $enc_hash{VDID} = $vdid;
    $enc_hash{serial} = $hpsys->{objectwwn};
    $enc_hash{model} = $hpsys->{systemtype};
    $enc_hash{displayname} = $hpsys->{objectname};
    $enc_hash{product} = "EVA";
    $enc_hash{vendor} = "HP";

    $enc_map->{$vdid} =  \%enc_hash;

    return;

}

sub _create_probe_script
{
    my ($eva_args, $cmd_args, $sssu_passwdfile_dir, $script) = @_;

    my $r = open(my $fh, '>', $script);
    if (not defined($fh)) {
        VRTS::Logger::log('error', 'Unable to open file for writing');
        exit $r;
    }

    binmode $fh;
    print $fh <<EOF;
#!/opt/VRTSsfmh/bin/perl
BEGIN { \@INC = ('/opt/VRTSsfmh/lib/modules') if ( \$^O !~ /Win32/i ) }
require VRTS::Paths;
require File::Spec;
my \$pass = '$pass';
my \$passwdfile_dir = '$sssu_passwdfile_dir';
if (defined(\$pass) && (\$pass ne '')) {
    require VRTS::Util;
    my (\$rc, \$passwd) = VRTS::Util::decrypt_encoded_passwd(\$pass, '$id');
    if (\$rc) {
        print "decrypt_encoded_passwd failed with error code = [\$rc]";
        exit \$rc;
    }
    \$pass = "password=\$passwd";
}

print "Vendor CLI not found. Exit" if (! -x "$sssu_path");
if ((defined \$passwdfile_dir) && (\$passwdfile_dir ne '')) {
    if ( \$^O !~ /Win32/i ) {
        \$ENV{HOME}=\$passwdfile_dir;
    }
    else {
        \$ENV{USERPROFILE}=\$passwdfile_dir;
    }
}

my \$vardir = VRTS::Paths::get_path('AppDir');
my \$sssu_cmd = File::Spec->catfile(\$vardir, 'sssu.txt');
my \$r = open (my \$fh, '>', \$sssu_cmd);
if (not defined(\$fh)) {
    print "Unable to open file for writing\n";
    exit \$r;
}
my \$eva_args = '$eva_args';
if (\$eva_args eq '-l') {
    if ( \$^O !~ /Win32/i ) {
        system("$sssu_path -l");
    } 
    else {
        system("\\\"$sssu_path\\\" -l");
    }
} 
else {
    \$eva_args .= ' ' . \$pass if (defined(\$pass));
    binmode \$fh;
    print \$fh <<EOF1;
\$eva_args
$cmd_args
EOF1
    close(\$fh);
    if ( \$^O !~ /Win32/i ) {
        system("$sssu_path 'file \$sssu_cmd'");
    } 
    else {
        system("\\\"$sssu_path\\\" \\\"file \$sssu_cmd\\\"");
    }
    unlink(\$sssu_cmd);
}
EOF
    close($fh);
}

1;

