#$Id$
# VT50-2000-2016-84-28-4
# $Copyright: Copyright (c) 2016 Veritas Technologies LLC.
# All rights reserved.
#
# THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF
# VERITAS TECHNOLOGIES LLC.  USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED
# WITHOUT THE PRIOR EXPRESS WRITTEN PERMISSION OF VERITAS TECHNOLOGIES LLC.
#
# The  Licensed  Software  and  Documentation  are  deemed  to be commercial
# computer  software  as  defined  in  FAR  12.212 and subject to restricted
# rights  as  defined in FAR Section 52.227-19 "Commercial Computer Software
# - Restricted  Rights"  and  DFARS 227.7202,  et seq.  "Commercial Computer
# Software  and  Commercial Computer Software Documentation," as applicable,
# and any successor regulations, whether delivered by Veritas as on premises
# or  hosted  services.  Any  use,  modification,  reproduction  release,
# performance,  display  or  disclosure  of  the  Licensed  Software  and
# Documentation by the U.S. Government shall be solely in accordance with
# the terms of this Agreement. $

package VRTS::Enclosure::Discover::NaviSphere;
use strict;
require XML::SAX::ExpatXS;
require VRTS::Util;
require VRTS::XMLHandler;
require VRTS::Enclosure::VDID;
require VRTS::Enclosure::Utils;

my ($xml, $vdid_obj);
my %enc_hash;
my $enc_map;

my %capabilities = ("VOL_CAPABLE" => 0, "SHOW_RG_LDEV" => 1, "SHOW_RG_PDEV" => 1, "SHOW_PDEV" => 1, "SHOW_LDEV" => 1,
    "SHOW_REPLICATION" => 1, "SHOW_TP_SOURCELDEV" => 0, "SHOW_TP_THINLDEV" => 1, "SHOW_LDEV_PDEV" => 1, "LDEV_PDEV_MAP" => 0,
    "PRIMARY_REMOTE_REP_CAPABLE" => 1, "SHOW_VSET" => 1, "SHOW_TP" => 1, "SHOW_RG" => 1, "SECONDARY_REMOTE_REP_CAPABLE" => 1,
    "SHOW_PDEV_FOR_THIN_LDEV" => 0, "SHOW_PDEV_FOR_TP_SOURCE_THINLDEV" => 0,
    "SHOW_DEVICE_GROUP_REPLICATION" => 0, "STATS_CAPABLE" => 1, 'SHOW_TP' => 1,'GROUP_RG_TP' => 1, 'SHOW_TP_TLDEV' => 1, 
    "SHOW_TP_PDEV" => 1);


# Map for Scoping details
my %scope_map = ("Global" => "0", "Local" => "1", "Ldap" => "2");

my $vardir= VRTS::Paths::get_path("AppDir");
my $cli_path = undef;
my $model = undef;
my $macro_version = 6.29;
my $err_no = undef;

sub new 
{
    my $class = shift;
    my $self = {};

    bless ($self, $class);
    return $self;
}

sub initialize
{
    my $top = shift;
    
    my $module = new VRTS::Enclosure::Discover::NaviSphere;
    return $module;
}

sub get_enclosure_listing
{
    my $self = shift;
    my $params = shift;

    my $ch = $params->{"control-host"};
    my $id = $params->{ID};
    my $secfilepath = $params->{"pem-file"};
    my $scope;
    my $port = $params->{"port"};
    $port = "443" if (not defined $port or $port eq "");

    $cli_path = $params->{"install-path"};
    $cli_path = "/opt/Navisphere/bin/" if (not defined $cli_path or $cli_path eq "");

    my $pass = quotemeta($params->{'pass'});

    VRTS::Logger::log('trace', "control-host = [$ch]"); 
    VRTS::Logger::log('trace', "ID = [$id]"); 
    VRTS::Logger::log('trace', "server = [$params->{server}]"); 
    VRTS::Logger::log('trace', "port = [$port]"); 
    VRTS::Logger::log('trace', "user = [$params->{user}]"); 
    VRTS::Logger::log('trace', "scope = [$params->{'scope'}]"); 

    my $probe_timeout = $self->{"probe_timeout"};

    my $script1 = "$vardir/navicli.pl";
    _create_probe_script('-help', $script1);

    my $out1 = VRTS::Enclosure::Utils::push_exec($ch, $script1, "\$VARDIR\/navicli.pl", $probe_timeout);
    if ($out1->[0] != 0)
    {
        my $err = $out1->[0] >> 8;
        if ($err == 192)
        {
            print "Request timed out due to no activity for timeout period\n";
            return undef, &VRTS::Enclosure::Discover::SESSION_TIMEDOUT;
        }
         print "Error in push_exec\n";
         return undef, &VRTS::Enclosure::Discover::MH_NOT_REACHABLE;
    }

    if ($out1->[1] =~ /not found/ or $out1->[1] =~ /No such file/
	    or $out1->[1] =~ /The system cannot find the path specified/)
    {
        # Naviseccli path not found.
        return undef, &VRTS::Enclosure::Discover::VENDOR_COMP_NOT_FOUND;
    }
    unlink($script1);

    VRTS::Logger::log('trace', 'COMMAND = [help]'); 
    VRTS::Logger::log('trace', "OUTPUT = [$out1->[1]]"); 
    
    my $data = $out1->[1];
    $data =~ m/Revision\s*([0-9]+\.[0-9]+)\s*/;

    my $version = $1;
    if ($version < $macro_version)
    {
        # The minimum supported macro version is 6.29
        # for naviseccli  
        return undef, &VRTS::Enclosure::Discover::VENDOR_COMP_VERSION_INCOMPAT;
    }
    
    
    $vdid_obj = new VRTS::Enclosure::VDID ();
    $vdid_obj->initialize("Clariion", 0);

    if (defined $scope_map{$params->{'scope'}})
    {
        $scope = $scope_map{$params->{'scope'}};
    }
    else
    {
        $scope = 0;
    }
    if ((defined($secfilepath)) && ($secfilepath ne '')) {
        $secfilepath =~ s/^\s*//;
        $secfilepath =~ s/\s*$//;
        $secfilepath =~ s/\\/\\\\/g;
    }
    
    #When naviseccli is configured at medium security level, 
    #below command will accept certificate permanently.    

    if (defined($pass) && ($pass ne '')) {
        VRTS::Enclosure::Utils::accept_cert_for_array("-h " . $params->{'server'} . " -Port $port -User " .
            $params->{'user'}, $id, $pass, "-Scope $scope getagent",
            "$vardir/cert.pl", $cli_path  );
        my $cert = VRTS::Enclosure::Utils::push_exec($ch, "$vardir/cert.pl", "\$VARDIR\/cert.pl");
        unlink("$vardir/cert.pl");

        VRTS::Logger::log('trace', 'COMMAND = [getagent]'); 
        VRTS::Logger::log('trace', "OUTPUT = [$cert->[1]]"); 

        {
            my (@subject, @cn, %cn_hash); 
            if ( (( $cert->[1] =~ /unable\s*to\s*validate\s*the\s*identity\s*of\s*the\s*server/i ) ||  
                    ( $cert->[1] =~ /cannot\s*confirm\s*that\s*the\s*connection\s*to\s*the\s*server\s*is\s*secure/i ) 
                ) 
        
                && ( @subject = ($cert->[1] =~ /(^Subject.*CN=.*)/mg))) {

                if ($#subject > 0 ) {
                    VRTS::Logger::log('warning', 'found more than 1 message with Name Subject :  ' . join("\n", @subject) );
                }
            
                @cn = map { ($_ =~ /CN=([^,]+)/g ) } @subject ;
                VRTS::Logger::log('trace', 'Common Names for array ' . join("," , @cn) );
                undef @cn_hash{@cn};
                if (!exists $cn_hash{$params->{'server'}}) {
                    return undef, &VRTS::Enclosure::Discover::ENCLOSURE_CERT_CN_MISMATCH;
                }
            }

        }
    }
    my $script = "$vardir/navicli.pl"; 
    _create_probe_script('-xml getagent',$script, $id, $params->{'server'}, $params->{'user'}, $pass, $port, $scope, $secfilepath);

    my $out = VRTS::Enclosure::Utils::push_exec($ch, $script, "\$VARDIR\/navicli.pl", $probe_timeout);

    VRTS::Logger::log('trace', 'COMMAND = [xml getagent]'); 
    VRTS::Logger::log('trace', "OUTPUT = [$out->[1]]"); 

    if ($out->[0] != 0)
    {
        my $err = $out->[0] >> 8;
        if ($err == 192)
        {
            print "Request timed out due to no activity for timeout period\n";
            return undef, &VRTS::Enclosure::Discover::SESSION_TIMEDOUT;
        }
         print "Error in push_exec\n";
         return undef, &VRTS::Enclosure::Discover::MH_NOT_REACHABLE;
    }
    unlink($script); 

    if ($out1->[1] =~ /not found/ or $out1->[1] =~ /No such file/
	    or $out1->[1] =~ /The system cannot find the path specified/)
    {
        # Naviseccli path not found.
        return undef, &VRTS::Enclosure::Discover::VENDOR_COMP_NOT_FOUND;
    }
    elsif ($out->[1] =~ /timed out/i or $out->[1] =~ /time out/i)
    {
        # No arrays found
        # naviseccli timed out
        return undef, &VRTS::Enclosure::Discover::ARRAY_NOT_REACHABLE;
    }
    elsif ($out->[1] =~ /Authentication failed/i or $out->[1] =~ /The specified security file path doesn't exist/i)
    {
        # Authentication failed
        # username or password or scope or security file is incorrect 
        return undef, &VRTS::Enclosure::Discover::AUTHENTICATION_FAILED;
    }
    elsif ($out->[1] =~ /security role privilege violation/i)
    {
        return undef, &VRTS::Enclosure::Discover::SECURITY_ROLE_NOT_PRIVILEGE;
    }
    elsif ($out->[1] =~ /Command is not supported/i or $out->[1] =~ /Feature software is not installed/i)
    {
        # naviseccli is not supported on this array
        return undef, &VRTS::Enclosure::Discover::VENDOR_COMP_VERSION_INCOMPAT;
    }
    elsif (($out->[1] =~ /network error occurred/i) or ($out->[1] =~ /Error occurred/i) or
                ($out->[1] =~ /Could not connect/i) or ($out->[1] =~ /Abort/i)
                or ($out->[1] =~ /Management Server is not running/i) or 
                ($out->[1] =~ /connection refused/i) ) 
    {
        # network error occured while communicating with the array
        # port may not be accessible 
        return undef, &VRTS::Enclosure::Discover::ARRAY_AUTHORIZATION_FAILED;
    }
   
    my $handler = VRTS::XMLHandler->new();
    $xml = XML::SAX::ExpatXS->new(Handler => $handler);

    my %markers = ("PARAMVALUE" => "");
    $handler->initialize(undef, \%markers, "VRTS::Enclosure::Discover::NaviSphere::fill_enc_details");

    my %parser_args = (Source => {String => $out->[1]});
    eval {$xml->parse(%parser_args)};
    if ($@)
    { 
        print "Error in parsing XML\n";
        return undef;
    } 

    return undef, $err_no if (defined $err_no and $err_no ne "");

    return $enc_map; 
}

sub fill_enc_details
{
    my $handler = shift;
    my $token = shift;
    my $depth = shift;

    my $agent_rev = "6.26";

    return if (not defined $handler->{DATA}->{PARAMVALUE}->{VALUE});

    if($handler->{DATA}->{PARAMVALUE}->{VALUE} =~ m/Agent Rev:\s*([0-9]+\.[0-9]+)\s*/) 
    {
         if($1 < $agent_rev) 
         {
             $err_no = &VRTS::Enclosure::Discover::VENDOR_COMP_VERSION_INCOMPAT;
         }
    }

    if ($handler->{DATA}->{PARAMVALUE}->{NAME} eq "Agent Rev") 
    {
        my $fw_version = $handler->{DATA}->{PARAMVALUE}->{VALUE};
        $fw_version =~ m/\s*([0-9]\.[0-9]+)\./;
        $fw_version = $1;

        if($fw_version < $agent_rev) 
        {
            $err_no = &VRTS::Enclosure::Discover::VENDOR_COMP_VERSION_INCOMPAT;
        }
    }
    if ($handler->{DATA}->{PARAMVALUE}->{NAME} eq "Model")
    {
        $model = $handler->{DATA}->{PARAMVALUE}->{VALUE};
    }
    elsif ($handler->{DATA}->{PARAMVALUE}->{NAME} eq "Serial No") 
    { 
        my $vdid = $vdid_obj->get_enclosure_vdid("EMC", "CLARiiON",
                                 $handler->{DATA}->{PARAMVALUE}->{VALUE});
        
        $enc_hash{VDID} = $vdid;
        $enc_hash{serial} = $handler->{DATA}->{PARAMVALUE}->{VALUE};

        if (defined $model) 
        {
            $enc_hash{model} = $model;
        }

        $enc_hash{displayname} = $handler->{DATA}->{PARAMVALUE}->{VALUE};
        $enc_hash{product} = "CLARiiON";
        $enc_hash{vendor} = "EMC"; 

        $enc_map->{$vdid} =  \%enc_hash;
     }
}    

sub _create_probe_script {
    my ($cmd, $script, $id, $server, $user, $pass, $port, $scope, $secfilepath) = @_;
    my $unix_cli = "$cli_path/naviseccli";
    my $win_cli = "$cli_path/NaviSECCli.exe";

    $win_cli =~ s/^\s*//;
    $win_cli =~ s/\s*$//;
    $win_cli =~ s/\\/\\\\/g;

    $unix_cli =~ s/^\s*//;
    $unix_cli =~ s/\s*$//;
    $unix_cli =~ s/\\/\\\\/g;

    my $r = open(my $fh, '>', $script);
    if (not defined($fh)) {
        VRTS::Logger::log('error', "Unable to open file $script for writing");
        exit $r;
    }

    binmode $fh;
    print $fh <<EOF;
#!/opt/VRTSsfmh/bin/perl
BEGIN { \@INC = ('/opt/VRTSsfmh/lib/modules') if ( \$^O !~ /Win32/i ) }
my \$cmd = '$cmd';
my \$secfilepath = "$secfilepath";
if ((defined(\$cmd)) && (\$cmd eq '-help')) {
    if ( \$^O !~ /Win32/i ) {
        system('$unix_cli', \$cmd);
    } else {
        system("$win_cli", \$cmd);
    }
} elsif ((defined(\$cmd)) && (\$cmd eq '-xml getagent')) {
    require VRTS::Util;
    my \$sys_cmd;
    my (\$rc, \$pass);
    if ((not defined(\$secfilepath)) || (\$secfilepath eq '')) {
        (\$rc, \$pass) = VRTS::Util::decrypt_encoded_passwd('$pass', '$id');
        if (\$rc) {
            print "decrypt_encoded_passwd failed with error code = [\$rc]";
            exit \$rc;
        }
    }
    if ( \$^O !~ /Win32/i ) {
        if (! -x '$unix_cli') {
            print 'not found'; 
            return;
        }
        \$sys_cmd = '$unix_cli';
    } else {
        if (! -x "$win_cli") {
            print 'not found'; 
            return;
        }
        \$sys_cmd = '"$win_cli"';
        \$pass = VRTS::Util::esc_special_chars(\$pass);
    }
    \$sys_cmd .= " -h $server";
    if ((defined(\$secfilepath)) && (\$secfilepath ne '')) {
        \$sys_cmd .= " -secfilepath \$secfilepath";
    } else {
        \$sys_cmd .= " -User $user -Password \$pass -Port $port -Scope $scope";
    }    
    \$sys_cmd .= " \$cmd";
    system(\$sys_cmd);
}    
EOF
    close($fh);
}

sub provide_capabilities
{
    my $self = shift;
    print STDERR "Provide capabilities\n";
    return \%capabilities;
}

1;
