#$Id$
# VT50-2000-2016-84-28-4
# $Copyright (c) 2017 Veritas Technologies LLC.  All rights reserved. Veritas
# and the Veritas Logo are trademarks or registered trademarks of Veritas
# Technologies LLC or its affiliates in the U.S. and other countries. Other 
# names may be trademarks of their respective owners.
#
# The Licensed Software and Documentation are deemed to be commercial computer
# software as defined in FAR 12.212 and subject to restricted rights as defined
# in FAR Section 52.227-19 "Commercial Computer Software - Restricted Rights" and
# DFARS 227.7202, et seq. "Commercial Computer Software and Commercial Computer
# Software Documentation," as applicable, and any successor regulations, whether
# delivered by Veritas as on premises or hosted services. Any use, modification,
# reproduction release, performance, display or disclosure of the Licensed Software
# and Documentation by the U.S. Government shall be solely in accordance with the
# terms of this Agreement. $

package VRTS::Enclosure::Utils;
use strict;
use Carp;
require Exporter;
use File::Spec;
use Data::Dumper;
use FileHandle;
use Time::Local 'timelocal_nocheck';
use Date::Parse;
require VRTS::Lock;

use vars qw($VERSION @ISA @EXPORT);

require VRTS::Paths;
require JSON;
require VRTS::Util;
require VRTS::Enclosure::Discover;
require VRTS::Enclosure::Faults;
require VRTS::Task;

@ISA = qw(Exporter);
@EXPORT = qw(push push_exec get_config_by_tech get_config_by_id get_arrays_for_config 
        get_ldev_details set_tp_thresholds _exec_sql exec_sql exec_sql_file get_capability_col_id 
        get_capacity_column_id check_config_health get_encl_name_using_vendorid
        setup_stat_cfg push_stat_script _get_enclosures_by_configuration
        get_ldev_details_by_target get_remote_ldev_details_by_name check_if_ha_nodes terminate_task todo_tp_monitor 
        dont_report_tables put_discovery_timestamp get_mh_version get_ldev_details_by_ldev_name get_epoch_time
        set_refresh_task_state_failed raise_discovery_fault get_enclosure_details_by_name accept_cert_for_array 
        get_enclosure_vdid_by_partner_node_id get_ldev_name_by_id get_remote_devicegroup_details 
        get_remote_filesystem_details read_file write_file sec_to_hhmmss get_fqdn);
$VERSION = '3.0';

my %sanitize_hash = ("METHOD" => "method", "DEVICE_MANAGER" => "server", "SECONDARY_HOST" => "secondary-ch",
     "PASSWD" => "pass", "CONTROL_HOST" => "control-host", "SCOPE" => "scope", "INSTALL_DIR" => "install-path",
     "PORT" => "port", "USERNAME" => "user", "PEMFILE" => "pem-file");

my %method_map = ("HICOMMAND" => "HiCommand", "NAVISPHERE" => "NaviSphere");

my %stat_keymap = (
    DEVICE_MANAGER  => 'DeviceManager',
    PASSWD          => 'Password',
    SCOPE           => 'Scope',
    INSTALL_DIR     => 'InstallDir',
    PORT            => 'Port',
    USERNAME        => 'UserName',
    PEMFILE         => 'Pem-File',
    ARRAY_VNX_IP_BLOCK => 'BlockIP',
    ARRAY_VNX_BLOCK_ENCLOSURE_ID => 'BlockID'

 );


my %capacities_col_map = ("physical_size" => {"index" => 1, "prop_name" => "PHYSICAL_SIZE"},
     "allocated_size" => {"index" => 2, "prop_name" => "ALLOCATED_SIZE"},
     "spare_size" => {"index" => 3, "prop_name" => "SPARE_SIZE"},
     "unallocated_size" => {"index" => 4, "prop_name" => "UNALLOCATED_SIZE"},
     "rg_total_size" => {"index" => 5, "prop_name" => "RG_TOTAL_SIZE"},
     "rg_addressable_size" => {"index" => 6, "prop_name" => "RG_ADDRESSABLE_SIZE"},
     "rg_overhead_size" => {"index" => 7, "prop_name" => "RG_OVERHEAD_SIZE"},
     "rg_free_size" => {"index" => 8, "prop_name" => "RG_FREE_SIZE"},
     "logical_total_size" => {"index" => 9, "prop_name" => "LOGICAL_TOTAL_SIZE"},
     "logical_configured_size" => {"index" => 10, "prop_name" => "LOGICAL_CONFIGURED_SIZE"},
     "logical_assigned_size" => {"index" => 11, "prop_name" => "LOGICAL_ASSIGNED_SIZE"},
     "logical_unassigned_size" => {"index" => 12, "prop_name" => "LOGICAL_UNASSIGNED_SIZE"},
     "logical_admin_size" => {"index" => 13, "prop_name" => "LOGICAL_ADMIN_SIZE"},
     "logical_reptarget_size" => {"index" => 14, "prop_name" => "LOGICAL_REPTARGET_SIZE"},
     "subs_logical_total_size" => {"index" => 15, "prop_name" => "SUBS_LOGICAL_TOTAL_SIZE"},
     "subs_logical_configured_size" => {"index" => 16, "prop_name" => "SUBS_LOGICAL_CONFIGURED_SIZE"},
     "subs_logical_assigned_size" => {"index" => 17, "prop_name" => "SUBS_LOGICAL_ASSIGNED_SIZE"},
     "subs_logical_unassigned_size" => {"index" => 18, "prop_name" => "SUBS_LOGICAL_UNASSIGNED_SIZE"},
     "subs_logical_admin_size" => {"index" => 19, "prop_name" => "SUBS_LOGICAL_ADMIN_SIZE"},
     "subs_logical_reptarget_size" => {"index" => 20, "prop_name" => "SUBS_LOGICAL_REPTARGET_SIZE"},
     "ext_physical" => {"index" => 34, "prop_name" => "EXT_PHYSICAL"},
     "rg_nas_file_capacity" => {"index" => 48, "prop_name" => "RG_FILE_CAPACITY"},
     "rg_block_capacity" => {"index" => 49, "prop_name" => "RG_BLOCK_CAPACITY"},
     "logical_nas_filesystem_free" => {"index" => 50, "prop_name" => "LOGICAL_NAS_FILESYSTEM_FREE"},
     "logical_nas_used" => {"index" => 51, "prop_name" => "LOGICAL_NAS_USED"},
     "array_vol_capacity" => {"index" => 56, "prop_name" => "ARRAY_VOL_CAPACITY"},
     "array_vol_allocated" => {"index" => 57, "prop_name" => "ARRAY_VOL_ALLOCATED"},
     "array_vol_available" => {"index" => 58, "prop_name" => "ARRAY_VOL_AVAILABLE"},
     "volume_snap_reserve_cap" => {"index" => 59, "prop_name" => "VOLUME_SNAP_RESV_CAP"},
     "volume_overhead" => {"index" => 60, "prop_name" => "VOLUME_OVERHEAD"},
     "overwrite_reserve_logical" => {"index" => 61, "prop_name" => "VOLUME_OVERWRITE_RESERVE"},
     "physical_overhead_size" => {"index" => 71, "prop_name" => "PHYSICAL_OVERHEAD_SIZE"}
);

my %capabilities_col_map = ("VOL_CAPABLE" => {"index" => 21, "prop_name" => "VOL_CAPABLE"},
     "SHOW_RG_LDEV" => {"index" => 22, "prop_name" => "SHOW_RG_LDEV"},
     "SHOW_RG_PDEV" => {"index" => 23, "prop_name" => "SHOW_RG_PDEV"},
     "SHOW_PDEV" => {"index" => 24, "prop_name" => "SHOW_PDEV"},
     "SHOW_LDEV" => {"index" => 25, "prop_name" => "SHOW_LDEV"},
     "SHOW_REPLICATION" => {"index" => 26, "prop_name" => "SHOW_REPLICATION"},
     "SHOW_TP_SOURCELDEV" => {"index" => 27, "prop_name" => "SHOW_TP_SOURCELDEV"},
     "SHOW_TP_THINLDEV" => {"index" => 28, "prop_name" => "SHOW_TP_THINLDEV"},
     "SHOW_LDEV_PDEV" => {"index" => 29, "prop_name" => "SHOW_LDEV_PDEV"},
     "PRIMARY_REMOTE_REP_CAPABLE" => {"index" => 30, "prop_name" => "PRIMARY_REMOTE_REP_CAPABLE"},
     "SHOW_VSET" => {"index" => 31, "prop_name" => "SHOW_VSET"},
     "SHOW_TP" => {"index" => 32, "prop_name" => "SHOW_TP"},
     "SHOW_RG" => {"index" => 33, "prop_name" => "SHOW_RG"},
     "SHOW_PDEV_FOR_THIN_LDEV" => {"index" => 35, "prop_name" => "SHOW_PDEV_FOR_THIN_LDEV"},
     "SHOW_PDEV_FOR_TP_SOURCE_THINLDEV" => {"index" => 36, "prop_name" => "SHOW_PDEV_FOR_TP_SOURCE_THINLDEV"},
     "SHOW_DEVICE_GROUP_REPLICATION" => {"index" => 37, "prop_name" => "SHOW_DEVICE_GROUP_REPLICATION"},
     "SHOW_NAS_FS_REPLICATION" => {"index" => 70, "prop_name" => "SHOW_NAS_FS_REPLICATION"},
     "SECONDARY_REMOTE_REP_CAPABLE" => {"index" => 38, "prop_name" => "SECONDARY_REMOTE_REP_CAPABLE"},
     "LDEV_PDEV_MAP" => {"index" => 39, "prop_name" => "LDEV_PDEV_MAP"},
     "SHOW_VFILER" => {"index" => 42, "prop_name" => "SHOW_VFILER"},
     "SHOW_RANK" => {"index" => 52, "prop_name" => "SHOW_RANK"},
     "NAS_CAPABLE" => {"index" => 53, "prop_name" => "NAS_CAPABLE"},
     "STATS_CAPABLE" => {"index" => 54, "prop_name" => "STATS_CAPABLE"},
     "CONSIDER_SEGMENT_LUN" => {"index" => 62, "prop_name" => "CONSIDER_SEGMENT_LUN"},
     "SHOW_NODE" => {"index" => 63, "prop_name" => "SHOW_NODE"},
     "SHOW_NODE_CAPACITY" => {"index" => 64, "prop_name" => "SHOW_NODE_CAPACITY"},
     "SHOW_INITIATOR" => {"index" => 65, "prop_name" => "SHOW_INITIATOR"},
     "SHOW_PORT" => {"index" => 66, "prop_name" => "SHOW_PORT"},
     "SHOW_VFILER_OVERVIEW" => {"index" => 67, "prop_name" => "SHOW_VFILER_OVERVIEW"},
     "SHOW_QTREE_FILE_CAP" => {"index" => 68, "prop_name" => "SHOW_QTREE_FILE_CAP"},
     "SHOW_TIER" => {"index" => 55, "prop_name" => "SHOW_TIER"},
     "SHOW_PHYSICAL_OVERHEAD_SIZE" => {"index" => 72, "prop_name" => "SHOW_PHYSICAL_OVERHEAD_SIZE"},
     "SHOW_LOGICAL_PDEV" => {"index" => 73, "prop_name" => "SHOW_MDISK_PHYSICAL_DISK"},
     "SHOW_TP_TLDEV" => {"index" => 74, "prop_name" => "SHOW_TP_TLDEV"},
     "GROUP_RG_TP" => {"index" => 75, "prop_name" => "GROUP_RG_TP"},
     "HIDE_SV_RAID_OVERHEAD" => {"index" => 77,"prop_name" => "HIDE_SV_RAID_OVERHEAD"},
     "SHOW_TP_PDEV" => {"index" => 76, "prop_name" => "SHOW_TP_PDEV"});

# physical disk vendor media types
use constant {
     PD_VENDOR_MEDIA_SED => "SED",
     PD_VENDOR_MEDIA_FC => "FC",
     PD_VENDOR_MEDIA_SATA => "SATA",
     PD_VENDOR_MEDIA_BD => "BD",
     PD_VENDOR_MEDIA_SAS => "SAS",
     PD_VENDOR_MEDIA_NL_SAS => "NL-SAS",
     PD_VENDOR_MEDIA_SAS_SED => "SAS(SED)"
};
use constant VENDOR_PD_MEDIA_TYPES =>
[&PD_VENDOR_MEDIA_FC,&PD_VENDOR_MEDIA_SATA,&PD_VENDOR_MEDIA_BD,
 &PD_VENDOR_MEDIA_SAS,&PD_VENDOR_MEDIA_SAS_SED, &PD_VENDOR_MEDIA_SED , 
 &PD_VENDOR_MEDIA_NL_SAS ];

# physical disk media types
use constant {
   PD_MEDIA_HDD => "HDD",
   PD_MEDIA_SSD => "SSD"
};
use constant PD_MEDIA_TYPES => [&PD_MEDIA_HDD, &PD_MEDIA_SSD];

# Adapter types
use constant {
   ADAPTER_TYPE_FC => "FC",
   ADAPTER_TYPE_ISCSI => "ISCSI",
   ADAPTER_TYPE_SAS => "SAS",
   ADAPTER_TYPE_FCOE => "FCOE"
};
use constant ADAPTER_TYPES => [&ADAPTER_TYPE_FC, &ADAPTER_TYPE_ISCSI, 
 &ADAPTER_TYPE_SAS, &ADAPTER_TYPE_FCOE];

# replication modes
use constant {
   REP_MODE_SYNC => "Sync",
   REP_MODE_ASYNC => "Async"
};
use constant REPLICATION_MODES => [&REP_MODE_SYNC, &REP_MODE_ASYNC];

# replication status
use constant {
     REP_STATE_SPLIT => "Split",
     REP_STATE_SYNC => "Synchronized",
     REP_STATE_COPYONWRITE => "CopyOnWrite",
     REP_STATE_COPIED => "Copied",
     REP_STATE_INACTIVE => "Inactive",
     REP_STATE_REV_SYNCING => "ReverseSynchronizing",
     REP_STATE_REV_OUT_OF_SYNC => "ReverseOutofSync",
     REP_STATE_SYNCING => "Synchronizing",
     REP_STATE_OUT_OF_SYNC => "OutofSync",
     REP_STATE_CONSISTENT => "Consistent",
     REP_STATE_ACTIVE => "Active",
     REP_STATE_SUSPENDED => "Suspended",
     REP_STATE_FAILED_OVER => "FailedOver",
     REP_STATE_INIT => "Initialization",
     REP_STATE_UNSYNC => "Unsynchronized",
     REP_STATE_PARTITIONED => "Partitioned",
     REP_STATE_PRECOPY => "PreCopy",
     REP_STATE_COPY_IN_PROGRESS => "CopyInProgress",
     REP_STATE_COPY_ON_ACCESS => "CopyOnAccess"
};
use constant REPLICATION_STATES =>
[ &REP_STATE_SPLIT."=0", &REP_STATE_SYNC."=3", &REP_STATE_COPYONWRITE."=4", 
  &REP_STATE_COPIED."=5", &REP_STATE_INACTIVE."=6", &REP_STATE_REV_SYNCING."=7", 
  &REP_STATE_REV_OUT_OF_SYNC."=8" , &REP_STATE_SYNCING."=9", 
  &REP_STATE_OUT_OF_SYNC."=10", &REP_STATE_CONSISTENT."=11", &REP_STATE_ACTIVE."=12", 
  &REP_STATE_SUSPENDED."=13", &REP_STATE_FAILED_OVER."=14", &REP_STATE_INIT."=15", 
  &REP_STATE_UNSYNC."=16", &REP_STATE_PARTITIONED."=18", &REP_STATE_PRECOPY."=19", 
  &REP_STATE_COPY_IN_PROGRESS."=20", &REP_STATE_COPY_ON_ACCESS."=21"];

# Replication access modes
use constant {
   REP_ACCESS_MODE_RDONLY => "Rdonly",
   REP_ACCESS_MODE_RDWRITE => "Rdwrite",
   REP_ACCESS_MODE_NONE => "-"
};
use constant REPLICATION_ACCESS_MODES =>
[&REP_ACCESS_MODE_RDONLY."=0", &REP_ACCESS_MODE_RDWRITE."=1", &REP_ACCESS_MODE_NONE."=2"];

#We create this file in enclosure vardir, for which we do metering.
#This file contains JSON data like which is currnt CH and when 
#we fetched array stats from CH to CMS last time. 

use constant STATS_MERGE_FILE => '.stats_merge.json' ;



sub get_capacities_hash
{
    return \%capacities_col_map;
}

sub get_capabilities_hash
{
    return \%capabilities_col_map;
}

sub pull_nfs_storage_pools
{
    my $nfs_sql = "SELECT ID,NAME,POOL_UUID FROM HABDBSYNC.VT_VIEW_STORAGE_POOL_R WHERE TYPE = 'NFS'";
    my $out = _exec_sql($nfs_sql);
}

sub pull_vardir_from_id
{
   my $ID = shift;
   my $array_sql = "SELECT VENDOR_ID FROM P_ARRAY_CONFIG_ARRAY_MAPPING WHERE ID = '$ID'";
   my $out = _exec_sql($array_sql);
   VRTS::Logger::log('debug', "pull_vardir_from_id: executing sql [$array_sql]");
   return $out;
}

#################################
# Function to pull all the shares
# consumed by a host and their hostname
# and host id which will be used by 
# agentlet to correlate
##################################
sub pull_nfs_shares
{
    my $share_sql = "SELECT SHARE_ID, ID, NAME, IP,
                    HOST_ID FROM HABDBSYNC.AM_VIEW_UI_NFS_SHARE_FILESYSTEM"; 
    
    my $out = _exec_sql($share_sql);
    VRTS::Logger::log('debug', "pull_nfs_shares: executing sql [$share_sql]");
    return $out;
}

# This return the information of remote ldev with name 
# $ldev_name which is not of enclosure with ID $enc_vdid
sub get_remote_ldev_details_by_name
{
    my $ldev_name = shift;
    my $enc_vdid = shift;

    my $sql = "SELECT ID, SUBSYS_ID FROM HABDBSYNC.P_ARRAY_LDEV WHERE NAME = '$ldev_name' and SUBSYS_ID != '$enc_vdid'";

    my $out = _exec_sql($sql);

    return $out;
}

sub get_array_details
{
    my $sql = "SELECT NAME, VDID, METHOD FROM HABDBSYNC.AM_VIEW_UI_ARRAY_LIST_CONFIG";
    my $out = _exec_sql($sql);

    return $out;
}

sub _get_xprtld_health_status
{
    my $host_name = shift;

    my $my_grep = "grep";
    $my_grep = "findstr" if ($^O =~ /Win32/i);

    my $xprtlc = File::Spec->catfile(VRTS::Paths::get_path("InstallDir"), "bin", "xprtlc");
    my $cmd = VRTS::Util::make_command($xprtlc, "-l", "https://localhost/agent/domain/enum", "|", $my_grep, $host_name);

    my $out = `$cmd`;
    chomp $out;

    VRTS::Logger::log('debug', "_get_xprtld_health_status: OP = [$out]");

    my @tmp_a = split/,/, $out;

    return $tmp_a[8] if (defined $tmp_a[8] and $tmp_a[8] ne "");
    return "DN";
}

sub _get_enclosures_by_configuration
{
    my $config_id = shift;

    my $sql = "SELECT METHOD, VENDOR_ID ".
          " FROM HABDBSYNC.ARRAY_VIEW_CONFIG_ARRAY_MAPPING_R,".
          " HABDBSYNC.ARRAY_VIEW_CONFIG_R ".
          " where ARRAY_VIEW_CONFIG_R.ID = '$config_id' ".
          " and ARRAY_VIEW_CONFIG_R.ID = ARRAY_VIEW_CONFIG_ARRAY_MAPPING_R.ID";

    my $op = _exec_sql($sql);
    return $op;
}

sub check_config_health
{
    my $agent = shift;

    my $context = $agent->get_context();
    my $conf_vals = read_config_params($context->{vardir});


    if (not defined $conf_vals)
    {
        VRTS::Logger::log('error', "Cant get array configuration details");
        exit 0;
    }

    my $method = $conf_vals->{method};
    if (defined $method_map{$method})
    {
        $method = $method_map{$method};
    }

    my ($out, $ret_code, $err_msg);
    my $enc_info = new VRTS::Enclosure::Discover();
    $enc_info->initialize("$method");
    # Even before we go the vendor specific module, going to check the health of the discovery hosts.
    # The primary is down and no secondary is specified, going to ignore any further required checks.

    # No check for primary control host existense as that is a must to be setup.
    my $check_sch = 1;
    my $report_fault = 0;
    my $op = _get_xprtld_health_status($conf_vals->{'control-host'});
    if ($op ne "DN")
    {
        # Primary discovery host is up.
        # Check communication between Primary and Array
        my $secondary_ch = $conf_vals->{'secondary-ch'};

        $check_sch = 0 if (not defined $secondary_ch or $secondary_ch eq "");

        $conf_vals->{'secondary-ch'} = undef; # Make sure we check with Primary only.

        ($out, $ret_code, $err_msg) = $enc_info->get_enclosure_listing($conf_vals);

        $conf_vals->{'secondary-ch'} = $secondary_ch; # Restore secondary ch value in hash map

        if (defined $ret_code and $ret_code == &VRTS::Enclosure::Discover::PROBE_SUCCESS)
        {
            VRTS::Logger::log('debug',"Basic communication with the array is okay, proceed with discovery");
            $conf_vals->{CURRENT_CH} = $conf_vals->{'control-host'};
            $check_sch = 0;
        }
        else
        {
            VRTS::Logger::log('error',"[$ret_code] Primary discovery host $conf_vals->{'control-host'} is unable to communicate with Array. Trying with Secondary");
            $report_fault = 1  if ($check_sch == 0);
        }
    }
    else
    {
        VRTS::Logger::log('error',"Primary discovery host $conf_vals->{'control-host'} is down. Trying with Secondary");
    }

    if ($check_sch == 1)
    {
        # Check if secondary discovery host is configured
        if (not defined $conf_vals->{'secondary-ch'} or $conf_vals->{'secondary-ch'} eq "")
        {
            VRTS::Logger::log('error',"Secondary discovery host not configured.");
            # Discovery host(s) are busted, no need to perform probe, report the fault on the enclosure object.
            $ret_code = &VRTS::Enclosure::Discover::MH_NOT_REACHABLE;
            $err_msg = "MH_NOT_REACHABLE";
            $report_fault = 1;
        }
        else
        {
            # Check if secondary discovery host is up
            $op = undef;
            $op = _get_xprtld_health_status($conf_vals->{'secondary-ch'});
            if ($op eq "DN")
            {
                # Discovery host(s) are busted, no need to perform probe, report the fault on the enclosure object.
                VRTS::Logger::log('error',"Secondary discovery host $conf_vals->{'secondary-ch'} is down");
                $ret_code = &VRTS::Enclosure::Discover::MH_NOT_REACHABLE;
                $err_msg = "MH_NOT_REACHABLE";
                $report_fault = 1;
            }
            else
            {
                # Secondary discovery host is up
                # Check communication bewteen Secondary and Array.
                my $primary_ch = $conf_vals->{'control-host'};
                $conf_vals->{'control-host'} = $conf_vals->{'secondary-ch'}; # Probe always checks communication
                                                                             # using $conf_vals->{'control-host'}
                $conf_vals->{'secondary-ch'} = undef; # Make sure we check using value in 'control-host'
                ($out, $ret_code, $err_msg) = $enc_info->get_enclosure_listing($conf_vals);
                # Restore Primary ch and Secondary ch in hash map
                $conf_vals->{'secondary-ch'} = $conf_vals->{'control-host'};
                $conf_vals->{'control-host'} = $primary_ch;
                if (defined $ret_code and $ret_code == &VRTS::Enclosure::Discover::PROBE_SUCCESS)
                {
                    VRTS::Logger::log('debug',"Basic communication with the array is okay, proceed with discovery");
                    $conf_vals->{CURRENT_CH} = $conf_vals->{'secondary-ch'};
                }
                else
                {
                    # Secondary is up but not able to communicate with Array
                    # Discovery hosts are busted, skip discovery
                    VRTS::Logger::log('debug',"Secondary discovery host $conf_vals->{'secondary-ch'} is unable to communicate with Array");
                    $report_fault = 1;
                }
            }
        }
    }

    # get the last update time for this array, and keep in the context so it can be used
    # while fetching the stats data
    if ($report_fault == 0) {
        $agent->set_context("array_config", $conf_vals);

        my $caps = $enc_info->provide_capabilities();
        if ((not defined $caps) || (not defined $caps->{STATS_CAPABLE})) {
            $agent->set_context("stats_capable", 0);
        } else {
            $agent->set_context("stats_capable", $caps->{STATS_CAPABLE});
        }

        $agent->set_context("FAULT_ALREADY_RAISED", 0);
    }
    else {
        $agent->set_context("_skip_discover", 1);

        VRTS::Logger::log('error',
             "Basic communication with the array is broken, need to report fault condition");
        _report_array_fault($agent,$err_msg);
    }
}


sub _report_array_fault {

    my $agent = shift;
    my $err_msg = shift;
    
    my $tb_list = $agent->get_table_report_order();
    my %skip_tbs = ("FAULT_OBJ" => "", "HOST_TO_FAULT" => "");
    my %bulkload_tbs = ("P_ARRAY_CONFIG" => "", "P_ARRAY_CONFIG_ARRAY_MAPPING" => "", "P_ARRAY_SUBSYSTEM" => "");
    foreach my $tb (@{$tb_list})
    {
        my $tb_h = $agent->gettable($tb);
        if (defined $bulkload_tbs{$tb})
        {
            $tb_h->bulkload_reported();
        }
        elsif (not defined $skip_tbs{$tb})
        {
            $tb_h->ignore();
        }
    }

    # Perform the fault computations.
    my $enc_tb = $agent->gettable("P_ARRAY_SUBSYSTEM");
    my $enc_rows = $enc_tb->each_current();

    foreach my $r (@{$enc_rows})
    {
        VRTS::Enclosure::Faults::compute_array_faults($agent, $err_msg,
                      $r->[1]->[&VRTS::Constants::P_ARRAY_SUBSYSTEM::NAME],
                      $r->[1]->[&VRTS::Constants::P_ARRAY_SUBSYSTEM::ID]);
    }

    $agent->set_context("FAULT_ALREADY_RAISED", 1);
}

sub push
{
    my $ch = shift;
    my $src = shift;
    my $dest = shift;
    my $timeout = shift;
    my $perm = shift;
    if(not defined $timeout)
    {
        $timeout = 1800;
    }
    my @out = (1, '');

    my $installdirmh = VRTS::Paths::get_path("InstallDir");
    my $xprtlc = VRTS::Util::make_command(File::Spec->catfile($installdirmh,"bin","xprtlc"));

    my $pushcmd = "$xprtlc -u vxss:///sfm_admin -l https://$ch/admin/cgi-bin/push_file.pl -f data=\@\"$src\" -d destfile='$dest' -d force=1 -t $timeout " ; 
 
   if(defined $perm) {
        $pushcmd = $pushcmd . " -d  permission=$perm";
    } else {
        $pushcmd = $pushcmd . ' -d permission=500' if ( $^O !~ /Win32/i );
    }

    $out[1] = `$pushcmd`;
    $out[0] = $?;
    return \@out;
}

sub set_tp_thresholds
{
    my $payload = shift;

    my $app_dir = VRTS::Paths::get_path("AppDir");
    my $tp_conf = undef;
    my $tp_sql = undef;
    my $tp_hash = undef;
    # We need to do couple of things here...
    ## 1) parse the JSON, construct SQLs and have them inserted into the thin pool records.
    ## 2) maintain this information on disk into a file to ensure that we dont have to go the DB very often to get these details
    while (my($key,$value) = each(%{$payload}))
    {
        while (my($k,$v) = each(%{$value}))
        {
            $tp_conf = File::Spec->catfile($app_dir, $key, $k, "thinpool.conf");
            # If the file is already there, then there are possibilities that there are
            #  some thresholds that already may be defined.
            if (-f $tp_conf)
            {
                open FILE, "<". $tp_conf;
                my $jsondata = do { local $/; <FILE> };

                my $jsonobject = new JSON ();
                $tp_hash = $jsonobject->decode($jsondata);
            }

            $tp_sql = File::Spec->catfile($app_dir, $key, $k, "thinpool.sql");
            open (SQL, "> $tp_sql");
            while (my($ik, $iv) = each(%{$v}))
            {
                $tp_hash->{$ik} = \%{$iv};

                print SQL "UPDATE HABDBSYNC.P_ARRAY_DEVICE_GROUP SET CONS_RISK_THRESHOLD=$iv->{CONS_RISK_THRESHOLD},".
                                       "CONS_FAULT_THRESHOLD=$iv->{CONS_FAULT_THRESHOLD},SUBS_RISK_THRESHOLD=$iv->{SUBS_RISK_THRESHOLD},".
                                       "SUBS_FAULT_THRESHOLD=$iv->{SUBS_FAULT_THRESHOLD} WHERE ID=\'$ik\' AND TYPE='TP';\n";

            }
            close(SQL);

            _exec_sql_file($tp_sql) if (-f $tp_sql);

            # Dump the contents of the hash into file in JSON format.
            my $final_buf = JSON::to_json($tp_hash, {pretty => 1, indent => 1});
            if (defined $final_buf)
            {
                open (HAN, "> $tp_conf");
                print HAN $final_buf;
                close(HAN);
            }
            unlink ($tp_sql) if (-f $tp_sql); # Purge the file once we done with the DB sync-up
            $tp_conf = $tp_hash = $tp_sql = undef;
            # set the scheduler.conf entry
            _set_tp_sched($key, $app_dir, $k);

        }
    }
    return 0;
}

sub read_config_params
{
    my $objid = shift;

    $objid =~ s/\"//g;

    my $config = get_config_by_vendor_id($objid);
    return undef if (not defined $config);

    return _sanitize_config_hash($config);
}

sub _sanitize_config_hash
{
    my $config = shift;
    my %final_hash = ();

    while (my($key,$value) = each(%{$config->[0]}))
    {
        if (defined $sanitize_hash{$key})
        {
            $final_hash{$sanitize_hash{$key}} = $value;
        }

        $final_hash{$key} = $value;
    }

    return \%final_hash;
}

sub read_tp_tuneables
{
    my $type = shift;
    my $vendor_id = shift;

    my $app_dir = VRTS::Paths::get_path("AppDir");   
    $app_dir = File::Spec->catdir($app_dir, $type, $vendor_id);
 
    my $tp_hash = undef; 
    my $tp_conf = File::Spec->catfile($app_dir, "thinpool.conf"); 

    if (-f $tp_conf)
    {
        open FILE, "<". $tp_conf;
        my $jsondata = do { local $/; <FILE> };
 
        my $jsonobject = new JSON ();
        $tp_hash = $jsonobject->decode($jsondata);
    }
    return $tp_hash;
}

sub push_exec
{
    my $ch = shift;
    my $src = shift;
    my $dest = shift;
    my $timeout = shift;
    my $format = shift;
    my $use_sep = shift;

    my $sitemp_dir = VRTS::Paths::get_path("AppDir"); 	 
    $sitemp_dir = File::Spec->catdir($sitemp_dir, "SITemp"); 	 
    eval ( File::Path::mkpath("$sitemp_dir", 0, 0755) );

    if(not defined $timeout)
    {
        $timeout = 1800;
    }
    my @out = (1, '', '');

    my $installdirmh = VRTS::Paths::get_path("InstallDir");
    my $xprtlc = File::Spec->catfile($installdirmh, "bin", "xprtlc");

    my $pushcmd = undef;
    my $file_name_token = File::Basename::fileparse ($dest);  # get file name from file path specified.
    $file_name_token =~ s/\..*//g; # Take out the extension.
    my $f_name = File::Spec->catfile($sitemp_dir, "cmd-op-$file_name_token-$$.output");
    my $xprtlcerr_file = File::Spec->catfile($sitemp_dir, "xprtlcerr-$file_name_token-$$.output");

    $pushcmd = VRTS::Util::make_command($xprtlc, '-u', 'vxss:///sfm_admin', '-l', "https://$ch/admin/cgi-bin/push_file.pl",
              '-f', 'data=@', $src, '-d', "destfile='$dest'", '-d', 'force=1', '-d', 'permission=700', '-d', 'exec=1');
    if (not defined ($ENV{'VOM_DISCOVERY_DEBUG'}))
    {
        $pushcmd = $pushcmd . ' -d delete=1';
    }
    my $sep = '';
    if (defined($use_sep) && ($use_sep == 1)) {
        $sep = '##XPRTLCEC##CLIOUT##CLIERR##XPRTLCERR##';
        $pushcmd = $pushcmd . " -d sep=$sep";
    }
    $pushcmd = $pushcmd . " -t $timeout";
    $pushcmd =~ s/data=\@\s+/data=\@/g;

    if (defined $format && $format == 1) {
        if ($ENV{'VOM_DISCOVERY_DEBUG'}) {
            VRTS::Logger::log('debug', "Executing $pushcmd");
        }
        if ( open (FH, "$pushcmd 2>&1 |") == 0) {
            # Failed to execute xprtlc
            $out[0] = &VRTS::Enclosure::Discover::PROBE_INTERNAL_ERROR;
            $out[1] = $!;
            VRTS::Logger::log('debug', "Failed to execute xprtlc to push the discovery script:".
                               "$out[0] :$out[1]");
            return \@out;
        }
        # Return file handle to script o/p and command return code
        $out[1] = *FH{IO};
        $out[0] = $?;
    } else {
        $pushcmd = $pushcmd . " > $f_name 2> $xprtlcerr_file";

        `$pushcmd`;
        $out[0] = $?;

        my $fh = undef;
        open $fh, "<$f_name";
        $out[1] =  do { local $/; <$fh> };
        close $fh;

        if (defined($use_sep) && ($use_sep == 1)) {
            if ($out[1] =~ /(.*)$sep(.*)/s) {
                $out[1] = $1;
                $out[2] = $2;
            }
        }

        if(defined $format and $format == 2) {
            open $fh, ">$f_name";
            print $fh $out[1];
            close $fh;
            $out[1] = $f_name;
        } else {
            unlink $f_name;
        }

        open $fh, "<$xprtlcerr_file";
        my $xerr = do { local $/; <$fh> }; 
        close $fh;
        unlink $xprtlcerr_file;

        if ($xerr ne '') {
            VRTS::Logger::log('debug', "xprtlc execution message = [$xerr]");
        }
    }

    return \@out;
}

sub get_arrays_for_config
{
    my $id = shift;

    my $sql = "SELECT * FROM HABDBSYNC.ARRAY_VIEW_CONFIG_ARRAY_MAPPING_R WHERE ID = '$id'";
    my $out = _exec_sql($sql);
    return $out;
}

sub get_config_by_id
{
    my $id = shift;

    my $sql = "SELECT * FROM HABDBSYNC.ARRAY_VIEW_CONFIG_R WHERE ID = '$id'";
    my $out = _exec_sql($sql);
    return $out;
}

sub get_config_by_tech
{
    my $tech = shift;

    my $sql = "SELECT * FROM HABDBSYNC.ARRAY_VIEW_CONFIG_R WHERE METHOD = '$tech'";
    my $out = _exec_sql($sql);
    return $out;
}

sub get_ldev_details
{
    my $hba_port = shift;
    my $scsi_lun_id = shift;
 
    my $sql = "SELECT DISTINCT(LDEV_ID), ldev.size, ldev.name, ldev.consumed_lun_size, ldev.subsys_id FROM HABDBSYNC.P_LUN_MAPPING".
                 " inner join habdbsync.p_array_ldev as ldev on ".
                 " ldev.id = LDEV_ID and HBA_PORT = '$hba_port' and SCSI_LUN_ID = '$scsi_lun_id'";

    my $out = _exec_sql($sql);
    return $out;
}
sub get_ldev_details_by_target
{
    my $hba_port = shift;
    my $scsi_lun_id = shift;
    my $tgt_hport = shift;
 
    my $sql = "SELECT DISTINCT(ldev_id), ldev.size, ldev.consumed_lun_size, ldev.is_ondemand, ldev.name, ldev.subsys_id FROM HABDBSYNC.p_lun_mapping".
                 " inner join habdbsync.p_array_ldev as ldev on ".
                 " ldev.id = ldev_id and target_id = '$hba_port' and scsi_lun_id = '$scsi_lun_id' and hba_port in ($tgt_hport)" ;
                 

    my $out = _exec_sql($sql);
    return $out;
}


sub get_config_by_vendor_id
{
    my $id = shift;
    my $sql = "SELECT * FROM HABDBSYNC.ARRAY_VIEW_CONFIG_R WHERE ID in (SELECT ID FROM HABDBSYNC.ARRAY_VIEW_CONFIG_ARRAY_MAPPING_R WHERE VENDOR_ID = '$id')";
    my $out = _exec_sql($sql);
    $out = _get_config_by_vendorid_ex($id, $out);
    return $out;
}

sub _get_config_by_vendorid_ex
{
    my $id = shift;
    my $op_data = shift;

    my $sql = "SELECT * FROM HABDBSYNC.ARRAY_VIEW_CONFIG_PARAMS_R WHERE AC_ID in (SELECT ID FROM HABDBSYNC.ARRAY_VIEW_CONFIG_ARRAY_MAPPING_R WHERE VENDOR_ID = '$id')";

    my $new_op = _exec_sql($sql);

    foreach my $elem (@{$new_op})
    {
        $op_data->[0]->{$elem->{DC_ID}} = $elem->{PARAM_VALUE};
    }

    return $op_data;
}

sub get_encl_name_using_vendorid
{
    my $vendor_id = shift;
    my $sql = "select COALESCE(NAME,DESCRIPTION) AS ENCL_NAME from HABDBSYNC.ARRAY_VIEW_SUBSYSTEM_R ".
                  " where VENDOR_ID = '$vendor_id'";

    my $new_op = _exec_sql($sql);
    if (defined $new_op->[0]->{ENCL_NAME})
    {
        return $new_op->[0]->{ENCL_NAME};
    }

    return undef;
}

sub get_arrays_all
{
    my $sql = "SELECT * FROM HABDBSYNC.ARRAY_VIEW_CONFIG_ARRAY_MAPPING_R";
    my $out = _exec_sql($sql);
    return $out;
}

sub exec_sql
{
    my $self = shift;
    my $sql = shift;

    _exec_sql($sql);
}

sub exec_sql_file
{
    my $self = shift;
    my $sql_file = shift;

    _exec_sql_file($sql_file);
}

sub _exec_sql
{
    my $sql = shift;
    if ($^O !~ /Win32/i)
    {
        $ENV{"PATH"} .= ":/opt/VRTSsfmcs/asa11/bin:/opt/VRTSsfmcs/asa11/res";
    }

    my $cmd = "/opt/VRTSsfmh/bin/xdbadm -x \"$sql\" -o /etc/vx/VRTSsfmcs/.odbc.ini -c /var/opt/VRTSsfmcs/conf -d SFMdb3 -v 2>/dev/null";
    if ($^O =~ /Win32/i)
    {
        my $installDirMH = VRTS::Paths::get_path("InstallDir");
        my $appDirCS     = VRTS::Paths::get_path("AppDir");   
        $appDirCS        =~ s/VRTSsfmh/VRTSsfmcs/g;

        $cmd = "\"".$installDirMH."bin\\xdbadm.exe"."\"".
               " -x \"$sql\"".
               " -o ".
               "\"".$appDirCS.".odbc.ini"."\"".
               " -c ".
               "\"".$appDirCS."conf"."\"".
               " -d SFMdb3 -v 2>NUL";
    }
    my $out = `$cmd`;
    if($? != 0)
    {
        return undef;
    }
    my $js = new JSON;
    my $obj = $js->decode($out);

    return $obj->{Records};
}

sub _exec_sql_file
{
    my $sql_file = shift;

    if ($^O !~ /Win32/i)
    {
        $ENV{"PATH"} .= ":/opt/VRTSsfmcs/asa11/bin:/opt/VRTSsfmcs/asa11/res";
    }

    my $cmd = "/opt/VRTSsfmh/bin/xdbadm -f \"$sql_file\" -o /etc/vx/VRTSsfmcs/.odbc.ini -c /var/opt/VRTSsfmcs/conf -d SFMdb3 -v 2>/dev/null";
    if ($^O =~ /Win32/i)
    {
        my $installDirMH = VRTS::Paths::get_path("InstallDir");
        my $appDirCS     = VRTS::Paths::get_path("AppDir");
        $appDirCS        =~ s/VRTSsfmh/VRTSsfmcs/g;

        $cmd = "\"".$installDirMH."bin\\xdbadm.exe"."\"".
               " -f \"$sql_file\"".
               " -o ".
               "\"".$appDirCS.".odbc.ini"."\"".
               " -c ".
               "\"".$appDirCS."conf"."\"".
               " -d SFMdb3 -v  ";
    }

    my $out = `$cmd`;

    return $?;
}

sub get_capacity_column_id
{
    my $key = shift;
    my %capacities_col_map = ("physical_size" => {"index" => 1, "prop_name" => "PHYSICAL_SIZE"},
            "allocated_size" => {"index" => 2, "prop_name" => "ALLOCATED_SIZE"},
            "spare_size" => {"index" => 3, "prop_name" => "SPARE_SIZE"},
            "unallocated_size" => {"index" => 4, "prop_name" => "UNALLOCATED_SIZE"},
            "rg_total_size" => {"index" => 5, "prop_name" => "RG_TOTAL_SIZE"},
            "rg_addressable_size" => {"index" => 6, "prop_name" => "RG_ADDRESSABLE_SIZE"},
            "rg_overhead_size" => {"index" => 7, "prop_name" => "RG_OVERHEAD_SIZE"},
            "rg_free_size" => {"index" => 8, "prop_name" => "RG_FREE_SIZE"},
            "logical_total_size" => {"index" => 9, "prop_name" => "LOGICAL_TOTAL_SIZE"},
            "logical_configured_size" => {"index" => 10, "prop_name" => "LOGICAL_CONFIGURED_SIZE"},
            "logical_assigned_size" => {"index" => 11, "prop_name" => "LOGICAL_ASSIGNED_SIZE"},
            "logical_unassigned_size" => {"index" => 12, "prop_name" => "LOGICAL_UNASSIGNED_SIZE"},
            "logical_admin_size" => {"index" => 13, "prop_name" => "LOGICAL_ADMIN_SIZE"},
            "logical_reptarget_size" => {"index" => 14, "prop_name" => "LOGICAL_REPTARGET_SIZE"},
            "subs_logical_total_size" => {"index" => 15, "prop_name" => "SUBS_LOGICAL_TOTAL_SIZE"},
            "subs_logical_configured_size" => {"index" => 16, "prop_name" => "SUBS_LOGICAL_CONFIGURED_SIZE"},
            "subs_logical_assigned_size" => {"index" => 17, "prop_name" => "SUBS_LOGICAL_ASSIGNED_SIZE"},
            "subs_logical_unassigned_size" => {"index" => 18, "prop_name" => "SUBS_LOGICAL_UNASSIGNED_SIZE"},
            "subs_logical_admin_size" => {"index" => 19, "prop_name" => "SUBS_LOGICAL_ADMIN_SIZE"},
            "subs_logical_reptarget_size" => {"index" => 20, "prop_name" => "SUBS_LOGICAL_REPTARGET_SIZE"},
            "ext_physical" => {"index" => 34, "prop_name" => "EXT_PHYSICAL"},
            "rg_nas_file_capacity" => {"index" => 48, "prop_name" => "RG_FILE_CAPACITY"},
            "rg_block_capacity" => {"index" => 49, "prop_name" => "RG_BLOCK_CAPACITY"},
            "logical_nas_filesystem_free" => {"index" => 50, "prop_name" => "LOGICAL_NAS_FILESYSTEM_FREE"},
            "logical_nas_used" => {"index" => 51, "prop_name" => "LOGICAL_NAS_USED"},
            "array_vol_capacity" => {"index" => 56, "prop_name" => "ARRAY_VOL_CAPACITY"},
            "array_vol_allocated" => {"index" => 57, "prop_name" => "ARRAY_VOL_ALLOCATED"},
            "array_vol_available" => {"index" => 58, "prop_name" => "ARRAY_VOL_AVAILABLE"},
            "volume_snap_reserve_cap" => {"index" => 59, "prop_name" => "VOLUME_SNAP_RESV_CAP"},
            "volume_overhead" => {"index" => 60, "prop_name" => "VOLUME_OVERHEAD"});

    return undef if (!exists $capacities_col_map{$key});
    return $capacities_col_map{$key}->{index};
}

sub get_capability_col_id
{
    my $key = shift;

    my %capabilities_col_map = ("VOL_CAPABLE" => {"index" => 21, "prop_name" => "VOL_CAPABLE"},
            "SHOW_RG_LDEV" => {"index" => 22, "prop_name" => "SHOW_RG_LDEV"},
            "SHOW_RG_PDEV" => {"index" => 23, "prop_name" => "SHOW_RG_PDEV"},
            "SHOW_PDEV" => {"index" => 24, "prop_name" => "SHOW_PDEV"},
            "SHOW_LDEV" => {"index" => 25, "prop_name" => "SHOW_LDEV"},
            "SHOW_REPLICATION" => {"index" => 26, "prop_name" => "SHOW_REPLICATION"},
            "SHOW_TP_SOURCELDEV" => {"index" => 27, "prop_name" => "SHOW_TP_SOURCELDEV"},
            "SHOW_TP_THINLDEV" => {"index" => 28, "prop_name" => "SHOW_TP_THINLDEV"},
            "SHOW_LDEV_PDEV" => {"index" => 29, "prop_name" => "SHOW_LDEV_PDEV"},
            "PRIMARY_REMOTE_REP_CAPABLE" => {"index" => 30, "prop_name" => "PRIMARY_REMOTE_REP_CAPABLE"},
            "SHOW_VSET" => {"index" => 31, "prop_name" => "SHOW_VSET"},
            "SHOW_TP" => {"index" => 32, "prop_name" => "SHOW_TP"},
            "SHOW_RG" => {"index" => 33, "prop_name" => "SHOW_RG"},
            "SHOW_PDEV_FOR_THIN_LDEV" => {"index" => 35, "prop_name" => "SHOW_PDEV_FOR_THIN_LDEV"},
            "SHOW_PDEV_FOR_TP_SOURCE_THINLDEV" => {"index" => 36, "prop_name" => "SHOW_PDEV_FOR_TP_SOURCE_THINLDEV"},
            "SHOW_DEVICE_GROUP_REPLICATION" => {"index" => 37, "prop_name" => "SHOW_DEVICE_GROUP_REPLICATION"},
            "SHOW_NAS_FS_REPLICATION" => {"index" => 70, "prop_name" => "SHOW_NAS_FS_REPLICATION"},
            "SECONDARY_REMOTE_REP_CAPABLE" => {"index" => 38, "prop_name" => "SECONDARY_REMOTE_REP_CAPABLE"},
            "LDEV_PDEV_MAP" => {"index" => 39, "prop_name" => "LDEV_PDEV_MAP"},
            "SHOW_RANK" => {"index" => 52, "prop_name" => "SHOW_RANK"},
            "SHOW_VFILER" => {"index" => 42, "prop_name" => "SHOW_VFILER"},
            "NAS_CAPABLE" => {"index" => 53, "prop_name" => "NAS_CAPABLE"},
            "VOLUME_OVERWRITE_RESERVE" => {"index" => 61, "prop_name" => "VOLUME_OVERWRITE_RESERVE"},
            "CONSIDER_SEGMENT_LUN" => {"index" => 62, "prop_name" => "CONSIDER_SEGMENT_LUN"},
            "GROUP_RG_TP" => {"index" => 75, "prop_name" => "GROUP_RG_TP"},
            "SHOW_TP_PDEV" => {"index" => 76, "prop_name" => "SHOW_TP_PDEV"},
            "HIDE_SV_RAID_OVERHEAD" => {"index" => 77,"prop_name" => "HIDE_SV_RAID_OVERHEAD"},
            "STATS_CAPABLE" => {"index" => 54, "prop_name" => "STATS_CAPABLE"});

    return undef if (!exists $capabilities_col_map{$key});
    return $capabilities_col_map{$key}->{index};
}

#
# setup_stat_cfg
#  set up the metering configuration on the current control host
#  and gather the stat log from last discovery cycle from that control host
#
sub setup_stat_cfg {
    my ($agent) = @_;

    my $ctx     = $agent->get_context();
    my $e_vdid  = $ctx->{arrayid};

    my $cap_t   = $agent->gettable('P_ARRAY_CAPABILITY');
    my $config  = $ctx->{array_config};

    my $old_module = VRTS::Logger::set_module('Enclosure::Utils::setup_stat_cfg');

    if (not defined $config) {
        # metering is not yet supported for this agent
        VRTS::Logger::log('debug', 'array config not found');
        VRTS::Logger::set_module($old_module);
        return;
    }

    if ($ctx->{stats_capable} != 1) {
        # metering not supported for this array
        VRTS::Logger::log('debug', 'stats capable flag not enabled for array');
        VRTS::Logger::set_module($old_module);
        return;
    }

    my $type    = uc($config->{METHOD});
    my $ch      = $config->{CURRENT_CH};

    my $cms_info = VRTS::Util::get_cms_config_info();

    if (($cms_info->{'ha'} == 1) && ($ch ne $cms_info->{'mh_hostname'})) {

        # check if ch and cms are ha nodes
        if (check_if_ha_nodes($cms_info->{'mh_hostname'}, $ch)) {
            $ch = $cms_info->{'mh_hostname'};
            VRTS::Logger::log('debug', "Using hostname [$ch] for metering setup, as "
                . "$config->{CURRENT_CH} is a CMS HA node");
        }
    }

    my $mcfg = {};

    # script name is derived from the family name
    my $scriptname = $type . 'stat.pl';
    my $admdir  = VRTS::Paths::get_path("ADMBin");
    my $scriptpath = File::Spec->catfile($admdir, $scriptname);


    $mcfg->{Script} = $scriptname;

    my %lease_info = get_lease_info($ch);
    if ( ! scalar( %lease_info)) {
        # could not contact control host to get local time
        VRTS::Logger::log('error', 'failed to get '
            . "vital statistics from the control host [$ch]"
            . "Skipped stat config and fetch of stat logs"
        );
        VRTS::Logger::set_module($old_module);
        return;
    }

    while(my ($k, $v) = each %{$config}) {
        if ($v eq '') {
            next;
        }
        if (defined $stat_keymap{$k}) {
            my $key =  $stat_keymap{$k};
            if ($key eq 'Password') {
                # Decrypt the password with array config id as key 
                # since passwords are now encrypted in db
                # Stat uses hostguid as key for encryption
                my ($rc, $pass) = VRTS::Util::decrypt_passwd($v, $config->{ID});
                if ($rc) {
                    VRTS::Logger::log('error', "decrypt_passwd failed with error code = [$rc]");
                    $v = "__INVALID_PASSWD__";
                } else {
                    $v = $pass;
                    # encrypt the password in $v using $lease_info{hostguid}
                    my ($ret, $enc_pswd) = VRTS::Util::get_encoded_passwd($v, $lease_info{hostguid});
                    if (defined $enc_pswd) {
                        $v = $enc_pswd;
                    } else {
                        $v = "__INVALID_PASSWD__";
                        VRTS::Logger::log('error', 'failed to encrypt the '
                            . "password string, error code [$ret]");
                    }
                }
            }

            $mcfg->{$key} = $v;
        }
    }

    $mcfg->{Lease} = $lease_info{lease};

    my $cms_host = $cms_info->{cs_name};
    $mcfg->{CMS} = $cms_host;

    my $sql = "SELECT ENABLE_STATS, STATS_SCHED FROM HABDBSYNC.P_ARRAY_CONFIG_ARRAY_MAPPING WHERE VENDOR_ID = '$e_vdid'";
    my $props = VRTS::Enclosure::Utils::_exec_sql($sql);
    my $ref = $props->[0]; 

    if( $ref->{'ENABLE_STATS'} ne '') {
        $mcfg->{Enable} = $ref->{'ENABLE_STATS'};
    }
    else {
        $mcfg->{Enable} = &JSON::true;
    }

    if ($ref->{'STATS_SCHED'} != 0) {
        $mcfg->{Schedule} = $ref->{'STATS_SCHED'};
    }
    else {
        $mcfg->{Schedule} = 30;
    }

    # push scripts only if control host is not the cms itself
    my @match = grep(/^$ch/, $cms_host);
    if (scalar(@match) == 0) {
        push_stat_script($type,$ch);
    }

    my ($rc, $out) = VRTS::Util::update_perf_config($e_vdid, $type, $ch, $mcfg);
    if (not defined $rc) {
        VRTS::Logger::log('error', 'failed to update PERF '
            . "configuration: [$out]"
        );
    }

    # Fetch the array stats from the current control host to the cms since its last update time
    # only if the control host is not the cms itself
    my $chk_cs = 1;
    if (scalar(@match) == 0) {
        # PERF data is on CH and CMS != CH.
        # We check that if array discovery agentlet has set chld_array context. If yes,
        # we would fetch stats for that array vdid and not stats for $e_vdid. 
        # Otherwise good old call to fetch_stats.

        if (exists($agent->{'CONTEXT'}->{'chld_array'}) && (defined($agent->{'CONTEXT'}->{'chld_array'}))) {
            VRTS::Logger::log('info', 'Calling fetch stats with child ids : ' . join (',' , ( @{ $agent->{'CONTEXT'}->{'chld_array'} } )));
            fetch_stats($e_vdid, $type, $lease_info{timeskew}, $ch, undef, $chk_cs, $agent->{'CONTEXT'}->{'chld_array'});
        } else {

            VRTS::Logger::log('info', 'No childs for this array calling fetch_stats with ' . $e_vdid);
            #So good old fetch_stats call. 
            fetch_stats($e_vdid, $type, $lease_info{timeskew}, $ch, undef, $chk_cs);
        }


    } else {
        # Set current CH in stats cache
        _set_stats_merge_details({type => $type, e_vdid => $e_vdid, CURRENT_CH => $ch});
    }

    VRTS::Logger::set_module($old_module);
    return;
}

#
#  get_lease_info
#  Gets lease timestamp and the guid for the control host.
#  The lease time stamp is calculated as:
#       current time on control_host + 13 hours
#
sub get_lease_info {
    my $ch = shift;
    my %info;

    # get the current time on control host.
    my $bindir = VRTS::Paths::get_path('MHBin');
    my $mytime = time;
    my $xcmd = VRTS::Util::make_command(
                            File::Spec->catfile($bindir, 'xprtlc'))
                            . " -l https://$ch:5634/world/getvitals";
    my $json = `$xcmd`;
    my $gv;
    eval {
        $gv = JSON::from_json($json);
        1;
    } or do {
        #Return an empty list
        return ();
    };

    $info{ch} = $ch;
    $info{lease} = $gv->{LOCAL_TIME} + 46800; # local time + 13 hours
    $info{hostguid} = $gv->{OSUUID};
    $info{timeskew} = $mytime - $gv->{LOCAL_TIME};

    return %info;
}

#
# fetch_stats
#  Fetch the array stats from current control host since the last discovery cycle
#
# Inputs:
#   e_vdid - array vdid
#   type   - array discovery family
#   ch     - control host
#   ch_timeskew - time skew between cms and ch
#   last_update_time - the last update time of the array
#   chk_cs - check discovery host is CMS itself 
#
sub fetch_stats {
    my ($e_vdid, $type, $ch_timeskew, $ch, $last_update_time, $chk_cs, $chld_vdid) = @_;
    my ($last_update_time_cache, $stats_merge_file_name, $stats_merge_file_dir, $lck);

    #This function is not reentrant. If multiple requests are made to see perf charts 
    #simultaneously from different user agent for a single enclosure,
    #We may end up fetching perf data from Discovery host multiple times. 
    #to make sure only one process enters this function at any given time,
    #we acquire lock first on stats_merge file.  

    ($stats_merge_file_dir, $stats_merge_file_name) = _get_stats_merge_file_name($e_vdid, $type );

    VRTS::Logger::log('info', "trying to acquire lock on dir $stats_merge_file_dir name $stats_merge_file_name");
    $lck = VRTS::Lock->new($stats_merge_file_name, $stats_merge_file_dir);
    if ($lck->trylock()) {
        VRTS::Logger::log('debug', "returning because fetch_stats already in progress for this enclosure [$e_vdid]");
        return 1;
    }

    VRTS::Logger::log('info', "acquired lock on dir $stats_merge_file_dir name $stats_merge_file_name");


    my $stats_hash = _get_stats_merge_details({e_vdid => $e_vdid, type => $type});
    if(!defined($ch) && !defined($stats_hash)) {
        VRTS::Logger::log('error', "Fetch stats: Unable to query stats merge details from cache");
        $lck->unlock();
        return;
    }
    if(!defined($ch)) {
        $ch = $stats_hash->{CURRENT_CH};
        # Untaint parameter as this could get called from CGI context
        if($ch =~ /^([\w\d\-_\.\:]+)$/) {
            $ch = $1;
        }
    }
    if(defined($stats_hash->{LAST_UPDATE})
        && defined($stats_hash->{LAST_UPDATE}->{$stats_hash->{CURRENT_CH}})) {
        $last_update_time_cache = $stats_hash->{LAST_UPDATE}->{$ch};
        # Untaint parameter as this could get called from CGI context
        if($last_update_time_cache =~ /^([\d]+)$/) {
            $last_update_time_cache = $1;
        }
    }
    if(defined($last_update_time_cache)) {
        $last_update_time = $last_update_time_cache;
    } else {
        $last_update_time = time(); # In case of first discovery, cache file will be missing
    }
    # Return if stats are already synced less than 30 minutes ago
    # Do not return in case of first discovery
    my $cur_time = time();
    if(defined($last_update_time_cache) && ($cur_time - 30*60 < $last_update_time)) {
        _set_stats_merge_details({type => $type, e_vdid => $e_vdid,
            CURRENT_CH => $ch});

        VRTS::Logger::log('info', "returning because stats are synced less than 30 mins ago  ");
        $lck->unlock();
        return;
    }
    if(!defined($ch_timeskew)) { # Determine time skew if called from cgi
        my %lease_info = get_lease_info($ch);
        $ch_timeskew = $lease_info{timeskew};
        if($ch_timeskew =~ /^([\d]+)$/) {
            $ch_timeskew = $1;
        }
    }

    VRTS::Logger::log('debug', "Fetch stats: control host [$ch], "
        . "last update time [$last_update_time]");

    if ((not defined $last_update_time) || $last_update_time eq '') {
        VRTS::Logger::log('error', "Fetch stats: cannot get last update time for [$e_vdid]");
        $lck->unlock();
        return;
    }

    # we need to account for time skew between control host time and the CMS time
    VRTS::Logger::log('debug', "Fetch stats: Time skew for control host $ch is [$ch_timeskew]");
    my $last_stats_time = $last_update_time - $ch_timeskew;

    # fetch stats from the ch
    my $bindir = VRTS::Paths::get_path('MHBin');
    my $xcmd = VRTS::Util::make_command(File::Spec->catfile($bindir, 'xprtlc'));
    my $statdir = File::Spec->catfile(VRTS::Paths::get_path('AppDir'), 'shared', 'stats');
    my $statdb = VRTS::Util::make_command(
        File::Spec->catfile($statdir, '1day.array'));
    my $insert_cmd = VRTS::Util::make_command(File::Spec->catfile($bindir, 'statlog'))
    . " --insert $statdb";

    # TODO: Pick up the list of metrics from some external conf file in future
	my @metrics = ('iops', 'iokbps');
    my @all_array_id; 
    
    
    #If we have got list of child array as well, we would meter their vdid 
    #instead of $e_vdid.  
    
    if (defined($chld_vdid) && ($#{@{$chld_vdid}} > -1) ) {
        #so we have one or more child array for this e_vdid. 
        push (@all_array_id, @{$chld_vdid});
        VRTS::Logger::log('info', "Child array " . join(':' , @{$chld_vdid} ));
    } else {
        VRTS::Logger::log('info', "No child array specified");
        push (@all_array_id, $e_vdid);
    }

    VRTS::Logger::log('info', "Going for fetching metering data from CH");
    if(-d $statdir) {
        foreach my $array_id (@all_array_id) {
            foreach my $metric (@metrics) {
                my $staturl = " -l https://$ch:5634/pub/cgi-bin/stats.pl/logdump"
                . " -d prefix=$metric.$array_id -d format=csv"
                . " -d logname=1day.array -d gt=$last_stats_time -d chkcs=$chk_cs";
                # invoke the command to get the statlog data,
                # manipulate the time stamp, insert into the local statlog
                my $fetch_cmd = $xcmd . $staturl;
                open(my $fh, '-|', $fetch_cmd) or do {
                    VRTS::Logger::log('error', "Fetch stats: cannot fork cmd to fetch stats [$fetch_cmd]");
                    $lck->unlock();
                    return;
                };

                if (not defined $fh) {
                    VRTS::Logger::log('error', "Fetch stats: cannot fetch [$metric] stats for [$array_id] from [$ch]");
                    $lck->unlock();
                    return;
                }

                my @stat_data = <$fh>;

                VRTS::Logger::log('debug', "Fetch stats: read [" . scalar(@stat_data) . "] stat records for [$metric] for array [$array_id]");

                close $fh;
                if ($?) {
                    VRTS::Logger::log('error', "Fetch stats: Error closing read pipe [$? - $!]");
                }

                $SIG{PIPE} = 'IGNORE';
                    
                VRTS::Logger::log('info', "insert command $insert_cmd ");
                open(my $ih, '|-', $insert_cmd) or do {
                    VRTS::Logger::log('error', "Fetch stats: cannot fork cmd to insert [$metric] stats [$insert_cmd for array [$array_id]");
                    $lck->unlock();
                    return;
                };

                if (not defined $ih) {
                    VRTS::Logger::log('error', "Fetch stats: cannot insert [$metric] stats for [$array_id] from [$ch]");
                    $lck->unlock();
                    return;
                }

                foreach (@stat_data) {
                    chomp;
                    $_ =~ /(\d+),(.*)/ && do {
                        # adjust the timestamp with skew
                        my $ts = $1 + $ch_timeskew;
                        print $ih "$ts,$2\n";
                    };
                }

                close $ih;
                if ($?) {
                    VRTS::Logger::log('error', "Fetch stats: Error closing write [$metric] pipe [$? - $!]");
                }
            }
        }
    }
    else {
        VRTS::Logger::log('error', "Fetch stats: Statlog directory [$statdir] not present");
    }
    _fetch_stats_err_msg($ch,\@metrics,"",$type, $chk_cs);
    _set_stats_merge_details({type => $type, e_vdid => $e_vdid,
        CURRENT_CH => $ch, 
        LAST_UPDATE => {$ch => time()}});

    VRTS::Logger::log('debug', "Fetch stats: Done fetching stats for [$e_vdid]");
    $lck->unlock();
    return;
}

sub _get_stats_merge_file_name
{
    my ($e_vdid, $type) = (@_);

    my $vardir = VRTS::Paths::get_path('AppDir');
    my $enc_vardir = File::Spec->catfile($vardir, $type, $e_vdid);

    return ($enc_vardir, STATS_MERGE_FILE);

}

sub _get_stats_merge_details {
    my($data) = @_;
    my $hash_ref;
    my $fh;
    my $stats_file;
    # Check for valid parameters
    if(!defined($data->{type}) || !defined($data->{e_vdid})) {
        return(undef);
    }
    my $family_vardir;
    my $vardir = VRTS::Paths::get_path('AppDir');
    $family_vardir = File::Spec->catfile($vardir, $data->{type}, $data->{e_vdid});

    $stats_file = File::Spec->catfile($family_vardir, STATS_MERGE_FILE);
    # Read data from existing file (if file already exists)
    open($fh, "$stats_file");
    if(defined($fh)) {
        eval {
            my($json) = <$fh>;
            $hash_ref = JSON::from_json($json);
        };
        if($@) { #Log and return undef
            VRTS::Logger::log('debug', 
                "_get_stats_merge_details : " . 
                "Error reading existing stats_merge data file [$stats_file]");
            close($fh);
            return(undef);
        }
        close($fh);
        return($hash_ref);
    }
    return(undef);
}

sub _set_stats_merge_details {
    my($data) = @_;
    my($hash_ref, $json_ref);
    my $fh;
    my $stats_file;
    my $family_vardir;
    my $vardir = VRTS::Paths::get_path('AppDir');
    $family_vardir = File::Spec->catfile($vardir, $data->{type}, $data->{e_vdid});
    $stats_file = File::Spec->catfile($family_vardir, STATS_MERGE_FILE);
    # Read data from existing file (if file already exists)
    open($fh, "$stats_file");
    if(defined($fh)) {
        eval {
            $hash_ref = JSON::from_json(<$fh>);
        };
        if($@) { #Log and ignore error in reading existing data
            VRTS::Logger::log('debug', "_set_stats_merge_details : " .
                "Error reading existing stats_merge data file [$stats_file]");
        }
        close($fh);
    }
    # Overwrite existing data read from file
    if(defined($data->{CURRENT_CH})) {
        $hash_ref->{CURRENT_CH} = $data->{CURRENT_CH};
    }
    if(defined($data->{LAST_UPDATE})) {
        my $time_hash = $data->{LAST_UPDATE};
        foreach my $key (keys%$time_hash) {
            $hash_ref->{LAST_UPDATE}->{$key} = $time_hash->{$key};
        }
    }
    eval {
        $json_ref = JSON::to_json($hash_ref);
    };
    if($@) {
        VRTS::Logger::log('error', "_set_stats_merge_details : " .
                "Error converting stats config to json: $@");
        return(1);
    }
    open($fh, ">$stats_file");
    if(!defined $fh) {
        VRTS::Logger::log('error', "_set_stats_merge_details : " .
                "Error opening stats file [$stats_file]");
        close($fh);   
        return 1;   
    }
    print $fh $json_ref;
    close($fh);
    return(0);
    
}

#
# push_stat_script
#   Push the metering script to the given control host
#
# Input
#   tech   - Technology e.g. SYMCLI or NETAPP
#   host   -  Control host
#   scriptpath - Absolute path of the script file (optional)
#
sub push_stat_script
{
    my ($tech, $host, $scriptname) = @_;
    if (not defined $tech) {
        VRTS::Logger::log('error', 'push_stat_script: invalid tech input');
        return undef;
    }

    if (not defined $host || $host eq '') {
        VRTS::Logger::log('error', 'push_stat_script: invalid control host');
        return undef;
    }

    my $perlbin =  VRTS::Util::make_command(VRTS::Paths::get_path('PerlExe'));
    my $bindir  = VRTS::Paths::get_path('MHBin');
    my $xdistc  = VRTS::Util::make_command(File::Spec->catfile(
                                            $bindir, 'xdistc'));

    if (not defined $scriptname) {

        # if $scriptname is not specified, then it is derived from
        # the input tech as <TECH>stat.pl
        $scriptname = $tech . 'stat.pl';
    }

    # the stat script is expected to be present in the ADMBin directory installed during
    # Addon setup
    my $admdir  = VRTS::Paths::get_path("ADMBin");
    my $scriptpath = File::Spec->catfile($admdir, $scriptname);
    
    if (! -f $scriptpath) {

        # no script found for this technology, perhaps stats collection
        # is not supported for it or the addon may not have installed it
        # properly at the given path
        VRTS::Logger::log('debug', "push_stat_script: skipped for [$tech], "
            . "script file [$scriptpath] not found"
        );
        return undef;
    }
    
    # the live script is expected to be present in the VRTS::Stats directory installed during
    # Addon setup
    my $livestat_script = $tech . 'stats.pm';
    my $livestat_dir = File::Spec->catdir(VRTS::Paths::get_path('InstallDir'),'lib', 'modules', 'VRTS', 'Stats');
    my $livestat_scriptpath = File::Spec->catfile($livestat_dir, $livestat_script);
    
    my $dest = "\$VARDIR/$scriptname";
    #$scriptpath = VRTS::Util::make_command($scriptpath);
    
    my $livestat_dest = "\$VARDIR/$livestat_script";
    #$livestat_scriptpath = VRTS::Util::make_command($livestat_scriptpath);
    
    my $ret = 0;
    # push the stat script to the control host
    VRTS::Logger::log('debug', "push_stat_script: pushing [$scriptpath] to host [$host]");

    my $result = VRTS::Enclosure::Utils::push($host, $scriptpath, $dest, undef, 700);
    if ($result->[0] != 0) {
        VRTS::Logger::log('error', 'push_stat_script: push to host '
               . "[$host] failed with error code $result->[0], output $result->[1]"
        );
        $ret = undef;
    }
            
    # push the livestat script to the control host
    VRTS::Logger::log('debug', "push_livestat_script: pushing [$livestat_scriptpath] to host [$host]");

    $result = VRTS::Enclosure::Utils::push($host, $livestat_scriptpath, $livestat_dest, undef, 700);
    if ($result->[0] != 0) {
        VRTS::Logger::log('error', 'push_livestat_script: push to host '
               . "[$host] failed with error code $result->[0], output $result->[1]"
        );
        $ret = undef;
    }
        
    my $var_dir = VRTS::Paths::get_path('AppDir');
    my $set_stat_script = File::Spec->catfile($var_dir, "setup_array_stat$$.pl");
    my $err = _create_arraystat_setup_script($set_stat_script, $scriptname, $livestat_script);
    if($err != 0) {
        VRTS::Logger::log('error', "push_stat_script: Error in creating $set_stat_script ");
        return undef;       
    }
    my $out = push_exec($host,$set_stat_script,"\$VARDIR\/setup_array_stat.pl");
    if($out->[0] != 0) {
        VRTS::Logger::log('error', "push_stat_script: Error in running $set_stat_script");
    }
    unlink($set_stat_script);
    return $ret;
}

#
#Creates temp script to be pushed on control host to 
#copy array stat collection scripts to appropriate location
#
sub _create_arraystat_setup_script
{
    my ($script, $stat_script, $livestat_script) = @_;
    open(my $fh, ">$script");
    if(!defined $fh) {
        return 1;   
    }
    binmode $fh;
    print $fh <<EOF;
#!/opt/VRTSsfmh/bin/perl
BEGIN { \@INC = ("/opt/VRTSsfmh/lib/modules") if ( \$^O !~ /Win32/i ) }
use File::Copy;
require File::Spec;
require VRTS::Paths;
my \$vardir = VRTS::Paths::get_path('AppDir');
my \$admdir = VRTS::Paths::get_path('ADMBin');
my \$stat_src = File::Spec->catfile(\$vardir, "$stat_script");
my \$stat_dest = File::Spec->catfile(\$admdir, "$stat_script");
copy(\$stat_src, \$stat_dest);
my \$livestat_dir = File::Spec->catdir(VRTS::Paths::get_path('InstallDir'),'lib', 'modules', 'VRTS', 'Stats');
unless(-d \$livestat_dir) { mkdir(\$livestat_dir); }
my \$livestat_src = File::Spec->catfile(\$vardir, "$livestat_script");
my \$livestat_dest = File::Spec->catfile(\$livestat_dir, "$livestat_script");
copy(\$livestat_src, \$livestat_dest);
EOF

    close($fh);
    return 0;
}

#
# _fetch_stats_err_msg
# Function for fetchin error messages from array control host
# Param : ch => control host name
#         id => identifier of the metrics
#         e_vdid => enclosure vdid
#         type => configuration type of the enclosure
#         chk_cs => check discovery host is CMS itself
#

sub _fetch_stats_err_msg {
    my $ch     = shift;
    my $id     = shift;
    my $e_vdid = shift;
    my $type   = shift;
    my $chk_cs = shift;
    my $bindir = VRTS::Paths::get_path('MHBin');
    my $xcmd = VRTS::Util::make_command(File::Spec->catfile($bindir, 'xprtlc'));

    #If id or type is not defined then return

    if ((not defined $id) || (not defined $type)) {
        VRTS::Logger::log('error',
            "_fetch_stats_err_msg : id or type is not defined");
    }
    if (not defined $e_vdid) {
        $e_vdid = "";
    }

    foreach my $metric (@$id) {

        my $staturl = " -l https://$ch:5634/pub/cgi-bin/stats.pl/fetch_err_msg"
          . " -d id=$metric.$e_vdid -d type=$type -d chkcs=$chk_cs";

  # invoke the command to get all the error messages from the array control host

        my $fetch_cmd = $xcmd . $staturl;

        VRTS::Logger::log('debug',
            " _fetch_stats_err_msg : executing [$fetch_cmd]");
        open(my $fh, '-|', $fetch_cmd) or do {
            VRTS::Logger::log('error',
                    "_fetch_stats_err_msg:cannot fork cmd to fetch err msg [$fetch_cmd]"
            );
            return;
        };

        if (not defined $fh) {
            VRTS::Logger::log('error',
                    "_fetch_stats_err_msg cannot fetch [$metric]"
                  . " error messages for [$e_vdid] from [$ch]");
            return;
        }

        my $err_msg_json = <$fh>;
        close $fh;
        if ($?) {
            VRTS::Logger::log('error',
                "_fetch_stats_err_msg: Error closing read pipe [$? -  $!]");
        }
        if (defined $err_msg_json) {
            eval {
                my $err_msg_list = JSON::from_json($err_msg_json);

                if(ref($err_msg_list) ne 'ARRAY')  {
                    VRTS::Logger::log('error',"_fetch_stats_err_msg: Recieved empty list of error messages");
                    return;
                }


          #iterate over the message list and add the messages in cms error table
                foreach my $err_msg (@$err_msg_list) {

                    #Extract the id of the object
                    if ($err_msg->{id} =~ /^($metric\S*)_$type$/) {
                        my $objid = $1;
                        my $cfg   = $type;
                        my $msg   = $err_msg->{msg};
                        if (not defined $msg) {
                            VRTS::Logger::log('error',
                                  "_fetch_stats_err: msg not defined for [$objid] [$cfg]"
                            );
                            next;
                        }
                        VRTS::Logger::log('debug',
                            "_fetch_stats_err for [$objid] [$cfg] [$msg]");
                        VRTS::Util::set_perf_error(
                            {
                                metrics => "$objid",
                                type    => "$cfg",
                                msg     => $msg
                            }
                        );
                    }
                }

            };
            if ($@) {
                VRTS::Logger::log('error',
                        "_fetch_stats_error: failed to convert the"
                      . "err_msg_json[$@]");
                return;
            }
        }
        else {
            VRTS::Logger::log('error',
                '_fetch_stats_error: failed to retrieve error messages');
        }
    }
}

#
# returns true if the given hosts are nodes of the same VCS cluster
#
sub check_if_ha_nodes {
    my ($hostname1, $hostname2) = @_;

    my $sql = "SELECT hc2.host_name FROM HABDBSYNC.SF_VIEW_HOST_TO_HC hc1"
    . " JOIN HABDBSYNC.SF_VIEW_HOST_TO_HC hc2 ON hc2.hc_id = hc1.hc_id AND hc1.ha_type = 'VCS'"
    . " AND hc1.host_name = '$hostname1' AND hc2.host_name != '$hostname1'"
    #. " JOIN HABDBSYNC.P_VCS_HOST_SG_RELATION "
    . " JOIN HABDBSYNC.P_VCS_SERVICEGROUP sg on"
    . " WHERE sg.name = 'SFM_Services'";

    my $recs = VRTS::Enclosure::Utils::_exec_sql($sql);

    if (not defined $recs || scalar(@{ $recs }) == 0) {
        return 0;
    }
    foreach my $h_ref (@{ $recs }) {
        if ($h_ref->{host_name} eq $hostname2) {
            return 1;
        }
    }
    return 0;
}

# PreDiff Hook to the Agent module.
# Used to terminate the discovery task.
sub terminate_task
{
    my $agent = shift;

    my $ctx     = $agent->get_context();
    my $taskid  = $ctx->{taskid};

    if ($taskid and $taskid ne "")
    {
        my $task = new VRTS::Task;
        my $rc = $task->load(undef,$taskid);
        if ($rc == 0) # task was loaded
        {
            my $taskinfo =
            {
                "state" => "FINISHED"
            };
            $task->update($taskid, $taskinfo);
        }
    }
}

sub _set_tp_sched
{
    my $family = shift;
    my $appdir = shift;
    my $array_id = shift;

    my $app_dir = VRTS::Paths::get_path("AppDir");
    my $array_sched_file;
    my $mh_ctl;
    my $bindir = VRTS::Paths::get_path("InstallDir");

    VRTS::Logger::log('debug', "Setting new schedule for family [$family] appdir [$appdir] array [$array_id]");

    my $sch_mh_ctl;
    if ($^O =~ m/Win32/i)
    {
        my $perl = $bindir . "\\bin\\perl.exe";

        # If thin pool thresholds has been set then forcefully trigger full disc
        # this is achieved through tp_tuneables_defined context.
        $mh_ctl = "\"$perl\" \"$bindir\\bin\\array_driver.pl\" --family $family --hidden --context \"vardir=$array_id\" --context \"vardir=$array_id:tp_tuneables_defined=1\"";
        $sch_mh_ctl = "\\\"$perl\\\" \\\"$bindir\\bin\\array_driver.pl\\\" --family $family --hidden --context vardir=$array_id";
        $array_sched_file = $app_dir . "\\array_scheduler.conf";
    }
    else
    {
        $mh_ctl = "/opt/VRTSsfmh/bin/array_driver.pl --family $family --hidden --context \"vardir=$array_id\" --context \"vardir=$array_id:tp_tuneables_defined=1\"";
        $sch_mh_ctl = "/opt/VRTSsfmh/bin/array_driver.pl --family $family --hidden --context vardir=$array_id";
        $array_sched_file = $app_dir . "/array_scheduler.conf";
    }

    # run discovery in background
    if ($^O =~ /Win32/i) 
    {
        require Win32::Process;
        #Run cmd in background
        my $child_proc;
        my $perl_install_dir = $bindir."bin\\perl.exe";
        Win32::Process::Create($child_proc,
            $perl_install_dir,
            $mh_ctl,
            0,
            &Win32::Process::DETACHED_PROCESS |
            &Win32::Process::CREATE_NO_WINDOW |
            &Win32::Process::NORMAL_PRIORITY_CLASS,
             ".") || VRTS::Logger::log('error' , "Failed to trigger thinpool discovery");
    }
    else
    {
        `$mh_ctl  1>/dev/null 2>&1 &`;
    }
    
    my $xprtlc_cmd = VRTS::Paths::get_path("xprtlc");
    if($^O =~ /Win32/i)
    {
        $xprtlc_cmd = "\"".VRTS::Paths::get_path("xprtlc")."\"";
    }
    my $cmd = $xprtlc_cmd." -l https://localhost:5634/admin/cgi-bin/schedule.pl/set/$array_id -d sched=\"60\" -d command=\"$sch_mh_ctl\"";
    VRTS::Logger::log('debug', "Executing $cmd ");
    my $out=`$cmd`;
    my $rc = $?;

    if($? != 0)
    {
        VRTS::Logger::log('debug', "Failed to set schedule: $out");
        return 100;
    }

    # Make a copy of the scheduler entry in array_scheduler.conf
    open (my $AFHAN, ">> $array_sched_file");
    VRTS::Logger::log('debug', "Arrayschedfile [$array_sched_file]");
    my $found = 0;
    while(<$AFHAN>) {
        if($_ =~ /$array_id/) {
            $found = 1;
            last;
        }
    }
    if($found == 0) {
        print $AFHAN "$array_id:360\n";
    }
    close($AFHAN);
}

sub todo_tp_monitor
{
  my $agent = shift;
  my $exec_dir = shift;
  my $vardir = shift;
  my $arrayid = shift;

  my $only_thin_monitor = 0;

  if(($agent->{CONFIGOPTIONS}->{RESCAN} == 1) || ($agent->{CONTEXT}->{optimize} eq 'off')){
     VRTS::Logger::log('debug', "Array rescan. Skipping Thinpool monitoring");
     return $only_thin_monitor;
  }
  my $time_f = File::Spec->catfile($exec_dir, "timestamp.conf");

  my $last_time;
  if (-e $time_f)
  {
	  open (my $AHAN, $time_f);
	  $last_time = <$AHAN>;
	  close($AHAN);
	  VRTS::Logger::log('debug', "Last refreshed time [$last_time]");
  }

  my $lc_time = time;
  my $timeinterval;
  my $sched_f = File::Spec->catfile($vardir, "array_scheduler.conf");
  if( -e $sched_f)
  {
      open (my $HAN, "$sched_f");
      while(<$HAN>){
          if($_ =~ /^$arrayid:(.*)/) {
	      VRTS::Logger::log('debug', "Matched arrayid [$arrayid] in [$sched_f] [$1]");
	      $timeinterval = $1;
	      last;
          }
      }
      $timeinterval = $timeinterval * 60;
      VRTS::Logger::log('debug', "Timeinterval in $sched_f = [$timeinterval] and current time [$lc_time]");
      if($lc_time >= ($last_time + $timeinterval)) {
          VRTS::Logger::log('debug', "Normal full discovery required.");
          $only_thin_monitor = 0;
      } else {
          VRTS::Logger::log('debug', "Thinpool monitoring required.");
          $only_thin_monitor = 1;
      } 
  }
  return $only_thin_monitor;  
}

#######################################################
# This function will ignore the tables
# PARAM1: $agent
# PARAM2: list of tables to be ignored. (optional)
# 
sub dont_report_tables
{
    my $agent =  shift;
    my $ignore_tb_list = shift;

    my @ignore_tables = ();

    # If caller hasn't set the ignore_tb list then
    # get all tables of the agentlet and ignore them.
    # If caller wants particular tables to be ignored then 
    # they need to pass reference of table list as param
    if(not defined $ignore_tb_list) {
        my $table_order = $agent->get_table_report_order();
        foreach my $tb_name (@$table_order) {
           push (@ignore_tables, $tb_name);
        }
    } else {
        @ignore_tables = @$ignore_tb_list;
    }
    $agent->ignoretables(\@ignore_tables);
    $agent->set_context("dontcompute_capacities", 1);
}

sub put_discovery_timestamp
{
    my $agent = shift;
    my $timestampfile = File::Spec->catfile( $agent->{CONTEXT}->{driver_var_dir}, "timestamp.conf");
    open my $AHAN, "> $timestampfile";
    print $AHAN time;
    close($AHAN);

}

sub get_mh_version
{
    my $discovery_host = shift;
    my $installdirmh = VRTS::Paths::get_path("InstallDir");
    my $xcmd = VRTS::Util::make_command(File::Spec->catfile($installdirmh,"bin","xprtlc"),"-l","https://$discovery_host:5634/world/getvitals");
    my $json = `$xcmd`;
    my $hash_ref;
    eval {
        $hash_ref = JSON::from_json($json);
    };
    if($@) { 
        VRTS::Logger::log('debug', "get_mh_version : Error reading getvitals output");
        return undef;
    }
    return $hash_ref->{XPRTLD_VERSION};
} 
    
sub get_ldev_details_by_ldev_name
{
    my $ldev_name = shift;
    my $enc_vdid = shift;

    my $sql = "SELECT ID, SUBSYS_ID FROM HABDBSYNC.P_ARRAY_LDEV WHERE NAME = '$ldev_name' and SUBSYS_ID = '$enc_vdid'";

    my $out = _exec_sql($sql);

    return $out;
}

# This function accepts a hash of device=>snapshot_time
# and converts the snapsht_time into epoch
# It will return an array out of above hash, each element of which
# will have format: device##epoch_snap_time
sub get_epoch_time
{
    my $time_data = shift;
    my $ch = shift;

    my $vardir = VRTS::Paths::get_path('AppDir');
    my $pnumber = $$;
    my $script = $vardir."/tempUtils$pnumber.pl";
    my %local_hash = %$time_data;

    # If discovery host is CMS, don't do push exec but use the Parse module directly.
    my $cms_info = VRTS::Util::get_cms_config_info();

    if(($cms_info->{'cs'} != 1) && ($ch ne $cms_info->{'cs_name'})) {
      open (S, ">$script");
      binmode S;
      print S "\#\!\//opt\/VRTSsfmh\/bin\/perl\n";
      print S "BEGIN { \@INC = (\"/opt/VRTSsfmh/lib/modules\") if ( \$\^O !~ /Win32/i ) }\n\n";
      print S "use Date\:\:Parse\;\n";
      print S "my \%data\;\n";
      foreach my $k (keys %local_hash) {
         print S "\$data{\"$k\"} = \"$local_hash{\"$k\"}\"\;\n";
      }
      print S "foreach my \$key (keys \%data)
            {
              print \"\$key|##|\"\.str2time(\$data{\$key})\.\" \"\;
            }";
      close(S);

      my $out = push_exec($ch, $script, "\$VARDIR\/tempUtils$pnumber.pl");
	  unlink($script);
      if($out->[0] != 0)
      {
         VRTS::Logger::log('debug', "Error in converting to UTC from [$ch]");
         return "";
      } else {         
         return $out->[1];
      }
    } else {
        my $data;
        foreach my $key (keys %local_hash) {
            $data = $data . " $key\|##\|". str2time($local_hash{$key});
        }
      return $data; 
    }
}

#
# Mark the refresh task state as FAILED (state => FINISHED, Result => 1) for the agent
#
sub set_refresh_task_state_failed
{
    my ($agent, $result) = @_;

    my $context = $agent->get_context();
    my $taskid = $context->{taskid};
    if (defined $taskid) {
        my $taskobj = new VRTS::Task;
        if (defined $taskobj) {
            my $ret = $taskobj->load(undef, $taskid);
            if ($ret == 0) {
                $taskobj->update($taskid, {state => &VRTS::Task::TASK_STATE_FINISHED, Result => 1});
            } else {
                VRTS::Logger::log('error', "set_refresh_task_state_failed: cannot load task object for taskid [$taskid]");
                return 1;
            }
        } else {
            VRTS::Logger::log('error', "set_refresh_task_state_failed: cannot create task object for taskid [$taskid]. Task not updated");
            return 1;
        }
    } else {
        return 1;
    }
    return 0;
}

#
# raise a fault against all arrays (except child arrays) discovered previously by the agent
#
sub raise_discovery_fault
{
    my ($agent, $reason) = @_;

    VRTS::Logger::log('error', "raise_discovery_fault: Array discovery failed - [$reason]");

    my $array_tb = $agent->gettable("P_ARRAY_SUBSYSTEM");

    # report fault for all parent arrays (filer_type = null)
    my $rows = $array_tb->getallrowswithvalue_r(&VRTS::Constants::P_ARRAY_SUBSYSTEM::FILER_TYPE, "COLUMNDEFAULTVALUE");

    if (not defined $rows) {
        VRTS::Logger::log('debug', 'raise_discovery_fault: no arrays discovered, no fault is generated');
        return;
    }

    foreach my $r (@{$rows}) {
        my $array_id = $r->[&VRTS::Constants::P_ARRAY_SUBSYSTEM::ID];
        my $array_name = $r->[&VRTS::Constants::P_ARRAY_SUBSYSTEM::NAME];

        VRTS::Summary->generate($agent, 'vrts_ddl_enclosure', $array_id, 'vrts_ddl_enclosure',
            $array_id, $array_name, $array_name, 'ARRAY_NOT_REACHABLE');
    }

    # when we raise discovery fault, ignore all tables except array config, config mapping and faults
    # for delta computation
    my $table_order = $agent->get_table_report_order();
    my @ignore_tbls = grep(!/P_ARRAY_CONFIG_ARRAY_MAPPING$|P_ARRAY_CONFIG$|FAULT_OBJ$|HOST_TO_FAULT$/, @{$table_order});
    dont_report_tables($agent, \@ignore_tbls);

    return;
}

sub get_enclosure_details_by_name 
{
    my $enc_name = shift;

    my $sql = "SELECT ID, SERIAL_NUMBER, VENDOR_ID FROM HABDBSYNC.P_ARRAY_SUBSYSTEM WHERE NAME = '$enc_name'";

    my $out = _exec_sql($sql);

    return $out;
}

#This sub routine creates a perl script, when runs on ch, 
#accepts certificate permanently for them
sub accept_cert_for_array
{
    my ($arg1, $id, $pass, $arg2, $script, $cli_path) = (@_); 
    my $unix_cli = "$cli_path/naviseccli";
    my $win_cli = "$cli_path/NaviSECCli.exe";

    $win_cli =~ s/^\s*//;
    $win_cli =~ s/\s*$//;
    $win_cli =~ s/\\/\\\\/g;

    $unix_cli =~ s/^\s*//;
    $unix_cli =~ s/\s*$//;
    $unix_cli =~ s/\\/\\\\/g;

    open(my $S, ">$script");
    binmode $S;
    print $S  <<EOF;
#!/opt/VRTSsfmh/bin/perl
BEGIN { \@INC = ("/opt/VRTSsfmh/lib/modules") if ( \$^O !~ /Win32/i ) }
require VRTS::Util;
my (\$id, \$pass, \$rc) = ('$id', '$pass');
(\$rc, \$pass) = VRTS::Util::decrypt_encoded_passwd(\$pass, \$id);
if (\$rc) {
    print "decrypt_passwd failed with error code = [\$rc]\n";
    exit \$rc;
}
if ( \$^O !~ /Win32/i ) {
    system("echo 2 | $unix_cli $arg1 -Password \$pass $arg2");
}
else {
    \$pass = VRTS::Util::esc_special_chars(\$pass);
    open(FH , "|-", " \\\"$win_cli\\\" $arg1 -Password \$pass $arg2");
    print FH "2\\n";
    close(FH);
}
EOF
close($S);

}


sub get_enclosure_vdid_by_partner_node_id
{
    my $partner_node_id = shift;
	
    my $sql = "SELECT ID FROM HABDBSYNC.P_ARRAY_SUBSYSTEM WHERE PARTNER_NODE_ID = '$partner_node_id'";
	
    my $out = _exec_sql($sql);
	
    return $out;
}

sub get_ldev_name_by_id
{
   my $ldev_id = shift;

   my $sql = "SELECT NAME FROM HABDBSYNC.P_ARRAY_LDEV WHERE ID = '$ldev_id'";

   my $out = _exec_sql($sql);
	
   return $out;	
}

sub get_remote_devicegroup_details
{
    my $dg_name = shift;
    my $enc_vdid = shift;

    my $sql = "SELECT * FROM HABDBSYNC.P_ARRAY_DEVICE_GROUP WHERE NAME = '$dg_name' and SUBSYS_ID != '$enc_vdid'";

    my $out = _exec_sql($sql);

    return $out;
}

sub get_remote_filesystem_details
{
    my $fs_name = shift;
    my $enc_vdid = shift;

    my $sql = "SELECT * FROM HABDBSYNC.P_ARRAY_NAS_FILESYSTEM WHERE NAME = '$fs_name' and SUBSYS_ID != '$enc_vdid'";

    my $out = _exec_sql($sql);

    return $out;
}

sub read_file {
    my $f_name = shift;
    my $out = undef;
    if (-e $f_name) {
        my $fh = undef;
        open $fh, '<', $f_name;
        $out =  do { local $/; <$fh> };
        close $fh;
    }
    return $out;
}

sub write_file {
    my ($out, $f_name) = @_;
    my $fh = undef;
    open $fh, '>', $f_name;
    print $fh $out;
    close $fh;
}

sub sec_to_hhmmss {
    my $ss = shift;
    my ( $hh, $mm );

    return -1 if ( $ss < 0 );

    $hh = int( $ss / 3600 );
    $ss = $ss % 3600;

    $mm = int( $ss / 60 );
    $ss = $ss % 60;

    my $out_str = sprintf "%0.2u:%0.2u:%0.2u", $hh, $mm, $ss;

    return $out_str;
}

sub get_fqdn {
    my $ip = shift;
    my @dns;
    require Net::DNS;
    my $res = Net::DNS::Resolver->new;

    # create the reverse lookup DNS name (octets in the IP address need to be reversed).
    my $target_IP = join('.', reverse split(/\./, $ip)).'.in-addr.arpa';
    my $query = $res->query($target_IP, 'PTR');
    if ($query)
    {
        foreach my $rr ($query->answer)
        {
            next unless $rr->type eq 'PTR';
            my %tmp = %{$rr};
            push(@dns, $tmp{ptrdname});
        }
    }
    return \@dns;
}

1;
