#$Id$
# VT50-2000-2016-84-28-4
# $Copyright (c) 2017 Veritas Technologies LLC.  All rights reserved. Veritas
# and the Veritas Logo are trademarks or registered trademarks of Veritas
# Technologies LLC or its affiliates in the U.S. and other countries. Other 
# names may be trademarks of their respective owners.
#
# The Licensed Software and Documentation are deemed to be commercial computer
# software as defined in FAR 12.212 and subject to restricted rights as defined
# in FAR Section 52.227-19 "Commercial Computer Software - Restricted Rights" and
# DFARS 227.7202, et seq. "Commercial Computer Software and Commercial Computer
# Software Documentation," as applicable, and any successor regulations, whether
# delivered by Veritas as on premises or hosted services. Any use, modification,
# reproduction release, performance, display or disclosure of the Licensed Software
# and Documentation by the U.S. Government shall be solely in accordance with the
# terms of this Agreement. $

package VRTS::Enclosure::SIPlugin::Parser::KeyValue;
use strict;
use Carp;
require Exporter;
require VRTS::Paths;
require File::Spec;
require JSON;
require VRTS::Enclosure::SIPlugin::Catalog;

my $inst_dir = VRTS::Paths::get_path("InstallDir");

sub new 
{
    my $class = shift;
    my $self = {};

    bless ($self, $class);
    return $self;
}

# instantiates a Parser plugin that can read key value pairs
# e.g. say key1 is the marker field that defines start of an object
#   key1 : value -> obj1 start
#   key2 : value
#   key3 : value
#   key1 : value -> obj2 start
#   key2 : value
# the object definitions are read from the file cliobjects.def
sub initialize
{
    my $top = shift;
    
    my $module = new VRTS::Enclosure::SIPlugin::Parser::KeyValue;
    my $file = File::Spec->catfile($inst_dir,"etc","SIPlugin","cliobjects.def");
    if (open (H, $file) == 0)
    {
        die "Failed to open file $file";
    }
    
    my $buf = do { local $/; <H> };

    my $defs;
    eval { $defs = JSON::from_json($buf)};
    if ($@)
    {
        die "$@";
    }
    $module->{Defs} = $defs;
    return $module;
}

# Creates object hash for the specified object type from the o/p stream given
sub get_object
{
    my $self = shift;
    my $obj = shift;
    my $fh = shift;
    my $marker_key = shift;
    my $marker_val = shift;

    my $obj_hash = ();
    my $mkey = $self->{Defs}->{$obj}->{$marker_key};
    my $match = 0;
    my $id;
    my $data_found = 0;
    my $error_msg = undef;
    my $cnt = 0;

    while(<$fh>)
    {
        chomp;
        if ($data_found == 0)
        {
            $error_msg .= $_;
        }
         
        if (/^[\s]*$mkey[\s]*\:[\s]*(.+)/)
        {
            # Todo Allow for spaces in object Id ?
            $data_found = 1;
            $error_msg = undef;
            if (defined $marker_val && $1 ne $marker_val)
            {
                $match = 0;
                next;
            }
            $id = $cnt;
            $cnt = $cnt + 1;
            $obj_hash->{$id}->{$marker_key} = $1;
            $match = 1;
            #print "id $id\n";
            next;
        }
        next if ($match == 0);
        foreach my $key (keys%{$self->{Defs}->{$obj}})
        {
            my $keyname = $self->{Defs}->{$obj}->{$key};
            my $val;
            if (/^[\s]*$keyname[\s]*\:[\s]*(.+)/)
            {
                $val = $1;
                $val =~ s/[\s]*$//;
                next if ($val eq "");
            }
            else
            {
                next;
            }
            #print "id $id..key $key..val $val\n";
            $obj_hash->{$id}->{$key} = $val;
            next;
        }
    }
    close($fh);
    if ($? || $error_msg =~ /Command not found/ || $error_msg =~ /Failed to execute CLI\:/
       || $error_msg =~ /CLI failed with error/)
    {
        # $? is the return code of xprtlc command and not that of the pushed discovery script.
        # to know if the discovery script failed look for specific error messages.
        VRTS::Logger::log('error', "Parser: Discovery command failed with error: [$error_msg]\n");
        if ($? == 0)
        {
            if ($error_msg =~ /Failed to execute CLI\:/)
            {
                # Discovery CLI is present but could not execute it
                return (undef, &VRTS::Enclosure::SIPlugin::Catalog::FAILED_TO_EXEC_DSC_CMD);
            }
            elsif ($error_msg =~ /CLI failed with error/)
            {
                # Invoked the discovery command but it failed (returned non zero exit code)
                return (undef, &VRTS::Enclosure::SIPlugin::Catalog::DSC_CMD_FAILED);
            } 
            return (undef, &VRTS::Enclosure::SIPlugin::Catalog::DSC_SCRIPT_NOT_FOUND);
        }
        VRTS::Logger::log('error', "Parser: failed to push discovery script on MH: $?");
        return (undef, &VRTS::Enclosure::SIPlugin::Catalog::MH_NOT_REACHABLE);
    }
    return ($obj_hash,0);
}

# Returns definitions for the specified object
sub get_object_meta
{
    my $self = shift;
    my $obj = shift;

    if (defined $obj && exists $self->{Defs}->{$obj})
    {
        return $self->{Defs}->{$obj};
    }
    else
    {
        return $self->{Defs};
    }
}

1;
