#$Id$
# VT50-2000-2016-84-28-4
# $Copyright (c) 2017 Veritas Technologies LLC.  All rights reserved. Veritas
# and the Veritas Logo are trademarks or registered trademarks of Veritas
# Technologies LLC or its affiliates in the U.S. and other countries. Other 
# names may be trademarks of their respective owners.
#
# The Licensed Software and Documentation are deemed to be commercial computer
# software as defined in FAR 12.212 and subject to restricted rights as defined
# in FAR Section 52.227-19 "Commercial Computer Software - Restricted Rights" and
# DFARS 227.7202, et seq. "Commercial Computer Software and Commercial Computer
# Software Documentation," as applicable, and any successor regulations, whether
# delivered by Veritas as on premises or hosted services. Any use, modification,
# reproduction release, performance, display or disclosure of the Licensed Software
# and Documentation by the U.S. Government shall be solely in accordance with the
# terms of this Agreement. $

package VRTS::SISdk::Fabricate;

use strict;
use Exporter;
use VRTS::Paths;
use File::Spec;
use VRTS::Logger;
use JSON;
use Inline::Files;
use VRTS::OS;
use VRTS::Ini;
use VRTS::Util;

our $config_fname = "plugin_def.conf";

# Description: Fabricates a custom agentlet that can discover the array based on specified capabilities
# Input :
#	family : The family of the array 
#       workarea : Location where the fabricated agentlet will be stored.
#       data: JSON hash which represents the array capabilities. 
#             It is JSON hash representation of the array configuration file.
# Output: 
#       Fabricated Agentlet name 
#       dies on failure
sub create_discovery_agentlet
{
    my $family = shift;
    my $workarea = shift;
    my $data = shift;

    die "Missing arguments" if (!defined $family || ! defined $workarea || ! defined $data);

    my $agentlet = "S10discover";
    my $file = File::Spec->catfile($workarea, "$agentlet.pm"); 
    open(F, ">$file") or die "Failed to create agentlet file $file : $!";
    binmode (F);
    print F "package VRTS\:\:$family\:\:$agentlet;\n"; 

    # disover the enclosure object, aggregate capacities, capabilities 
    # and vendor terminologies 
    while (<AGENTLET_TEMPLATE>)
    {
        print F $_;
    }

    while (<DISCOVER_CAPABILITIES>)
    {
        print F $_;
    }

    my @capabilities;
 
    # discover adapters 
    if ($data->{CAPABILITIES}->{ADAPTER})
    {
        push (@capabilities, "discover_adapters(\$agent, \$vdid)");
        while (<DISCOVER_ADAPTERS>)
        {
            print F $_;
        }
    }

    # discover array ports
    if ($data->{CAPABILITIES}->{ARRAY_PORT})
    {
        push (@capabilities, "discover_aps(\$agent, \$vdid)");
        while (<DISCOVER_ARRAY_PORTS>)
        {
            print F $_;
        }
    }

    # discover pdevs
    if ($data->{CAPABILITIES}->{PDEV})
    {
        push (@capabilities, "discover_pdevs(\$agent, \$vdid)");
        while (<DISCOVER_PDEVS>)
        {
            print F $_;
        }
    }

    # discover raid groups
    if ($data->{CAPABILITIES}->{RG})
    {
        push (@capabilities, "discover_arraygrps(\$agent, \$vdid, \"RG\" )");
        while (<DISCOVER_ARRAY_GROUPS>)
        {
            print F $_;
        }
    }

    # discover raid group to pdev mapping
    if ($data->{CAPABILITIES}->{RG_PDEV_MAP})
    {
        push (@capabilities, "discover_rgpdev_map(\$agent, \$vdid)");
        while (<DISCOVER_RGPDEV_MAP>)
        {
            print F $_;
        }
    }

    # discover ldevs
    if ($data->{CAPABILITIES}->{LDEV})
    {
        push (@capabilities, "discover_ldevs(\$agent, \$vdid)");
        while (<DISCOVER_LDEVS>)
        {
            print F $_;
        }
        if ($data->{CAPABILITIES}->{REPLICATION})
        {
            push (@capabilities, "discover_replications(\$agent, \$vdid)");
            while (<DISCOVER_REPLICATIONS>)
            {
                print F $_;
            }
        }
    }

    # discover meta lun assoication with its segments
    if ($data->{CAPABILITIES}->{META_LDEV})
    {
        push (@capabilities, "discover_meta_ldevs(\$agent, \$vdid)");
        while (<DISCOVER_META_LDEVS>)
        {
            print F $_;
        }
    }

    # discover thinpools
    if ($data->{CAPABILITIES}->{TP})
    {
        push (@capabilities, "discover_arraygrps(\$agent, \$vdid, \"TP\")");
        while (<DISCOVER_ARRAY_GROUPS>)
        {
            print F $_;
        }
    }

    # discover raid group to ldev mapping
    if ($data->{CAPABILITIES}->{RG_LDEV_MAP})
    {
        push (@capabilities, "discover_devgrp_ldev_map(\$agent, \$vdid, \"RG\", \"LDEV\")");
        while (<DISCOVER_DEVGRP_LDEV_MAP>)
        {
            print F $_;
        }
    }

    # discover thin pool to ldev mapping
    if ($data->{CAPABILITIES}->{TP_LDEV_MAP})
    {
        push (@capabilities, "discover_devgrp_ldev_map(\$agent, \$vdid, \"TP\", \"LDEV\")");
        while (<DISCOVER_DEVGRP_LDEV_MAP>)
        {
            print F $_;
        }
    }

    # discover thin pool to source ldev mapping
    if ($data->{CAPABILITIES}->{TP_SOURCE_LDEV_MAP})
    {
        push (@capabilities, "discover_devgrp_ldev_map(\$agent, \$vdid, \"TP\", \"SOURCE_LDEV\")");
        while (<DISCOVER_DEVGRP_LDEV_MAP>)
        {
            print F $_;
        }
    }


    # discover ldev pdev mapping.
    # LDEV_PDEV_MAP will not be set if either pdevs or ldevs are
    # not supported
    if ($data->{CAPABILITIES}->{LDEV_PDEV_MAP})
    {
        push (@capabilities, "discover_ldev_pdev_map(\$agent, \$vdid)");
        while (<DISCOVER_LDEV_PDEV_MAP>)
        {
            print F $_;
        }
    }

    # discover ldev host mapping.
    if ($data->{CAPABILITIES}->{HOST_LUN_MAPPING})
    {
        push (@capabilities, "discover_ldev_host_map(\$agent, \$vdid)");
        while (<DISCOVER_LDEV_HOST_MAP>)
        {
            print F $_;
        }
    }

    my $caps_fun = "sub _capabilities_based_discover {\n".
                   "    my \$agent = shift; \n".
                   "    my \$vdid = shift; \n".
                   "    my \$ret = 0 ; \n";
    foreach my $caps (@capabilities)
    {
       $caps_fun .= "    $caps;\n";
    }
    $caps_fun .= "    return \$ret;\n } \n";

    print F $caps_fun;
    print F "1;\n";
    close (F);
    return "$agentlet".".pm";
}

# Description: Fabricates a probe module for the array
# Input :
#       family : The family of the array
#       workarea : Location where the fabricated probe module will be stored.
#       data: JSON hash which represents the array capabilities.
#             It is JSON hash representation of the array configuration file.
# Output:
#       Fabricated probe module name
#       dies on failure
sub create_probe_module
{
    my $family = shift;
    my $workarea = shift;
    my $data = shift;

    my $install_dir = VRTS::Paths::get_path("InstallDir");

    die "Missing arguments" if (!defined $family || ! defined $workarea || ! defined $data);
    my $tmplfile = File::Spec->catfile($install_dir, "etc", "SISdk", "SIPluginProbe.tmpl");
    open (T, "$tmplfile") or die "Failed to open probe template file $tmplfile: $!";
    my $file = File::Spec->catfile($workarea, "$family.pm");
    if (open(F, ">$file") == 0)
    { 
        close (T); 
        die "Failed to create probe module file $file: $!"; 
    }
    binmode (F);
    my $pkg = "package VRTS\:\:Enclosure\:\:Discover\:\:$family;\n";
    while (<T>)
    {
        s/##PKG_MODEL##/$pkg/;
        s/##FAMILY##/$family/;       
        print F $_;
    }
    close (T);
    close (F);

    return "$family".".pm";
}

# Description: Private method to get directory listing of the specified base directory
sub _get_dir_listing
{
    my $base_dir = shift;
    my $dir = shift;
    my $tar_files = shift;

    my $rel_path;

    my $p = File::Spec->catdir ( $base_dir,$dir);
    chdir $p;

    opendir (D, $p);
    my @files = readdir(D);
    closedir (D);
    foreach my $f (@files)
    {
        next if ($f eq "." || $f eq "..");
        $rel_path = File::Spec->catfile ($dir, $f);

        push (@$tar_files, $rel_path);
        my $full_path =  File::Spec->catdir ( $base_dir, $dir, $f );
        if ( -d $full_path)
        {
            _get_dir_listing ($base_dir, $rel_path, $tar_files);
        }
   }
}

# Description: Creates the sfa for the SI Plugin
# Input :
#       basedir : the directory that contains the add-on structure for the SI plugin
#       addon_id : the add-on id to be used for constructing the add-on
# Output:
#       sfa file for the SI Plugin. this will be present in the basedir
#       dies on failure. basedir is not cleaned up on failure.
sub create_plugin_sfa
{
    my $basedir = shift;
    my $addon_id = shift;

    my $addon_dir = File::Spec->catdir ($basedir, $addon_id);

    my $targz = $addon_id.".tar.gz";

    my $rc;
    my $cmd;
    my $file;
    my @tar_files = ();
    my @final_tar;
    push @final_tar, $addon_id;

    chdir $basedir;
    unlink($targz) if (-f $targz);

    chdir ($addon_id);
    unlink($targz) if (-f $targz);


    # get list of files and directories relative to this directory
    opendir (D, $addon_dir);
    my @files = readdir(D);
    closedir(D);

    foreach $file (@files)
    {
        next if ($file eq "." || $file eq "..");
        if ($file eq "catalog.ini" || $file eq "Readme.html")
        {
            push @final_tar, File::Spec->catdir($addon_id, $file);
            next;
        }
        push @tar_files, $file;
        if (-d $file)
        {
            _get_dir_listing($addon_dir, $file, \@tar_files);
            chdir $addon_dir;
        }
    }

    chdir $addon_dir;
    my $tar = Archive::Tar->new;

    $tar->add_files( @tar_files);
    for my $f ( $tar->get_files )
    {
         $file = $f->name();
         if ($file =~ /\.pl$/ || $file =~ /\.sh$/ )
         {
              # Give executable permissions to the perl/sh scripts
              $f->mode( 0744 );
         }
    }

    die "Failed to tar contents of $addon_id" if (! defined $tar->write( $targz, 9));

    # Once the tar is created, cleanup the component files
    foreach $file (@tar_files)
    {
        if (-f $file)
        {
           unlink($file);
        }
        elsif(-d $file)
        {
           File::Path::rmtree($file);
        }
    }

    chdir $basedir;
    push @final_tar, File::Spec->catdir($addon_id, $targz);
    $tar = Archive::Tar->new;
    $tar->add_files( @final_tar);

    die "Failed to create plugin sfa" if (! defined $tar->write( $targz, 9));
    rename ($targz, "$addon_id".".sfa");
    File::Path::rmtree($addon_id);

    # TODO restrict the plugin add-on only to CMS platforms as the add-on is CS only
}

# Description: Extracts the SI Plugin sfa
# Input :
#       addon_sfa : the SI Plugin file name
#       tmp_dir : the location where the sfa should be extracted.
#                 the addon_sfa file should be present in this directory
# Output:
#       The full path of the SI Plugin configuration file extracted from the Plugin sfa.
#       dies on failure. tmp_dir is not removed.
sub extract_plugin
{
    my $addon_sfa = shift;
    my $work_area = shift;
    my $msg;
    my $family;
    my $tmp_dir = $work_area;

    my $upload_file = File::Spec->catfile ($tmp_dir, $addon_sfa);
    die "Cannot access addon sfa file $upload_file" if (! -f $upload_file);

    my $tar_gz = $upload_file;
    my $tar_gz =~ s/\.sfa$/\.tar\.gz/;
    rename($upload_file, $tar_gz);
 
    if (VRTS::OS->ungzip_and_untar($tmp_dir, $upload_file))
    {
        VRTS::Logger::log('error', "Failed to unzip add-on sfa file $upload_file");
        die "Failed to unzip the sfa";
    }

    # Remove the sfa and the tar file contained after extraction
    unlink($upload_file) if ( -f $upload_file);
    $upload_file =~ s/\.gz$//;
    unlink( $upload_file) if ( -f $upload_file);

    my $addon_name = $addon_sfa;
    $addon_name =~ s/\.sfa$//;

    $tmp_dir = File::Spec->catdir ($tmp_dir, $addon_name);
    $upload_file = File::Spec->catfile ($tmp_dir, "$addon_name.tar.gz");
    if (VRTS::OS->ungzip_and_untar($tmp_dir, $upload_file))
    {
        VRTS::Logger::log('error', "Failed to unzip add-on sfa file $upload_file");
        die "Failed to unzip the sfa";
    }

    # Remove the sfa and the tar file contained after extraction
    unlink($upload_file) if ( -f $upload_file);
    $upload_file =~ s/\.gz$//;
    unlink( $upload_file) if ( -f $upload_file);

    my $config_file = File::Spec->catfile($tmp_dir, $config_fname);
    if (!-f $config_file)
    {
        VRTS::Logger::log('error', "Plugin sfa does not contain the config file $config_file");
        die "Plugin sfa does not contain the config file";
    }

    if (open (H, $config_file) == 0)
    {
        VRTS::Logger::log('error', "failed to open file $config_file : $!");
        die  "failed to open config file";
    }

    my $buf = do { local $/; <H> };
    close (H);

    my $data;
    eval { $data = JSON::from_json($buf)};
    if ($@)
    {
        $msg = "Invalid configuration file : $@\n";
        VRTS::Logger::log('error', $msg);
        die $@;
    }

    $family = $data->{ARRAY_PLUGIN_FAMILY};
    if (!defined $family)
    {
        VRTS::Logger::log('error', "Invalid Plugin config file.Plugin Family Missing");
        die "Invalid Plugin config file.Plugin Family Missing";
    }

    my $discovery_cmddir = File::Spec->catfile($tmp_dir, $family);

    my $bck_cmdir = $discovery_cmddir."bck";
    eval { VRTS::Util::copy_recursively($discovery_cmddir, $bck_cmdir);};
    if ($@)
    {
        $msg = "Internal Error";
        VRTS::Logger::log('error', "Failed to back up $discovery_cmddir $@\n");
        die $msg;
    }

    my $file = File::Spec->catfile ($bck_cmdir, "setup-modules.pl");
    unlink ($file) if (-f $file);

    $file =  File::Spec->catfile ($bck_cmdir, "discovery-scripts.tar.gz");
    if ( -f $file)
    {
        if (VRTS::OS->ungzip_and_untar($bck_cmdir, $file))
        {
           VRTS::Logger::log('error', "Failed to unzip discovery scripts tar");
           die "Failed to unzip discovery scripts tar";
        }
        unlink ($file) if (-f $file);
        $file =~ s/\.gz$//;
        unlink ($file) if (-f $file);
    }

    $file = File::Spec->catfile ($bck_cmdir, $config_fname);
    unlink ($file) if (-f $file);

    $data->{DISCOVERY_INFO}->{COMMAND_LOCATION} = $bck_cmdir;

    my $config_data;
    eval {
        $config_data = JSON::to_json($data, {pretty => 1, indent => 1});
    };
    if ($@)
    {
        $msg = "Failed to update the configuration file: $@\n";
        VRTS::Logger::log('error', $msg);
        die $@;
    }
    if (open (H, ">$config_file") == 0)
    {
        VRTS::Logger::log('error', "failed to open file $config_file for update: $!");
        die  "failed to create config file";
    }
    binmode (H);
    print H $config_data;
    close (H);

    return $config_file; 
}

1;

############Template for plugin agentlet#################

__AGENTLET_TEMPLATE__
use strict; 

require Exporter;
require VRTS::Util;
require VRTS::CMD;


require VRTS::Constants::P_ARRAY_LDEV;
require VRTS::Constants::P_ARRAY_PDEV;
require VRTS::Constants::P_ARRAY_SUBSYSTEM;
require VRTS::Constants::P_ARRAY_ADAPTER;
require VRTS::Constants::P_ARRAY_AP;
require VRTS::Constants::P_ARRAY_DEVICEMAP;
require VRTS::Constants::P_ARRAY_RAIDGROUP;
require VRTS::Constants::P_ARRAY_DEVICEGROUP_MAP;
require VRTS::Constants::P_ARRAY_DEVICE_GROUP;
require VRTS::Constants::P_ARRAY_LDEV_MAP;
require VRTS::Constants::P_ARRAY_PDEV_MAP;
require VRTS::Constants::P_ARRAY_UI_STRING;
require VRTS::Constants::P_LUN_MAPPING;
require VRTS::Constants::P_ARRAY_CONFIG_ARRAY_MAPPING;
require VRTS::Constants::P_ARRAY_CONFIG;
require VRTS::Enclosure::SIPlugin::Parser;
require VRTS::Enclosure::SIPlugin::Util;
require VRTS::Enclosure::SIPlugin::Catalog;
require VRTS::Constants::P_ARRAY_CAPACITY;
require VRTS::Constants::P_ARRAY_CAPABILITY;

use VRTS::Agent;
require VRTS::Enclosure::Utils;
require VRTS::Enclosure::Discover;

use vars qw($VERSION @ISA @EXPORT );
$VERSION = '5.00'; 
@ISA = qw(Exporter);
@EXPORT = qw(scan initialize discover callback);

use constant BOOL_VALUE_TRUE => 1;
use constant BOOL_VALUE_FALSE => 0;
my %bool_values = (&BOOL_VALUE_TRUE=> 1, &BOOL_VALUE_FALSE=> 0);

my $pd_media_list = &VRTS::Enclosure::Utils::PD_MEDIA_TYPES;
my %pd_media = map { $_ => 1  } @$pd_media_list;

my $vendor_media_list = &VRTS::Enclosure::Utils::VENDOR_PD_MEDIA_TYPES;
my %pd_vendor_media = map { $_ => 1  } @$vendor_media_list;

my $adapter_type_list = &VRTS::Enclosure::Utils::ADAPTER_TYPES;
my %adapter_type = map { $_ => 1  } @$adapter_type_list;

my $rep_modes_list = &VRTS::Enclosure::Utils::REPLICATION_MODES;
my %rep_mode = map { $_ => 1  } @$rep_modes_list;

my $replication_state_list = &VRTS::Enclosure::Utils::REPLICATION_STATES;
my %replication_state = map {split("=",$_)} @$replication_state_list;
my @rep_states = keys %replication_state;

my $rep_access_mode_list = &VRTS::Enclosure::Utils::REPLICATION_ACCESS_MODES;
my %replication_access_mode = map {split("=",$_)} @$rep_access_mode_list;
my @rep_access_modes = keys %replication_access_mode;

use constant AP_STATUS_UP => 0;
use constant AP_STATUS_DN => 1;
my %port_status = (&AP_STATUS_UP => 1, &AP_STATUS_DN => 0);

use constant ADAPTER_MODE_NORMAL => 0;
use constant ADAPTER_MODE_ERROR => 1;
my %adapter_mode = (&ADAPTER_MODE_NORMAL => &ADAPTER_MODE_NORMAL, 
                    &ADAPTER_MODE_ERROR => &ADAPTER_MODE_ERROR);

use constant PDEV_STATUS_HEALTHY => 0;
use constant PDEV_STATUS_FAILED => 1;

use constant RG_STATUS_HEALTHY => 1;
use constant RG_STATUS_FAILED => 0;
my %rg_states = (&RG_STATUS_HEALTHY => 0, &RG_STATUS_FAILED => 1);

use constant REP_DESIG_SOURCE => 0;
use constant REP_DESIG_TARGET => 1;
my %rep_designations = (&REP_DESIG_SOURCE => 0, &REP_DESIG_TARGET => 1);

use constant REP_LOCAL => 0;
use constant REP_REMOTE => 1;
my %rep_across_encls = (&REP_LOCAL => 0, &REP_REMOTE => 1);

use constant EXEC_CONTEXT_DISCOVER => "discover";
use constant EXEC_CONTEXT_TEST => "sdktest";

use constant MSG_LEVEL_DBG => 0;
use constant MSG_LEVEL_DBG_1 => 1;
use constant MSG_LEVEL_DBG_2 => 2;
use constant MSG_LEVEL_DBG_3 => 3;
use constant MSG_LEVEL_WARN => 4;
use constant MSG_LEVEL_ERROR => 5;
use constant UNDEF_FIELD => "-";

my $num_warns = 0;
my $num_errors = 0;
my $debug_level = &MSG_LEVEL_DBG_2;
my $execution_ctx = &EXEC_CONTEXT_DISCOVER;
my $ch;
my $sch;
my $cli_path;
my ($enc_tb, $ui_tb, $pdev_tb, $ldev_tb, $adap_tb, $ap_tb, $pdevmap_tb, $lunmap_tb, $devgrp_tb, 
    $devgrp_map_tb, $capabilities_tb, $ldevmap_tb);
my $capbs;
my @errors;
my $vardir = VRTS::Paths::get_path("AppDir");
my $install_dir =  VRTS::Paths::get_path("InstallDir");
my $arrayid;
my $vendor;
my $family;
my $method;
my $model;
my $args;
my $config_data;
my $parser;
my $enc_discv;
my $vendor_caps;
my %caps;
my $objects_def;
my $invalid_value_msg;
my $pdevs_discovered;
my $ldevs_discovered;
my $autocompute_caps;
my $discovery_cmdline;
my $discovery_cmd_loc;
my $display_cmd_line;
my $cmd_name;
my $discovery_ctxt = {};
my $test_source;
my $tlog_hndl = undef;
my $applicable_caps;

# Listed device configuration parameters obtained from database will
# not be passed to the plugin discovery CLI
my %skip_cli_params =
(
    "CONTROL_HOST" => 1,
    "SECONDARY_HOST" => 1,
    "INSTALL_DIR" => 1,
    "METHOD" => 1
);

# Initialize method for the agentlet
sub initialize
{
    my $agent = shift;
    $invalid_value_msg = VRTS::Enclosure::SIPlugin::Util::get_error(undef,
           &VRTS::Enclosure::SIPlugin::Catalog::INVALID_FIELD_VALUE);
}

# Description : Method to log messages during the discovery of the SI Plugin array
# If agentlet executes in the SDK test context then they are printed on STDOUT/STDERR as well.
# All error/warning messages are logged by default
# All debug messages are logged if the message debug level specified is lesser than or equal to the
# the current debug level
# Input :
#       msg - Message to be logged
#       level - Message log level
# Ouput : None
sub log_msg
{
    my $msg = shift;
    my $level = shift;
    my $break_line = shift;
    my $br = "";
    my $time = localtime ();

    return if (!defined $msg);
    my $string;

    $break_line = 0 if (!defined $break_line );
    if ($break_line)
    {
        $br = "<br>";
    }
    $level = &MSG_LEVEL_DBG if (!defined $level);
    if ($level == &MSG_LEVEL_ERROR)
    {
        if ($execution_ctx eq &EXEC_CONTEXT_TEST)
        {
            $string = "<div> <div style='color: red;'>$br$msg</div></div>";
            print STDOUT "$string\n" ;
        }
        $num_errors += 1;
        VRTS::Logger::log('error' , $msg);
        print $tlog_hndl "$time $msg\n" if (defined $tlog_hndl);
    }
    elsif ($level == &MSG_LEVEL_WARN)
    {
        if ($execution_ctx eq &EXEC_CONTEXT_TEST)
        {
            $string = "<div> <div style='color: orange;'>$br$msg</div></div>";
            print STDOUT "$string\n";
        }
        $num_warns += 1;
        VRTS::Logger::log('warning' , $msg);
        print $tlog_hndl "$time $msg\n" if (defined $tlog_hndl);
    }
    elsif ($debug_level >= $level)
    {
        if ($execution_ctx eq &EXEC_CONTEXT_TEST)
        {
            if ($test_source == 1 && $msg =~ /^\(cmd:.+\-\-list\s*([^\s]+)/)
            {
                $msg = "Source data file: $1.txt";
            }
            if ($break_line)
            {
                $string = "<div> <div style='color: green;'><br><b>$msg</b></div></div>\n";
            }
            else
            {
                $string = "<div> <div style='color: green;'>$msg</div></div>";
            }
            print STDOUT $string."\n";
        }
        print $tlog_hndl "$time $msg\n" if (defined $tlog_hndl);
        VRTS::Logger::log('debug' , $msg);
    }
}

sub print_record
{
    my $obj_hash = shift;
    log_msg("" , undef, 1);
    my $msg;

    foreach my $fld (keys %$obj_hash)
    {
        if ($execution_ctx eq &EXEC_CONTEXT_TEST)
        {
            print "<div> <div style='color: green;'>".
                  "&nbsp;&nbsp;&nbsp;$obj_hash->{$fld}->{fld_name}:&nbsp;&nbsp;$obj_hash->{$fld}->{value}".
                  "</div></div>";
        }
        $msg = "\t\t$obj_hash->{$fld}->{fld_name}:$obj_hash->{$fld}->{value}";
        print $tlog_hndl $msg."\n" if (defined $tlog_hndl);
        VRTS::Logger::log('debug' ,$msg) if ($debug_level eq &MSG_LEVEL_DBG_2);
    }

    foreach my $fld (keys %$obj_hash)
    {
        if (defined $obj_hash->{$fld}->{dbg_level} && $obj_hash->{$fld}->{dbg_level} ne &MSG_LEVEL_DBG)
        {
            $msg =  $obj_hash->{$fld}->{error_msg};
            $msg = "Invalid value specified" if (!defined $msg or $msg eq "");
            log_msg("$obj_hash->{$fld}->{fld_name}: $msg",
             $obj_hash->{$fld}->{dbg_level},0);
        }
    }
}



# Description : Method to validate a capacity field in the Plugin discovery CLI
#               and return the capacity in bytes
# Input :
#       hash - containing parsed output of the CLI
#       sz_key - the capacity field name (key in the hash)
#       sz_unit - the unit in which the input capacity is specified. 
#                 currently MB is the only unit supported by the method
# Ouput:
#       (retcode, capacity)
#        retcode  - 0 on success, 1 on failure
#        capacity - undef if capacity not discovered , on failure returns the input capacity as is,
#                    on success converts the capacity in bytes and returns it
#                   
sub get_capacity
{
    my $hash = shift;
    my $sz_key = shift;
    my $sz_unit = shift;

    return 0, undef if (!exists $hash->{$sz_key} || $hash->{$sz_key} eq &UNDEF_FIELD);
    my $capacity = $hash->{$sz_key};

    return 1, $capacity if ($capacity !~ /^[0-9]*\.?[0-9]*$/);
    my $capacity = $capacity * 1024 * 1024;
    return 0, $capacity;
}

# Description : Method reads the array configuration parameters 
# (which were specified when the device for this array was configured) 
# if agentlet is running in the SDK test context then the test device configuration
# parameters are read on the array configuration file itself.
# Input :
#       agent - Reference to the Agentlet framework
#       objid - the VDID of the array which is getting discovered
# Ouput:
#       Initializes the global hash config_data with plugin configuration read from the config file.
#       exits on failure.
sub read_config
{
    my $agent = shift;
    my $objid = shift;

    my $error_msg;
    $objid =~ s/\"//g;

    $execution_ctx = $agent->{CONTEXT}->{'execution-ctxt'};
    $method = $agent->{NAME};

    my $test_output = File::Spec->catfile ($vardir, $agent->{CONTEXT}->{'test-log'});
    if (defined $test_output)
    {
        open ($tlog_hndl, ">$test_output");
    }
 
    # read the plugin configuration file
    my $config_file = File::Spec->catdir ($vardir, $method, "plugin_def.conf");
    if (open (H, $config_file) == 0)
    {
        VRTS::Logger::log('error', "Failed to open array module configuration file: $config_file:$!");
        $error_msg = VRTS::Enclosure::SIPlugin::Util::get_error(undef,
           &VRTS::Enclosure::SIPlugin::Catalog::PROBE_INTERNAL_ERROR);
        log_msg ($error_msg, &MSG_LEVEL_ERROR);
        exit(1);
    }

    my $buf = do { local $/; <H> };
    eval { $config_data = JSON::from_json($buf)};
    if ($@)
    {
        VRTS::Logger::log ('error', "Invalid config file $config_file:$@");
        $error_msg = VRTS::Enclosure::SIPlugin::Util::get_error(undef,
           &VRTS::Enclosure::SIPlugin::Catalog::PROBE_INTERNAL_ERROR);
        log_msg ($error_msg, &MSG_LEVEL_ERROR);
        exit(1);
    }

    # Get the device configuration parameters for this device
    my $config = VRTS::Enclosure::SIPlugin::Util::get_config_by_vendor_id($agent, $objid, $config_data);

    $ch = $config->[0]->{CONTROL_HOST};
    $sch = $config->[0]->{SECONDARY_HOST};

    if (!defined $ch)
    {
        VRTS::Logger::log('error',"Failed to fetch configuration parameters from db");
        $error_msg = VRTS::Enclosure::SIPlugin::Util::get_error(undef,
           &VRTS::Enclosure::SIPlugin::Catalog::PROBE_INTERNAL_ERROR);
        log_msg ($error_msg, &MSG_LEVEL_ERROR);
        exit(1);
    }
    $agent->set_context("configid", $config->[0]->{ID});

    $family = $config_data->{ARRAY_PLUGIN_FAMILY};
    $vendor = $config_data->{VENDOR};
    $model = $config_data->{MODEL};

    # Construct arguments to be passed to the plugin discovery CLI
    $discovery_cmdline = $config_data->{DISCOVERY_INFO}->{"ENCL_INFO"};
    my $config_param;
    $cmd_name = $config_data->{DISCOVERY_INFO}->{"DISCOVERY_CMD"};
    my $cmd_file_var;
    my $config_param_value;

    if ($discovery_cmdline =~ /\s+$cmd_name/ || $discovery_cmdline =~ /^$cmd_name/)
    {
        VRTS::Logger::log("debug", "User configurable location for discovery command not specified");
        # if command location is not configurable assume it to be present in the default location
        $discovery_cmd_loc = "default";
    }
    elsif ($discovery_cmdline =~ /\$\{([^\s\}]+)\}[\S]+$cmd_name/)
    {
        $cmd_file_var = $1;
    }

    $display_cmd_line = $discovery_cmdline;
    my $password_str = "\*\*\*";

    while (1)
    {
        $config_param_value = undef;
        if ($discovery_cmdline =~ /\$\{([^\s\}]+)\}/)
        {
            $config_param = $1;
            my $data_type = undef;
            foreach my $field (keys %{$config_data->{DEVICE_CONFIG_META}})
            {
               if ($config_data->{DEVICE_CONFIG_META}->{$field}->{cli_arg_name} eq $config_param)
               {
                   if (defined $config->[0]->{$field} && $config->[0]->{$field} ne "")
                   {
                       $config_param_value = $config->[0]->{$field};
                       $data_type = $config_data->{DEVICE_CONFIG_META}->{$field}->{datatype};
                   }
                   last;
               }
            }

            if (defined $cmd_file_var && $cmd_file_var eq $config_param &&
                 $discovery_cmdline =~ /\$\{([^\s\}]+)\}([\S]+)$cmd_name/)
            {
                $discovery_cmdline =~ s/\$\{([^\s\}]+)\}([\S]+)$cmd_name/$cmd_name/;
                $display_cmd_line =~ s/\$\{([^\s\}]+)\}([\S]+)$cmd_name/$cmd_name/;

                if (defined $config_param_value)
                {
                    $discovery_cmd_loc = $config_param_value;
                }
                else
                {
                    $discovery_cmd_loc = "default";
                }
                next;
            }

            if (defined $config_param_value)

            {
                $discovery_cmdline =~ s/\$\{([^\s\}]+)\}/$config_param_value/;
                if ($data_type eq "PASSWORD_STRING")
                {
                    $display_cmd_line =~ s/\$\{([^\s\}]+)\}/$password_str/;
                }
                else
                {
                    $display_cmd_line =~ s/\$\{([^\s\}]+)\}/$config_param_value/;
                }
            }
            else
            {
                if ($discovery_cmdline  =~ /-?-?$config_param\s+\$\{$config_param\}/)
                {
                    $discovery_cmdline =~ s/-?-?$config_param\s+\$\{$config_param\}//;
                    $display_cmd_line =~ s/-?-?$config_param\s+\$\{$config_param\}//;
                }
                else
                {
                    $discovery_cmdline =~ s/\$\{$config_param\}//;
                    $display_cmd_line =~ s/\$\{$config_param\}//;
                }
            }
        }
        else
        {
            last;
        }
    }

    $discovery_ctxt->{CMD} = $cmd_name;
    $discovery_ctxt->{FAMILY} = $method;

    if ($execution_ctx eq "sdktest")
    {
        $discovery_ctxt->{'execution-ctxt'} = "sdktest";
        $discovery_ctxt->{'test-source'} = $agent->{CONTEXT}->{"test-source"};
        $test_source = $discovery_ctxt->{'test-source'};
        if ($discovery_ctxt->{'test-source'}  == 1)
        {
            $discovery_ctxt->{CMD_LOC} = File::Spec->catfile ($vardir, $agent->{CONTEXT}->{"test-source-dir"});
        }
        else
        {
            $discovery_ctxt->{CMD_LOC} = $discovery_cmd_loc;
        }
    }
    elsif (defined $config_data->{TEST} && $config_data->{TEST} == 1)
    {
        $discovery_ctxt->{'execution-ctxt'} = "sdktest";
        $discovery_ctxt->{'test-source'} = 1;
        $test_source = $discovery_ctxt->{'test-source'};
        $discovery_ctxt->{CMD_LOC} = $config_data->{TEST_SOURCE};
    }
    else
    {
        $discovery_ctxt->{CMD_LOC} = $discovery_cmd_loc;
    }
}

# Description : Agentlet discover interface
# Discovers the array using commands specified in the plugin configuration file.
# Input :
#       agent - Reference to the Agentlet framework
# Ouput:
#       0 on success. non-zero on failure.
#       exits on failure to read device configuration parameters.
#       holds back deletes for the enclosure if it cannot be discovered in the current run.
sub discover 
{
    my $agent = shift;
    my $ret = 0;
    my $error_msg;
    my $cmd;

    my $vdid = $agent->{CONTEXT}->{vardir};
    read_config($agent, $vdid);

    $parser = new VRTS::Enclosure::SIPlugin::Parser();
    $parser->initialize("KeyValue");
    $objects_def = $parser->get_object_meta();


    # Get vendor capabilities
    $vendor_caps = $config_data->{CAPABILITIES};

    my ($msg, $ret) = discover_enclosure($agent, $vdid);
    if ($ret)
    { 
        # skip futher discovery if the enclosure object is not discovered
        $error_msg = VRTS::Enclosure::SIPlugin::Util::get_error($msg, $ret);
        log_msg ($error_msg, &MSG_LEVEL_ERROR);
        goto out;
    }

    # Give opportunity to discovery scripts to fetch all data in one shot if
    # so desired. this is a optional command to implement.
    if ($execution_ctx ne &EXEC_CONTEXT_TEST || $test_source == 0)
    {
        $cmd = $discovery_cmdline." --get-data";
        $discovery_ctxt->{CMD_LINE} = $cmd;

        log_msg ("Calling get-data...", &MSG_LEVEL_DBG,1);
        log_msg ("(cmd: $display_cmd_line --get-data)", &MSG_LEVEL_DBG);
        VRTS::Enclosure::SIPlugin::Util::exec_cmd($ch, $discovery_ctxt);
    }
    # Discover standard information like capabilities, aggregate capacities
    # and vendor terminlogies for the array objects

    VRTS::Logger::log('debug', "Report Capabilities");
    discover_capabilities($agent, $vdid);

    VRTS::Logger::log('debug', "Discover capacities");
    discover_capacities($agent, $vdid);

    VRTS::Logger::log('debug', "Setup vendor specific terminologies");
    discover_ui_strings($agent, $vdid);

    # Discover other array objects based on capabilities specified
    $ret = _capabilities_based_discover($agent, $vdid);

    if ($execution_ctx ne &EXEC_CONTEXT_TEST || $test_source == 0)
    {
        # Give opportunity to discovery scripts to perform any cleanup at the end of all discovery
        $cmd = $discovery_cmdline." --cleanup";
        $discovery_ctxt->{CMD_LINE} = $cmd;
        log_msg ("Calling discovery script cleanup...", &MSG_LEVEL_DBG,1);
        log_msg ("(cmd: $display_cmd_line --cleanup)", &MSG_LEVEL_DBG);
        VRTS::Enclosure::SIPlugin::Util::exec_cmd($ch, $discovery_ctxt);
    }
out:
    my $discovery_script = File::Spec->catfile($vardir,"disc_".$family.$$.".pl");
    if (-f $discovery_script)
    {
        unlink ($discovery_script);
    }
    return ($ret);
}

sub callback
{
    my $agent = shift;
    my $cap_tb = $agent->gettable("P_ARRAY_CAPACITY");
    my $cap_def = $objects_def->{ARRAY_CAPACITIES};

    my @caps = keys %$applicable_caps;
    if (scalar(@caps) == 0)
    {
        goto out;
    }

    log_msg ("Enclosure Storage Capacity Summary", &MSG_LEVEL_DBG,1);
    foreach my $capacity (keys %$cap_def)
    {
        next if (!exists ($applicable_caps->{$capacity}));
        my $index = VRTS::Enclosure::Utils::get_capacity_column_id($capacity);
        next if (!defined $index);

        my $cap_row;
        $cap_row = $cap_tb->getrowwithvalue(&VRTS::Constants::P_ARRAY_CAPACITY::ARRAY_PROP_ID, $index);
        if (defined $cap_row)
        {
            my $value = $cap_row->[&VRTS::Constants::P_ARRAY_CAPACITY::VALUE];
            $value = ($value / (1024 * 1024 * 1024));
            log_msg("$cap_def->{$capacity} : $value GB");
        }
   }

out:
   if ($num_errors || $num_warns)
   {
        log_msg ("$num_warns Warnings, $num_errors Errors found", &MSG_LEVEL_ERROR, 1);
   }
   else
   {
       log_msg ("$num_warns Warnings, $num_errors Errors found", &MSG_LEVEL_DBG, 1);
   }
}

# Description : method discovers the enclosure object by running the Plugin
# discovery CLIs on the specified discovery host.
# if primary discovery host is not reachable then secondary discovery host is used.
# Input :
#       agent - Reference to the Agentlet framework
#       vdid - VDID of the array
# Ouput:
#       msg - Error message on failure to discover enclosure
#       retval - 0 on success. non-zero on failure.
sub discover_enclosure
{
    my $agent = shift;
    my $vdid = shift;
    my $msg;

    my ($enc_info,$ret) = get_enclosure_details($ch, $vdid);
    if ($ret)
    {
        if (defined $sch and $sch ne "")
        {
            VRTS::Logger::log("error", "Could not discover from primary discovery host. trying secondary");
            ($enc_info,$ret) = get_enclosure_details($sch, $vdid);
        }
        if ($ret)
        {
             my @ignore_tbs = ("P_ARRAY_LDEV", "P_ARRAY_PDEV", "P_ARRAY_ADAPTER", "P_ARRAY_UI_STRING",
                              "P_LUN_MAPPING", "P_ARRAY_AP", "P_ARRAY_DEVICEMAP", "P_ARRAY_DEVICEGROUP_MAP",
                              "P_ARRAY_PDEV_MAP", "P_ARRAY_DEVICE_GROUP", "P_ARRAY_LDEV_MAP",
                              "P_ARRAY_CONFIG", "P_ARRAY_CONFIG_ARRAY_MAPPING", "P_ARRAY_CAPACITY", 
                              "P_ARRAY_CAPABILITY");

             $agent->ignoretables(\@ignore_tbs);
             $agent->set_context("dontcompute_capacities", 1);
             return undef, $ret;
        }
    }
    if (!defined $enc_info)
    {   
        return undef, &VRTS::Enclosure::SIPlugin::Catalog::NO_ARRAYS_DISCOVERED;
    }

    $enc_tb = $agent->gettable("P_ARRAY_SUBSYSTEM");

    my @enc_row;
    my $value;
    my $ret;

    my @encl_keys = keys%{$enc_info};
    my $encl_key = $encl_keys[0];

    $enc_row[&VRTS::Constants::P_ARRAY_SUBSYSTEM::ID] = $vdid;

    $enc_info->{$encl_key}->{name} = $vdid if (!exists $enc_info->{$encl_key}->{name});
    $enc_row[&VRTS::Constants::P_ARRAY_SUBSYSTEM::NAME] = $enc_info->{$encl_key}->{name};

    $enc_info->{$encl_key}->{vendor} = $vendor if (!exists $enc_info->{$encl_key}->{vendor});

    $enc_row[&VRTS::Constants::P_ARRAY_SUBSYSTEM::VENDOR_ID] = $vdid;
    $enc_row[&VRTS::Constants::P_ARRAY_SUBSYSTEM::VENDOR] = $enc_info->{$encl_key}->{vendor};

    # TODO Bail out (critical error) if product is not specified
    $enc_row[&VRTS::Constants::P_ARRAY_SUBSYSTEM::PRODUCT] = $enc_info->{$encl_key}->{product};

    $enc_row[&VRTS::Constants::P_ARRAY_SUBSYSTEM::FAMILY] = $enc_info->{$encl_key}->{vendor};

    $enc_info->{$encl_key}->{serial_num} = $vdid if (!exists $enc_info->{$encl_key}->{serial_num});
    $enc_row[&VRTS::Constants::P_ARRAY_SUBSYSTEM::SERIAL_NUMBER] = 
         $enc_info->{$encl_key}->{serial_num};

    $enc_row[&VRTS::Constants::P_ARRAY_SUBSYSTEM::MICROCODE_VERSION] = 
         $enc_info->{$encl_key}->{frmware_ver} if (exists $enc_info->{$encl_key}->{frmware_ver});

    $enc_row[&VRTS::Constants::P_ARRAY_SUBSYSTEM::IP] =
         $enc_info->{$encl_key}->{ip_address} if (exists $enc_info->{$encl_key}->{ip_address});

    $enc_row[&VRTS::Constants::P_ARRAY_SUBSYSTEM::ALT_ID] =
         $enc_info->{$encl_key}->{alt_id} if (exists $enc_info->{$encl_key}->{alt_id});

    if (exists $enc_info->{$encl_key}->{autocompute_caps} && $enc_info->{$encl_key}->{autocompute_caps} == 1)
    {
        VRTS::Logger::log('debug', "Agentlet wants to auto compute capacities");
        $autocompute_caps = 1;
    }

    $agent->set_context("enclosure_vdid", $vdid);
    $discovery_ctxt->{encl_vdid} = $vdid;

    $agent->set_context("enclosure_vdid", $vdid);
    $agent->set_context("arrayid", $vdid);
    
    $enc_tb->addrow(\@enc_row);
    $enc_tb->print_table() if ($debug_level == &MSG_LEVEL_DBG_3);
    return undef, 0;
}

# Fires the actual plugin specified CLI for getting enclosure information.
# The o/p of the CLI is parsed using the Parser
sub get_enclosure_details
{
    my $ch = shift;
    my $vdid = shift;
    my $msg;

    my $cmd_line = $discovery_cmdline." --list encls";
    $discovery_ctxt->{CMD_LINE} = $cmd_line;

    log_msg ("Discovering Enclosure details ...", &MSG_LEVEL_DBG,1);
    log_msg ("(cmd: $display_cmd_line --list encls )", &MSG_LEVEL_DBG);

    my $out = VRTS::Enclosure::SIPlugin::Util::exec_cmd($ch, $discovery_ctxt);
    if ($out->[0] != 0)
    {
        $msg = "Failed to push/exec CLI: $out->[1]. $out->[0]";
        $msg = VRTS::Enclosure::SIPlugin::Util::get_error($msg, $out->[0]);
        log_msg ($msg, &MSG_LEVEL_ERROR);
        return undef, $out->[0];
    }
    my $fh = $out->[1];
    my ($enc_hash, $ret) = $parser->get_object("ARRAY_SUBSYSTEM",$fh, "id" , $vdid);
    return ($enc_hash,$ret);
}

# Description : method discovers aggregate capacities by running the Plugin
# discovery CLIs on the specified discovery host.
# Input :
#       agent - Reference to the Agentlet framework
#       vdid - VDID of the array
# Ouput:
#       msg - Error message on failure to discover
#       retval - 0 on success. non-zero on failure.
sub discover_capacities
{
    my $agent = shift;
    my $vdid = shift;
    my $msg;
    my ($ret, $value);

    if ($autocompute_caps)
    {
        $agent->set_context("compute_capacities", 1);
        return undef, 0;
    }

    my $cmd = $discovery_cmdline." --list capacities --encl $vdid";
    $discovery_ctxt->{CMD_LINE} = $cmd;

    log_msg ("Discovering Capacities ...", &MSG_LEVEL_DBG,1);
    log_msg ("(cmd: $display_cmd_line --list capacities --encl $vdid)", &MSG_LEVEL_DBG);

    my $out = VRTS::Enclosure::SIPlugin::Util::exec_cmd($ch, $discovery_ctxt);
    if ($out->[0] != 0)
    {
        $msg = "Failed to push/exec CLI: $out->[1]. $out->[0]";
        $msg = VRTS::Enclosure::SIPlugin::Util::get_error($msg, $out->[0]);
        log_msg ($msg, &MSG_LEVEL_ERROR);
        return undef, $out->[0];
    }
    my $fh = $out->[1];
    my ($obj_hash, $ret) = $parser->get_object("ARRAY_CAPACITIES", $fh, "id");
    if ($ret)
    {
        $msg = "Vendor CLI failed to execute:$obj_hash:retcode $ret";
        $msg = VRTS::Enclosure::SIPlugin::Util::get_error($msg, $ret);
        log_msg ($msg, &MSG_LEVEL_ERROR);
        return (undef, $ret);
    }
    if (!defined $obj_hash )
    {
         $msg = VRTS::Enclosure::SIPlugin::Util::get_error($msg,
                &VRTS::Enclosure::SIPlugin::Catalog::NO_OBJECTS_DISCOVERED);
         log_msg ($msg, &MSG_LEVEL_ERROR);
         return undef, &VRTS::Enclosure::SIPlugin::Catalog::NO_OBJECTS_DISCOVERED;
    }

    my $capacities = $parser->get_object_meta("ARRAY_CAPACITIES");
    my $cap_tb = $agent->gettable("P_ARRAY_CAPACITY");
    my $invalid_physical_sz = 0;

    foreach my $capacity (keys %$capacities)
    {
        next if ($capacity eq "id");
        ($ret, $value) = get_capacity($obj_hash->{$vdid}, $capacity, "MB");
        if ($ret) {
            log_msg($invalid_value_msg."> $capacities->{$capacity}: $value", &MSG_LEVEL_ERROR);
            $value = 0;
            $invalid_physical_sz = 1 if ($capacity eq "physical_size");
        }

        $value = 0 if (!defined $value);
        $caps{$capacity} = $value;
    }

    if ($caps{physical_size} > 0)
    {
        my $row = $enc_tb->getrowwithvalue(&VRTS::Constants::P_ARRAY_SUBSYSTEM::ID, $vdid);
        $row->[&VRTS::Constants::P_ARRAY_SUBSYSTEM::PHYSICAL_SIZE] = $caps{physical_size};
        $enc_tb->addrow($row);
    }

    if ($caps{unallocated_size} > 0 || $caps{allocated_size} > 0 ||
        $caps{spare_size} > 0)
    {
        # If atleast total physical size and either of the other sizes
        # (physical configured, unconfigured or spare) is specified then
        # validate if total physical is the sum of all others.
        if ($caps{physical_size} == 0)
        {
            log_msg($invalid_value_msg."> $capacities->{physical_size}: $value", &MSG_LEVEL_ERROR)
                if (!$invalid_physical_sz);
            $caps{allocated_size} = $caps{unallocated_size} = $caps{spare_size} = 0;
        }
        if ($caps{physical_size} != $caps{allocated_size} + $caps{unallocated_size} + $caps{spare_size})
        {
            log_msg($invalid_value_msg."> $capacities->{physical_size}/$capacities->{allocated_size}".
            "/$capacities->{unallocated_size}/$capacities->{spare_size}".
            ":$capacities->{physical_size} should be $capacities->{allocated_size} + ".
            "$capacities->{unallocated_size} + $capacities->{spare_size}", &MSG_LEVEL_WARN);
        }
    }
    elsif ($caps{physical_size} > 0)
    {
        # if only total physical size is specified then assume all of it to be
        # Unconfigured
        $caps{unallocated_size} = $caps{physical_size};
    }

    if ($caps{logical_assigned_size} == 0 && $caps{logical_unassigned_size} == 0
        && $caps{logical_admin_size} == 0 && $caps{logical_reptarget_size} == 0)
    {
        # if only total logical size is specified then assume all of it to be
        # unassigned
        if ($caps{logical_total_size} > 0)
        {
            $caps{logical_unassigned_size} = $caps{logical_total_size}
        }
    }

    foreach my $capacity (keys %$capacities)
    {
        next if ($capacity eq "id");
        my @cap_row;
        my $index = VRTS::Enclosure::Utils::get_capacity_column_id($capacity);
        next if (!defined $index);
        $cap_row[&VRTS::Constants::P_ARRAY_CAPACITY::ARRAY_ID] = $vdid;
        $cap_row[&VRTS::Constants::P_ARRAY_CAPACITY::ARRAY_PROP_ID] = $index;
        $cap_row[&VRTS::Constants::P_ARRAY_CAPACITY::VALUE] = $caps{$capacity};
        $cap_tb->addrow(\@cap_row);
    }
    $cap_tb->print_table() if ($debug_level == &MSG_LEVEL_DBG_3);
    return (undef, 0);
}

__DISCOVER_CAPABILITIES__
# discovers the array capabiliites
sub discover_capabilities
{
    my $agent = shift;
    my $vdid = shift;

    $capbs = VRTS::Enclosure::Utils::get_capabilities_hash();

    my %vendor_caps;
    foreach my $cap (keys %$capbs)
    {
        $vendor_caps{$cap} = 0;
    }

    if ($config_data->{CAPABILITIES}->{ADAPTER})
    {
        $adap_tb = $agent->gettable("P_ARRAY_ADAPTER");
    }

    if ($config_data->{CAPABILITIES}->{ARRAY_PORT})
    {
        $ap_tb = $agent->gettable("P_ARRAY_AP");
    }

    if ($config_data->{CAPABILITIES}->{PDEV})
    {
        $pdev_tb = $agent->gettable("P_ARRAY_PDEV");
        $pdevmap_tb = $agent->gettable("P_ARRAY_PDEV_MAP");

        $vendor_caps{SHOW_PDEV} = 1;
        $applicable_caps->{physical_size} = 1;
        $applicable_caps->{allocated_size} = 1;
        $applicable_caps->{spare_size} = 1;
        $applicable_caps->{unallocated_size} = 1;
    }

    if ($config_data->{CAPABILITIES}->{LDEV})
    {
        $ldev_tb = $agent->gettable("P_ARRAY_LDEV");
        $lunmap_tb = $agent->gettable("P_LUN_MAPPING");
        $ldevmap_tb = $agent->gettable("P_ARRAY_LDEV_MAP");

        $vendor_caps{SHOW_LDEV} = 1;
        $applicable_caps->{logical_total_size} = 1;
        $applicable_caps->{logical_assigned_size} = 1;
        $applicable_caps->{logical_unassigned_size} = 1;
        $applicable_caps->{logical_admin_size} = 1;
        if ($config_data->{CAPABILITIES}->{REPLICATION})
        {
             $vendor_caps{SHOW_REPLICATION} = 1;
             $applicable_caps->{logical_reptarget_size} = 1;
             if ($config_data->{CAPABILITIES}->{PRIMARY_REMOTE_REP})
             {  
                 # Array supports remote replication
                 $vendor_caps{PRIMARY_REMOTE_REP_CAPABLE} = 1;
                 $vendor_caps{SECONDARY_REMOTE_REP_CAPABLE} = 1;
             }
        }
        if ($config_data->{CAPABILITIES}->{META_LDEV})
        {
            $vendor_caps{SHOW_VSET} = 1;
        }
    }

    if ($vendor_caps{SHOW_PDEV} && $vendor_caps{SHOW_LDEV})
    {
        if ($config_data->{CAPABILITIES}->{LDEV_PDEV_MAP})
        {
            $vendor_caps{SHOW_LDEV_PDEV} = 1;
            $vendor_caps{SHOW_PDEV_FOR_THIN_LDEV} = 1;
        }
    }

    $devgrp_tb = $agent->gettable("P_ARRAY_DEVICE_GROUP");
    $devgrp_map_tb = $agent->gettable("P_ARRAY_DEVICEGROUP_MAP");

    if ($config_data->{CAPABILITIES}->{RG})
    {
        $applicable_caps->{rg_total_size} = 1;
        $applicable_caps->{rg_addressable_size} = 1;
        $applicable_caps->{rg_overhead_size} = 1;
        $applicable_caps->{rg_free_size} = 1;

        $vendor_caps{SHOW_RG} = 1;
        if ($vendor_caps{SHOW_PDEV} && $config_data->{CAPABILITIES}->{RG_PDEV_MAP})
        {
            $vendor_caps{SHOW_RG_PDEV} = 1;
        }
        if ($vendor_caps{SHOW_LDEV} && $config_data->{CAPABILITIES}->{RG_LDEV_MAP})
        {
            $vendor_caps{SHOW_RG_LDEV} = 1;
        }
    }
    if ($config_data->{CAPABILITIES}->{TP})
    {
        $vendor_caps{SHOW_TP} = 1;
        if ($vendor_caps{SHOW_LDEV})
        {
            if ($config_data->{CAPABILITIES}->{TP_LDEV_MAP})
            {
                $vendor_caps{SHOW_TP_THINLDEV} = 1;
            }
            if ($config_data->{CAPABILITIES}->{TP_SOURCE_LDEV_MAP}) 
            {
                $vendor_caps{SHOW_TP_SOURCELDEV} = 1;
            }
        } 
    }
 
    $capabilities_tb = $agent->gettable("P_ARRAY_CAPABILITY");

    foreach my $cap (keys %$capbs)
    {
        my @cap_row;
        $cap_row[&VRTS::Constants::P_ARRAY_CAPABILITY::ARRAY_ID] = $vdid;
        $cap_row[&VRTS::Constants::P_ARRAY_CAPABILITY::ARRAY_PROP_ID] = $capbs->{$cap}->{index};
        $cap_row[&VRTS::Constants::P_ARRAY_CAPABILITY::VALUE] = $vendor_caps{$cap};
        $capabilities_tb->addrow(\@cap_row);
    }

    $agent->set_context("enforce_capability", \%vendor_caps);

    $capabilities_tb->print_table() if ($debug_level == &MSG_LEVEL_DBG_3);
}

# Discovers vendor terminology for array objects
sub discover_ui_strings
{
    my $agent = shift;
    my $enc_vdid = shift;

    $ui_tb = $agent->gettable("P_ARRAY_UI_STRING");

    foreach my $key (keys %{$config_data->{OBJECT_NAMES}})
    {
        my $value = $config_data->{OBJECT_NAMES}->{$key};
        my @ui_row;

        $ui_row[&VRTS::Constants::P_ARRAY_UI_STRING::ID] = $enc_vdid;
        $key = "LUSE" if ($key eq "META_LDEV");
        $ui_row[&VRTS::Constants::P_ARRAY_UI_STRING::PROP_KEY] = $key;
        $ui_row[&VRTS::Constants::P_ARRAY_UI_STRING::PROP_VALUE] = $value;

        $ui_tb->addrow(\@ui_row);
    }

    $ui_tb->print_table() if ($debug_level == &MSG_LEVEL_DBG_3);
}

__DISCOVER_PDEVS__
# Description : Discovers physical devices using vendor specified CLI.
# The o/p is parsed and stored in memory VOM PDEV object
# Input :
#       agent - Reference to the Agentlet framework
#       vdid - VDID of the array
# Ouput:
#       msg - Error message on failure to discover 
#       retval - 0 on success. non-zero on failure.
sub discover_pdevs
{
    my $agent = shift;
    my $vdid = shift;
    my $ret;
    my $value;
    my $msg;
    my $num_objects = 0;
    my $dbg_level;

    my $num_spares = 0;

    my $cmd = $discovery_cmdline." --list pdevs --encl $vdid";
    $discovery_ctxt->{CMD_LINE} = $cmd;

    log_msg ("Discovering Physical disks..",&MSG_LEVEL_DBG, 1);
    log_msg ("(cmd: $display_cmd_line --list pdevs --encl $vdid)", &MSG_LEVEL_DBG);

    my $out = VRTS::Enclosure::SIPlugin::Util::exec_cmd($ch, $discovery_ctxt);
    if ($out->[0] != 0)
    {
        $msg = "Failed to push/exec CLI: $out->[1]. $out->[0]";
        $msg = VRTS::Enclosure::SIPlugin::Util::get_error($msg, $out->[0]);
        log_msg ($msg, &MSG_LEVEL_ERROR);
        return undef, $out->[0];
    }

    my $fh = $out->[1];
    my ($obj_hash, $ret) = $parser->get_object("ARRAY_PDEV", $fh, "id");
    if ($ret)
    {
        $msg = "Vendor CLI failed to execute:$obj_hash:retcode $ret";
        $msg = VRTS::Enclosure::SIPlugin::Util::get_error($msg, $ret);
        log_msg ($msg, &MSG_LEVEL_ERROR);
        return (undef, $ret);
    }
    if (!defined $obj_hash )
    {
         $msg = VRTS::Enclosure::SIPlugin::Util::get_error($msg, 
                &VRTS::Enclosure::SIPlugin::Catalog::NO_OBJECTS_DISCOVERED);
         log_msg ($msg);
         return undef, &VRTS::Enclosure::SIPlugin::Catalog::NO_OBJECTS_DISCOVERED;
    }
    my $pdev_def = $objects_def->{ARRAY_PDEV};
    foreach my $pdev (keys%$obj_hash)
    {
        my $bad_record = ();
        my @phy_row;
        $phy_row[&VRTS::Constants::P_ARRAY_PDEV::ID] = $obj_hash->{$pdev}->{id};
        if (exists $obj_hash->{$pdev}->{name} && $obj_hash->{$pdev}->{name} ne "")
        {
            $phy_row[&VRTS::Constants::P_ARRAY_PDEV::NAME] = $obj_hash->{$pdev}->{name};
        }
        else
        {
            $phy_row[&VRTS::Constants::P_ARRAY_PDEV::NAME] = $phy_row[&VRTS::Constants::P_ARRAY_PDEV::ID];
        }

        # total size of the physical device
        ($ret, $value) = get_capacity($obj_hash->{$pdev}, "physical_sz", "MB");
        if ($ret || !defined $value) {
            $dbg_level = &MSG_LEVEL_ERROR;
            if ($ret)
            {
                $msg =  "Invalid value specified";
            }
            else
            {
                $msg =  "Physical disk size not specified";
            }
            $bad_record->{physical_sz} = {
                                        fld_name => $pdev_def->{physical_sz},
                                        value => $value,
                                        dbg_level => $dbg_level,
                                        error_msg => $msg
                                    };
        }
        else
        {
            $phy_row[&VRTS::Constants::P_ARRAY_PDEV::SIZE] = $value;
        } 
        $phy_row[&VRTS::Constants::P_ARRAY_PDEV::STATUS] = &PDEV_STATUS_HEALTHY;
        $phy_row[&VRTS::Constants::P_ARRAY_PDEV::STATUS] = $obj_hash->{$pdev}->{state}
        if (exists $obj_hash->{$pdev}->{state} && $obj_hash->{$pdev}->{state} == &PDEV_STATUS_FAILED);

        $phy_row[&VRTS::Constants::P_ARRAY_PDEV::SUBSYS_ID] = $vdid;

        $phy_row[&VRTS::Constants::P_ARRAY_PDEV::MEDIA_TYPE] = "Unknown";
        if (defined $obj_hash->{$pdev}->{media_type} && 
            $obj_hash->{$pdev}->{media_type} ne &UNDEF_FIELD)
        {
            if (!exists $pd_media{$obj_hash->{$pdev}->{media_type}})
            {
                 $msg = "Invalid value specified. Allowed values are: @$pd_media_list";
                 $bad_record->{media_type} = {
                                        fld_name => $pdev_def->{media_type},
                                        value => $obj_hash->{$pdev}->{media_type},
                                        dbg_level => &MSG_LEVEL_ERROR,
                                        error_msg => $msg
                                    };
            }
            else
            {
                $phy_row[&VRTS::Constants::P_ARRAY_PDEV::MEDIA_TYPE] = $obj_hash->{$pdev}->{media_type};
            }
        }

         if (defined $obj_hash->{$pdev}->{vendor_media_type} &&
             $obj_hash->{$pdev}->{vendor_media_type} ne &UNDEF_FIELD)
         {
             if (!exists $pd_vendor_media{$obj_hash->{$pdev}->{vendor_media_type}})
             {
                  $msg = "Invalid value specified. Allowed values are: @$vendor_media_list";
                  $bad_record->{vendor_media_type} = {
                                         fld_name => $pdev_def->{vendor_media_type},
                                         value => $obj_hash->{$pdev}->{vendor_media_type},
                                         dbg_level => &MSG_LEVEL_ERROR,
                                         error_msg => $msg
                                     };
            }
             else
             {
                 $phy_row[&VRTS::Constants::P_ARRAY_PDEV::VENDOR_MEDIA_TYPE] = $obj_hash->{$pdev}->{vendor_media_type};
             }
        }

 
        # available_sz is the size of the physical disk (out of the total size) that is not configured
        ($ret, $value) = get_capacity($obj_hash->{$pdev}, "available_sz", "MB");
        if ($ret || !defined $value) {
            if ($ret)
            {
                $dbg_level = &MSG_LEVEL_ERROR;
                $msg =  "Invalid value specified";
            }
            else
            {
                $dbg_level = &MSG_LEVEL_WARN;
                $msg =  "Physical disk available size not specified";
            }
            $bad_record->{available_sz} = {
                                        fld_name => $pdev_def->{available_sz},
                                        value => $value,
                                        dbg_level => $dbg_level,
                                        error_msg => $msg
                                    };
        }
        else
        {
            $phy_row[&VRTS::Constants::P_ARRAY_PDEV::AVAILABLE] = $value;

            # validate that available is less than total size of the disk
            if (defined $phy_row[&VRTS::Constants::P_ARRAY_PDEV::SIZE] &&
                $phy_row[&VRTS::Constants::P_ARRAY_PDEV::SIZE] ne "COLUMNDEFAULTVALUE" &&
                $value > $phy_row[&VRTS::Constants::P_ARRAY_PDEV::SIZE])
            {
                $msg = "Physical disk available size should not be greater than total disk size";
                $bad_record->{available_sz} = {
                                        fld_name => $pdev_def->{available_sz},
                                        value => $value,
                                        dbg_level => &MSG_LEVEL_ERROR,
                                        error_msg => $msg
                                    };
            }
        }

        # spare_sz size of the device that is designated as spare. could be a part of the disk or
        # the entire disk
        ($ret, $value) = get_capacity($obj_hash->{$pdev}, "spare_sz", "MB");
        if ($ret) {
            $bad_record->{spare_sz} = {
                                        fld_name => $pdev_def->{spare_sz},
                                        value => $value,
                                        dbg_level => &MSG_LEVEL_ERROR
                                    };
        }
        elsif (defined $value)
        {
            # GUI shows a disk with non zero spare size as a spare disk
            $phy_row[&VRTS::Constants::P_ARRAY_PDEV::SPARE_SIZE] = $value;

            # validate that spare size is less than total size of the disk
            if (defined $phy_row[&VRTS::Constants::P_ARRAY_PDEV::SIZE] &&
                $phy_row[&VRTS::Constants::P_ARRAY_PDEV::SIZE] ne "COLUMNDEFAULTVALUE" &&
                $value > $phy_row[&VRTS::Constants::P_ARRAY_PDEV::SIZE])
            {
                $msg = "Physical disk spare size should not be greater than total disk size";
                $bad_record->{spare_sz} = {
                                        fld_name => $pdev_def->{spare_sz},
                                        value => $value,
                                        dbg_level => &MSG_LEVEL_ERROR,
                                        error_msg => $msg
                                    };
            }
            $num_spares++ if ($value > 0);
        }

        # shared should be set to 1 if pdev is shared across array groups
        $phy_row[&VRTS::Constants::P_ARRAY_PDEV::IS_SHARED] = 0;
        if (exists $obj_hash->{$pdev}->{shared} && $obj_hash->{$pdev}->{shared} == 1)
        {
            $phy_row[&VRTS::Constants::P_ARRAY_PDEV::IS_SHARED] = 1;
        }

        # RPM of the physical device.
        if (defined $obj_hash->{$pdev}->{rpm} && $obj_hash->{$pdev}->{rpm} ne &UNDEF_FIELD)
        {
            if ($obj_hash->{$pdev}->{rpm} =~ /\d+/)
            {
                $phy_row[&VRTS::Constants::P_ARRAY_PDEV::RPM] = $obj_hash->{$pdev}->{rpm};
            }
            else
            {
                $msg = "RPM should be specified as an integer";
                $bad_record->{rpm} = {
                                        fld_name => $pdev_def->{rpm},
                                        value => $obj_hash->{$pdev}->{rpm},
                                        dbg_level => &MSG_LEVEL_ERROR,
                                        error_msg => $msg
                                    };
            }
        }
        $pdev_tb->addrow(\@phy_row);
        $num_objects ++;
        $pdevs_discovered = 1;

        my @errors = keys %$bad_record;
        if (scalar(@errors))
        {
            $bad_record->{id} = {
                                    fld_name => $pdev_def->{id},
                                    value => $phy_row[&VRTS::Constants::P_ARRAY_PDEV::ID],
                                    dbg_level  => &MSG_LEVEL_DBG
                                };
            print_record ($bad_record);
        }
    }

    my $row = $enc_tb->getrowwithvalue(&VRTS::Constants::P_ARRAY_SUBSYSTEM::ID,
                      $vdid);
    if (defined $row)
    {
        $row->[&VRTS::Constants::P_ARRAY_SUBSYSTEM::NOOFSPAREDRIVES] = $num_spares;
        $enc_tb->addrow($row);
    }

    $pdev_tb->print_table() if ($debug_level == &MSG_LEVEL_DBG_3);
    log_msg("Total Records discovered: $num_objects");
    return (undef, 0);
}

__DISCOVER_LDEVS__
# Description : method discovers logical devices by executing Plugin
# discovery CLIs on the specified discovery host.
# Input :
#       agent - Reference to the Agentlet framework
#       vdid - VDID of the array
# Ouput:
#       msg - Error message on failure to discover
#       retval - 0 on success. non-zero on failure.
sub discover_ldevs
{
    my $agent = shift;
    my $vdid = shift;
    my $ret;
    my $value;
    my $msg;
    my $num_objects = 0;
    my $dbg_level;

    my $cmd = $discovery_cmdline." --list ldevs --encl $vdid";
    log_msg ("Discovering Logical disks...", &MSG_LEVEL_DBG, 1);
    log_msg ("(cmd: $display_cmd_line --list ldevs --encl $vdid)", &MSG_LEVEL_DBG);

    $discovery_ctxt->{CMD_LINE} = $cmd;
    my $out = VRTS::Enclosure::SIPlugin::Util::exec_cmd($ch, $discovery_ctxt);

    if ($out->[0] != 0)
    {
        $msg = "Failed to push/exec CLI: $out->[1]. $out->[0]";
        $msg = VRTS::Enclosure::SIPlugin::Util::get_error($msg, $out->[0]);
        log_msg ($msg, &MSG_LEVEL_ERROR);
        return undef, $out->[0];
    }
    my $fh = $out->[1];
    my ($obj_hash, $ret) = $parser->get_object("ARRAY_LDEV", $fh, "id");
    if ($ret)
    {
        $msg = "Vendor CLI failed to execute:$obj_hash:retcode $ret";
        $msg = VRTS::Enclosure::SIPlugin::Util::get_error($msg, $ret);
        log_msg ($msg, &MSG_LEVEL_ERROR);
        return (undef, $ret);
    }
    if (!defined $obj_hash )
    {
         $msg = VRTS::Enclosure::SIPlugin::Util::get_error($msg,
                &VRTS::Enclosure::SIPlugin::Catalog::NO_OBJECTS_DISCOVERED);
         log_msg ($msg);
         return undef, &VRTS::Enclosure::SIPlugin::Catalog::NO_OBJECTS_DISCOVERED;
    }

    my $no_ldevs;
    my $def = $objects_def->{ARRAY_LDEV};

    foreach my $ldev (keys%$obj_hash)
    {
        my $bad_record = {};
        my @ldev_row;

        $ldev_row[&VRTS::Constants::P_ARRAY_LDEV::ID] = $obj_hash->{$ldev}->{id};
        $ldev_row[&VRTS::Constants::P_ARRAY_LDEV::ALT_ID] = $obj_hash->{$ldev}->{alt_id};
        $ldev_row[&VRTS::Constants::P_ARRAY_LDEV::SUBSYS_ID] = $vdid;
        $ldev_row[&VRTS::Constants::P_ARRAY_LDEV::TYPE] = "LDEV";
        $ldev_row[&VRTS::Constants::P_ARRAY_LDEV::NAME] = $obj_hash->{$ldev}->{name};

        $ldev_row[&VRTS::Constants::P_ARRAY_LDEV::RAID_LEVEL] = $obj_hash->{$ldev}->{raid_level};
        $ldev_row[&VRTS::Constants::P_ARRAY_LDEV::LAYOUT] = $obj_hash->{$ldev}->{layout};

        # logical_sz is the logical size of the logical device.
        # for a thin device this is the subscribed size of the device.
        ($ret, $value) = get_capacity($obj_hash->{$ldev}, "logical_sz", "MB");
        if ($ret || !defined $value)
        {
            if (! defined $value)
            {
                $msg = "Logical disk size not specified";
            }
            else
            {
                $msg = "Invalid value specified";
            }
            $bad_record->{logical_sz} = {
                                    fld_name => $def->{logical_sz},
                                    value => $obj_hash->{$ldev}->{logical_sz},
                                    dbg_level => &MSG_LEVEL_ERROR,
                                    error_msg => $msg
                                 };
        }
        else
        {
            $ldev_row[&VRTS::Constants::P_ARRAY_LDEV::SIZE] = $value;
        }

        # raw_sz is the physical size of the ldev (it includes the raid overhead).
        # for a thin device this size is relative to the subscribed size
        ($ret, $value) = get_capacity($obj_hash->{$ldev}, "raw_sz", "MB");
        if ($ret || !defined $value)
        {
            if (!defined $value)
            {
                $dbg_level = &MSG_LEVEL_WARN;
                $msg = "Logical disk raw size not specified";
            }
            else
            {
                $dbg_level = &MSG_LEVEL_ERROR;
                $msg = "Invalid value specified";
            }

            $bad_record->{raw_sz} = {
                                    fld_name => $def->{raw_sz},
                                    value => $obj_hash->{$ldev}->{raw_sz},
                                    dbg_level => $dbg_level,
                                    error_msg => $msg
                                 };
        }
        else
        {
            $ldev_row[&VRTS::Constants::P_ARRAY_LDEV::CONFIGURED_RAW_CAPACITY] = $value;
        }
        $ldev_row[&VRTS::Constants::P_ARRAY_LDEV::IS_ADMINISTRATIVE] = 0;
        $ldev_row[&VRTS::Constants::P_ARRAY_LDEV::IS_ONDEMAND] = 0;
        $ldev_row[&VRTS::Constants::P_ARRAY_LDEV::IS_FAULT_TOLERANT] = 0;
        if (defined $obj_hash->{$ldev}->{fault_tolerant} && $obj_hash->{$ldev}->{fault_tolerant} == 1)
        {
            $ldev_row[&VRTS::Constants::P_ARRAY_LDEV::IS_FAULT_TOLERANT] = 1;
        }
        $ldev_row[&VRTS::Constants::P_ARRAY_LDEV::STATUS] = 0;
        if (defined $obj_hash->{$ldev}->{assigned} && $obj_hash->{$ldev}->{assigned} ne &UNDEF_FIELD)
        {
            if (!exists $bool_values{$obj_hash->{$ldev}->{assigned}})
            {
                $bad_record->{assigned} = 
                {fld_name =>$def->{assigned},value =>$obj_hash->{$ldev}->{assigned},dbg_level=>&MSG_LEVEL_ERROR};
            }
            else
            { 
                $ldev_row[&VRTS::Constants::P_ARRAY_LDEV::STATUS] = $obj_hash->{$ldev}->{assigned};
            }
        }
        $ldev_row[&VRTS::Constants::P_ARRAY_LDEV::IS_VIRTUAL] = 0;
        if (defined $obj_hash->{$ldev}->{is_virtual} && $obj_hash->{$ldev}->{is_virtual} ne &UNDEF_FIELD)
        {
            if (!exists $bool_values{$obj_hash->{$ldev}->{is_virtual}})
            {
                $bad_record->{is_virtual} =
                {fld_name =>$def->{is_virtual},
                 value =>$obj_hash->{$ldev}->{is_virtual},dbg_level=>&MSG_LEVEL_ERROR};
            }
            else
            {
                $ldev_row[&VRTS::Constants::P_ARRAY_LDEV::IS_VIRTUAL] = $obj_hash->{$ldev}->{is_virtual};
            }
        }

        $ldev_row[&VRTS::Constants::P_ARRAY_LDEV::IS_ADMINISTRATIVE] = 0;
        if (defined $obj_hash->{$ldev}->{administrative} && 
            $obj_hash->{$ldev}->{administrative} ne &UNDEF_FIELD)
        {
            if (!exists $bool_values{$obj_hash->{$ldev}->{administrative}})
            {
                $bad_record->{administrative} =
                {fld_name =>$def->{administrative},
                 value =>$obj_hash->{$ldev}->{administrative},dbg_level=>&MSG_LEVEL_ERROR};
            } 
            elsif ($obj_hash->{$ldev}->{administrative} == &BOOL_VALUE_TRUE)
            {
                $ldev_row[&VRTS::Constants::P_ARRAY_LDEV::TYPE] = "ADMINLUN";
                $ldev_row[&VRTS::Constants::P_ARRAY_LDEV::IS_ADMINISTRATIVE] = 1; # required by VOM-A
            }
        }

        $ldev_row[&VRTS::Constants::P_ARRAY_LDEV::IS_ONDEMAND] = 0;
        if (defined $obj_hash->{$ldev}->{thin} && $obj_hash->{$ldev}->{thin} ne &UNDEF_FIELD)
        {
            if (!exists $bool_values{$obj_hash->{$ldev}->{thin}})
            {
                $bad_record->{thin} =
                {fld_name=>$def->{thin},value =>$obj_hash->{$ldev}->{thin},dbg_level=>&MSG_LEVEL_ERROR};
            }
            elsif ($obj_hash->{$ldev}->{thin} == &BOOL_VALUE_TRUE) {

            $ldev_row[&VRTS::Constants::P_ARRAY_LDEV::IS_ONDEMAND] = 1;

            # thin_consumed_sz is the actual consumed logical size of the ldev.
            # this is applicable only for thin devices
            ($ret, $value) = get_capacity($obj_hash->{$ldev}, "thin_consumed_sz", "MB");
            if ($ret || !defined $value)
            {
                if (!defined $value)
                {
                    $dbg_level = &MSG_LEVEL_WARN;
                    $msg = "Logical consumed size not specified";
                }
                else
                {
                    $dbg_level = &MSG_LEVEL_ERROR;
                    $msg = "Invalid value specified";
                }
                $bad_record->{thin_consumed_sz} = {
                                    fld_name => $def->{thin_consumed_sz},
                                    value => $obj_hash->{$ldev}->{thin_consumed_sz},
                                    dbg_level => $dbg_level,
                                    error_msg => $msg
                                 };
            }
            else
            {
                $ldev_row[&VRTS::Constants::P_ARRAY_LDEV::CONSUMED_LUN_SIZE] = $value;
                if (defined $ldev_row[&VRTS::Constants::P_ARRAY_LDEV::SIZE] &&
                        $ldev_row[&VRTS::Constants::P_ARRAY_LDEV::SIZE] ne "COLUMNDEFAULTVALUE" &&
                        $value > $ldev_row[&VRTS::Constants::P_ARRAY_LDEV::SIZE])
                {
                    $msg = "Consumed size should not be greater than logical size";
                    $dbg_level = &MSG_LEVEL_WARN;
                    $bad_record->{thin_consumed_sz} = {
                                    fld_name => $def->{thin_consumed_sz},
                                    value => $obj_hash->{$ldev}->{thin_consumed_sz},
                                    dbg_level => $dbg_level,
                                    error_msg => $msg
                                 };
                }
            }

            # thin_consumed_raw_sz is the actual consumed physical size of the ldev.
            # this is applicable only for thin devices
            ($ret, $value) = get_capacity($obj_hash->{$ldev}, "thin_consumed_raw_sz", "MB");
            if ($ret || !defined $value)
            {
                $dbg_level = &MSG_LEVEL_ERROR;
                if (!defined $value)
                {
                    $dbg_level = &MSG_LEVEL_WARN;
                    $msg = "Consumed raw size not specified";
                } 
                else
                {
                    $msg = "Invalid value specified";
                }
                $bad_record->{thin_consumed_raw_sz} = {
                                    fld_name => $def->{thin_consumed_raw_sz},
                                    value => $obj_hash->{$ldev}->{thin_consumed_raw_sz},
                                    dbg_level => $dbg_level,
                                    error_msg => $msg
                                 };
            }
            else
            {
                $ldev_row[&VRTS::Constants::P_ARRAY_LDEV::CONSUMED_RAW_CAPACITY] = $value;
            }
            }
        }
  
        $ldev_row[&VRTS::Constants::P_ARRAY_LDEV::REPLICATION] = &VRTS::Enclosure::Discover::NOT_REPLICATED;

        $ldev_tb->addrow(\@ldev_row);
        $ldevs_discovered = 1;
        $num_objects++;
        $no_ldevs++;
report_err:
        my @errors = keys %$bad_record;
        if (scalar(@errors))
        {
            $bad_record->{id} = {
                                    fld_name => $def->{id},
                                    value => $obj_hash->{$ldev}->{id},
                                    dbg_level  => &MSG_LEVEL_DBG
                                };
            print_record ($bad_record);
        }
    }
    log_msg("Total Records discovered: $num_objects");
    $ldev_tb->print_table() if ($debug_level == &MSG_LEVEL_DBG_3);
    return (undef, 0);
}

__DISCOVER_ADAPTERS__
# Description : method discovers adapters by executing Plugin
# discovery CLIs on the specified discovery host.
# Input :
#       agent - Reference to the Agentlet framework
#       vdid - VDID of the array
# Ouput:
#       msg - Error message on failure to discover
#       retval - 0 on success. non-zero on failure.
sub discover_adapters
{
    my $agent = shift;
    my $vdid = shift;
    my $num_objects = 0;
    my $msg;

    my $adapter_cmd = $discovery_cmdline." --list adapters --encl $vdid";
    log_msg ("Discovering Adapaters...", &MSG_LEVEL_DBG, 1);
    log_msg ("(cmd: $display_cmd_line --list adapters --encl $vdid)", &MSG_LEVEL_DBG);
    $discovery_ctxt->{CMD_LINE} = $adapter_cmd;

    my $out = VRTS::Enclosure::SIPlugin::Util::exec_cmd($ch, $discovery_ctxt);
    if ($out->[0] != 0)
    {
        $msg = "Failed to push/exec CLI: $out->[1]. $out->[0]";
        $msg = VRTS::Enclosure::SIPlugin::Util::get_error($msg, $out->[0]);
        log_msg ($msg, &MSG_LEVEL_ERROR);
        return undef, $out->[0];
    }
    my $fh = $out->[1];
    my ($obj_hash, $ret) = $parser->get_object("ARRAY_ADAPTER", $fh, "id");
    if ($ret)
    {
        $msg = "Vendor CLI failed to execute:$obj_hash:retcode $ret";
        $msg = VRTS::Enclosure::SIPlugin::Util::get_error($msg, $ret);
        log_msg ($msg, &MSG_LEVEL_ERROR);
        return (undef, $ret);
    }
    if (!defined $obj_hash )
    {
         $msg = VRTS::Enclosure::SIPlugin::Util::get_error($msg,
                &VRTS::Enclosure::SIPlugin::Catalog::NO_OBJECTS_DISCOVERED);
         log_msg ($msg);
         return undef, &VRTS::Enclosure::SIPlugin::Catalog::NO_OBJECTS_DISCOVERED;
    }
    my $adp_def = $objects_def->{ARRAY_ADAPTER};
 
    foreach my $adap (keys%$obj_hash)
    {
        my $bad_record = {};

        my @adp_row;
        $adp_row[&VRTS::Constants::P_ARRAY_ADAPTER::ID] = $obj_hash->{$adap}->{id};
        $adp_row[&VRTS::Constants::P_ARRAY_ADAPTER::SUBSYS_ID] = $vdid;

        # TODO UI does not show this. remove it. SDK guide does not document it.
        $adp_row[&VRTS::Constants::P_ARRAY_ADAPTER::CARD] = $obj_hash->{$adap}->{card}
            if (exists $obj_hash->{$adap}->{card});

        if (exists $obj_hash->{$adap}->{name} && $obj_hash->{$adap}->{name} ne &UNDEF_FIELD)
        {
            $adp_row[&VRTS::Constants::P_ARRAY_ADAPTER::DISPLAY_NAME] = $obj_hash->{$adap}->{name};
        }
        else
        {
            $adp_row[&VRTS::Constants::P_ARRAY_ADAPTER::DISPLAY_NAME] = 
               $adp_row[&VRTS::Constants::P_ARRAY_ADAPTER::ID];
        }

        $adp_row[&VRTS::Constants::P_ARRAY_ADAPTER::TYPE] = &VRTS::Enclosure::Utils::ADAPTER_TYPE_FC;
        if (exists $obj_hash->{$adap}->{type} && $obj_hash->{$adap}->{type} ne &UNDEF_FIELD)
        {
            if (!exists $adapter_type{$obj_hash->{$adap}->{type}})
            {
                $msg = "Invalid value specified. Allowed values are: @$adapter_type_list ";
                $bad_record->{type} = {
                                    fld_name => $adp_def->{type},
                                    value => $obj_hash->{$adap}->{type},
                                    dbg_level  => &MSG_LEVEL_ERROR,
                                    error_msg => $msg
                                };
            }
            else
            {
                $adp_row[&VRTS::Constants::P_ARRAY_ADAPTER::TYPE] = $obj_hash->{$adap}->{type};
            }
        }

        $adp_row[&VRTS::Constants::P_ARRAY_ADAPTER::ADAPTER_MODE] = &ADAPTER_MODE_NORMAL;
        if (exists $obj_hash->{$adap}->{mode} && $obj_hash->{$adap}->{mode} ne &UNDEF_FIELD)
        {
           VRTS::Logger::log('debug', "mode is $obj_hash->{$adap}->{mode}");
            if (!exists $adapter_mode{$obj_hash->{$adap}->{mode}})
            {
                $msg = "Invalid value specified. Allowed values are: ".
                       &ADAPTER_MODE_NORMAL.",".&ADAPTER_MODE_ERROR;
                $bad_record->{mode} = {
                                    fld_name => $adp_def->{mode},
                                    value => $obj_hash->{$adap}->{mode},
                                    dbg_level  => &MSG_LEVEL_ERROR,
                                    error_msg => $msg
                                };
            }
            else
            {
                $adp_row[&VRTS::Constants::P_ARRAY_ADAPTER::ADAPTER_MODE] = $obj_hash->{$adap}->{mode};
            }
        }

        # TODO UI does not show this. remove it. SDK guide does not document it.
        $adp_row[&VRTS::Constants::P_ARRAY_ADAPTER::CLUSTER] = 0;
        if (exists $obj_hash->{$adap}->{is_clustered} && $obj_hash->{$adap}->{is_clustered} == 1)
        {
            $adp_row[&VRTS::Constants::P_ARRAY_ADAPTER::CLUSTER] = $obj_hash->{$adap}->{is_clustered};
        }
        $adap_tb->addrow(\@adp_row);
        $num_objects++;

report_err:
        my @errors = keys %$bad_record;
        if (scalar(@errors))
        {
            $bad_record->{id} = {
                                    fld_name => $adp_def->{id},
                                    value => $obj_hash->{$adap}->{id},
                                    dbg_level  => &MSG_LEVEL_DBG
                                };
            print_record ($bad_record);
        }
    }
    log_msg("Total Records discovered: $num_objects");
    $adap_tb->print_table() if ($debug_level == &MSG_LEVEL_DBG_3);
    return (undef, 0);
}

__DISCOVER_ARRAY_PORTS__
# Description : method discovers array ports by executing Plugin
# discovery CLIs on the specified discovery host.
# Input :
#       agent - Reference to the Agentlet framework
#       vdid - VDID of the array
# Ouput:
#       msg - Error message on failure to discover
#       retval - 0 on success. non-zero on failure.
sub discover_aps
{
    my $agent = shift;
    my $vdid = shift;
    my $msg;  
    my $num_objects = 0;

    my $cmd = $discovery_cmdline." --list ports --encl $vdid";
    log_msg ("Discovering Array Ports...", &MSG_LEVEL_DBG,1);
    log_msg ("(cmd: $display_cmd_line --list ports --encl $vdid)...", &MSG_LEVEL_DBG);
    $discovery_ctxt->{CMD_LINE} = $cmd;

    my $out = VRTS::Enclosure::SIPlugin::Util::exec_cmd($ch, $discovery_ctxt);
    if ($out->[0] != 0)
    {
        $msg = "Failed to push/exec CLI: $out->[1]. $out->[0]";
        $msg = VRTS::Enclosure::SIPlugin::Util::get_error($msg, $out->[0]);
        log_msg ($msg, &MSG_LEVEL_ERROR);
        return undef, $out->[0];
    }
    my $fh = $out->[1];
    my ($obj_hash, $ret) = $parser->get_object("ARRAY_AP", $fh, "id");
    if ($ret)
    {
        $msg = "Vendor CLI failed to execute:$obj_hash:retcode $ret";
        $msg = VRTS::Enclosure::SIPlugin::Util::get_error($msg, $ret);
        log_msg ($msg, &MSG_LEVEL_ERROR);
        return (undef, $ret);
    }
    if (!defined $obj_hash )
    {
         $msg = VRTS::Enclosure::SIPlugin::Util::get_error($msg,
                &VRTS::Enclosure::SIPlugin::Catalog::NO_OBJECTS_DISCOVERED);
         log_msg ($msg);
         return undef, &VRTS::Enclosure::SIPlugin::Catalog::NO_OBJECTS_DISCOVERED;
    }

    my $ap_def = $objects_def->{ARRAY_AP};
    foreach my $ap (keys%$obj_hash)
    {
        my @ap_row;
        my $bad_record = {};

        $ap_row[&VRTS::Constants::P_ARRAY_AP::ID] = $obj_hash->{$ap}->{id};
        $ap_row[&VRTS::Constants::P_ARRAY_AP::SUBSYS_ID] = $vdid;
        if (exists $obj_hash->{$ap}->{name} && $obj_hash->{$ap}->{name} ne &UNDEF_FIELD)
        {
            $ap_row[&VRTS::Constants::P_ARRAY_AP::NAME] = $obj_hash->{$ap}->{name};
        }
        else
        {
            $ap_row[&VRTS::Constants::P_ARRAY_AP::NAME] = $ap_row[&VRTS::Constants::P_ARRAY_AP::ID];
        }

        $ap_row[&VRTS::Constants::P_ARRAY_AP::TYPE] = &VRTS::Enclosure::Utils::ADAPTER_TYPE_FC;
        if (exists $obj_hash->{$ap}->{type} && $obj_hash->{$ap}->{type} ne &UNDEF_FIELD)
        {
            if (!exists $adapter_type{$obj_hash->{$ap}->{type}})
            {
                $msg = "Invalid value specified. Allowed values are: @$adapter_type_list";
                $bad_record->{type} = {
                                    fld_name => $ap_def->{type},
                                    value => $obj_hash->{$ap}->{type},
                                    dbg_level  => &MSG_LEVEL_ERROR,
                                    error_msg => $msg
                                };
            }
            else
            {
                $ap_row[&VRTS::Constants::P_ARRAY_ADAPTER::TYPE] = $obj_hash->{$ap}->{type};
            }
        }

        if (!exists $obj_hash->{$ap}->{wwpn} || $obj_hash->{$ap}->{wwpn} eq &UNDEF_FIELD)
        {
            # Missing WWPN treated as error as this attribute is required for corelation
            $msg = "WWPN for array port not specified";
            $bad_record->{wwpn} = {
                                    fld_name => $ap_def->{wwpn},
                                    value => $obj_hash->{$ap}->{wwpn},
                                    dbg_level  => &MSG_LEVEL_ERROR,
                                    error_msg => $msg
                                };
        }
        else
        {
            $ap_row[&VRTS::Constants::P_ARRAY_AP::WWPN] = $obj_hash->{$ap}->{wwpn};
        }

        if (!exists $obj_hash->{$ap}->{wwnn} || $obj_hash->{$ap}->{wwnn} eq &UNDEF_FIELD)
        {
            $msg = "WWNN for array port not specified";
            $bad_record->{wwnn} = {
                                    fld_name => $ap_def->{wwnn},
                                    value => $obj_hash->{$ap}->{wwnn},
                                    dbg_level  => &MSG_LEVEL_WARN,
                                    error_msg => $msg
                                };
        }
        else
        {
            $ap_row[&VRTS::Constants::P_ARRAY_AP::WWNN] = $obj_hash->{$ap}->{wwnn};
        }

        if (!exists $obj_hash->{$ap}->{adapter_id} || $obj_hash->{$ap}->{adapter_id} eq &UNDEF_FIELD)
        {
            $msg = "Adapter for array port not specified";
            $bad_record->{adapter_id} = {
                                    fld_name => $ap_def->{adapter_id},
                                    value => $obj_hash->{$ap}->{adapter_id},
                                    dbg_level  => &MSG_LEVEL_WARN,
                                    error_msg => $msg
                                };
        }
        else
        {
            $ap_row[&VRTS::Constants::P_ARRAY_AP::ADAPTER_ID] = $obj_hash->{$ap}->{adapter_id};
        }

        $ap_row[&VRTS::Constants::P_ARRAY_AP::STATUS] = $port_status{&AP_STATUS_UP};
        if (exists $obj_hash->{$ap}->{status} && $obj_hash->{$ap}->{status} ne &UNDEF_FIELD)
        {
            if (!exists $port_status{$obj_hash->{$ap}->{status}}) 
            {
                $msg = "Invalid value specified. Allowed values are: ".
                       &AP_STATUS_UP.",".&AP_STATUS_DN;
                $bad_record->{status} = {
                                    fld_name => $ap_def->{status},
                                    value => $obj_hash->{$ap}->{status},
                                    dbg_level  => &MSG_LEVEL_ERROR,
                                    error_msg => $msg
                                };
            }
            else
            {
                $ap_row[&VRTS::Constants::P_ARRAY_AP::STATUS] = $port_status{$obj_hash->{$ap}->{status}};
            }
        }
        $ap_tb->addrow(\@ap_row);
        $num_objects++;

report_err:
        my @errors = keys %$bad_record;
        if (scalar(@errors))
        {
            $bad_record->{id} = {
                                    fld_name => $ap_def->{id},
                                    value => $obj_hash->{$ap}->{id},
                                    dbg_level  => &MSG_LEVEL_DBG
                                };
            print_record ($bad_record);
        }
    }
    log_msg("Total Records discovered: $num_objects");
    $ap_tb->print_table() if ($debug_level == &MSG_LEVEL_DBG_3);
    return (undef, 0);
}

__DISCOVER_LDEV_PDEV_MAP__
# Description : method discovers logical to physical device mapping by executing Plugin
# discovery CLIs on the specified discovery host.
# Input :
#       agent - Reference to the Agentlet framework
#       vdid - VDID of the array
# Ouput:
#       msg - Error message on failure to discover
#       retval - 0 on success. non-zero on failure.
sub discover_ldev_pdev_map
{
    my $agent = shift;
    my $vdid = shift;
    my $msg;
    my $num_objects = 0;

    my $cmd = $discovery_cmdline." --list ldevpdevmap --encl $vdid";
    if (!$pdevs_discovered)
    {
        log_msg ("\nSkipping discovery of logical to physical device mapping".
                 " as no physical devices were discovered");
        # Not returning error as it was intentionally skipped
        return undef, 0;
    }
    if (!$pdevs_discovered)
    {
        log_msg ("Skipping discovery of logical to physical device mapping".
                 " as no logical devices were discovered");
        return undef, 0;
    }

    log_msg ("Discovering Logical to Physical device mapping: ...", &MSG_LEVEL_DBG, 1);
    log_msg ("(cmd: $display_cmd_line --list ldevpdevmap --encl $vdid)...", &MSG_LEVEL_DBG);
    $discovery_ctxt->{CMD_LINE} = $cmd;

    my $out = VRTS::Enclosure::SIPlugin::Util::exec_cmd($ch, $discovery_ctxt);
    if ($out->[0] != 0)
    {
        $msg = "Failed to push/exec CLI: $out->[1]. $out->[0]";
        $msg = VRTS::Enclosure::SIPlugin::Util::get_error($msg, $out->[0]);
        log_msg ($msg, &MSG_LEVEL_ERROR);
        return undef, $out->[0];
    }
    my $fh = $out->[1];
    my ($obj_hash, $ret) = $parser->get_object("LDEV_PDEV_MAP", $fh, "ldev");
    if ($ret)
    {
        $msg = "Vendor CLI failed to execute:$obj_hash:retcode $ret";
        $msg = VRTS::Enclosure::SIPlugin::Util::get_error($msg, $ret);
        log_msg ($msg, &MSG_LEVEL_ERROR);
        return (undef, $ret);
    }
    if (!defined $obj_hash )
    {
         $msg = VRTS::Enclosure::SIPlugin::Util::get_error($msg,
                &VRTS::Enclosure::SIPlugin::Catalog::NO_OBJECTS_DISCOVERED);
         log_msg ($msg);
         return undef, &VRTS::Enclosure::SIPlugin::Catalog::NO_OBJECTS_DISCOVERED;
    }

    my $row = $capabilities_tb->getrowwithvalue(&VRTS::Constants::P_ARRAY_CAPABILITY::ARRAY_PROP_ID,
                      $capbs->{'LDEV_PDEV_MAP'}->{index});

    if (defined $row)
    {
        # if the ldevpdevmap command gives explicit mapping between pdevs and ldevs
        # then set LDEV_PDEV_MAP capability to 1 so that GUI always uses this
        # mapping even in cases where the enclosure is raid group capable and mapping between
        # RG-pdevs and RG-ldevs is available.
        $row->[&VRTS::Constants::P_ARRAY_CAPABILITY::VALUE] = 1;
        $capabilities_tb->addrow($row);
    }

    my $def = $objects_def->{LDEV_PDEV_MAP};

    foreach my $map (keys%$obj_hash)
    {
        my @row;
        my $bad_record = {};

        my $ldevrow = $ldev_tb->getrowwithvalue(&VRTS::Constants::P_ARRAY_LDEV::ID, $obj_hash->{$map}->{ldev});
        if (!defined $ldevrow)
        {
             $msg = "Invalid Logical disk specified. Ignoring the record";
             $bad_record->{ldev} = {
                                    fld_name => $def->{ldev},
                                    value => $obj_hash->{$map}->{ldev},
                                    dbg_level => &MSG_LEVEL_ERROR,
                                    error_msg => $msg
                                 };
             print_record ($bad_record);
             next;
        }
        my $pdevrow = $pdev_tb->getrowwithvalue(&VRTS::Constants::P_ARRAY_PDEV::ID, 
                      $obj_hash->{$map}->{pdev});
        if (!defined $pdevrow)
        {
             $msg = "Invalid Physical disk specified. Ignoring the record";
             $bad_record->{pdev} = {
                                    fld_name => $def->{pdev},
                                    value => $obj_hash->{$map}->{pdev},
                                    dbg_level => &MSG_LEVEL_ERROR,
                                    error_msg => $msg
                                 };
             goto report_err;
        }

        $row[&VRTS::Constants::P_ARRAY_PDEV_MAP::PRODUCER_ID] = $obj_hash->{$map}->{ldev};
        $row[&VRTS::Constants::P_ARRAY_PDEV_MAP::CONSUMER_TYPE] = "PDEV";
        $row[&VRTS::Constants::P_ARRAY_PDEV_MAP::CONSUMER_ID] = $obj_hash->{$map}->{pdev};
        $row[&VRTS::Constants::P_ARRAY_PDEV_MAP::PRODUCER_TYPE] = "LDEV";

        $pdevmap_tb->addrow(\@row);
        $num_objects++;
report_err:
        my @errors = keys %$bad_record;
        if (scalar(@errors))
        {
            $bad_record->{ldev} = {
                                    fld_name => $def->{ldev},
                                    value => $obj_hash->{$map}->{ldev},
                                    dbg_level  => &MSG_LEVEL_DBG
                                };
            print_record ($bad_record);
        }
    }
    log_msg("Total Records discovered: $num_objects");
    $pdevmap_tb->print_table() if ($debug_level == &MSG_LEVEL_DBG_3);
    return (undef, 0);
}

__DISCOVER_LDEV_HOST_MAP__
# Description : method discovers lun mapping to hosts
# Input :
#       agent - Reference to the Agentlet framework
#       vdid - VDID of the array
# Ouput:
#       msg - Error message on failure to discover
#       retval - 0 on success. non-zero on failure.
sub discover_ldev_host_map
{
    my $agent = shift;
    my $vdid = shift;
    my $msg;
    my $num_objects = 0;

    my $cmd = $discovery_cmdline." --list ldevhostmap --encl $vdid";

    log_msg ("Discovering Logical Disks to Host mapping...", &MSG_LEVEL_DBG,1 );
    log_msg ("(cmd: $display_cmd_line --list ldevhostmap --encl $vdid)...", &MSG_LEVEL_DBG);
    $discovery_ctxt->{CMD_LINE} = $cmd;

    my $out = VRTS::Enclosure::SIPlugin::Util::exec_cmd($ch, $discovery_ctxt);
    if ($out->[0] != 0)
    {
        $msg = "Failed to push/exec CLI: $out->[1]. $out->[0]";
        $msg = VRTS::Enclosure::SIPlugin::Util::get_error($msg, $out->[0]);
        log_msg ($msg, &MSG_LEVEL_ERROR);
        return undef, $out->[0];
    }
    my $fh = $out->[1];
    my ($obj_hash, $ret) = $parser->get_object("LDEV_HOST_MAP", $fh, "id");
    if ($ret)
    {
        $msg = "Vendor CLI failed to execute:$obj_hash:retcode $ret";
        $msg = VRTS::Enclosure::SIPlugin::Util::get_error($msg, $ret);
        log_msg ($msg, &MSG_LEVEL_ERROR);
        return (undef, $ret);
    }
    if (!defined $obj_hash )
    {
         $msg = VRTS::Enclosure::SIPlugin::Util::get_error($msg,
                &VRTS::Enclosure::SIPlugin::Catalog::NO_OBJECTS_DISCOVERED);
         log_msg ($msg);
         return undef, &VRTS::Enclosure::SIPlugin::Catalog::NO_OBJECTS_DISCOVERED;
    }
    my $def = $objects_def->{LDEV_HOST_MAP}; 

    foreach my $map (keys%$obj_hash)
    {
        my $bad_record = {};

        my @row;
        my $idrow = $ldev_tb->getrowwithvalue(&VRTS::Constants::P_ARRAY_LDEV::ID, $obj_hash->{$map}->{id});
        if (!defined $idrow)
        {
             $msg = "Invalid Logical disk specified. Ignoring the record";
             $bad_record->{id} = {
                                    fld_name => $def->{id},
                                    value => $obj_hash->{$map}->{id},
                                    dbg_level => &MSG_LEVEL_ERROR,
                                    error_msg => $msg
                                 };
             print_record ($bad_record);
             next;
        }

        if (!exists $obj_hash->{$map}->{array_wwpn} || $obj_hash->{$map}->{array_wwpn} eq &UNDEF_FIELD)
        {
             $msg = "Missing Array Port WWPN. Ignoring the record";
             $bad_record->{array_wwpn} = {
                                    fld_name => $def->{array_wwpn},
                                    value => $obj_hash->{$map}->{array_wwpn},
                                    dbg_level => &MSG_LEVEL_ERROR,
                                    error_msg => $msg
                                 };
             goto report_err;
        }

        $idrow = undef;
        $idrow = $ap_tb->getrowwithvalue(&VRTS::Constants::P_ARRAY_AP::ID,
                      $obj_hash->{$map}->{array_wwpn});
        if (!defined $idrow)
        {
             $msg = "Invalid Array Port specified. Ignoring the record";
             $bad_record->{array_wwpn} = {
                                    fld_name => $def->{array_wwpn},
                                    value => $obj_hash->{$map}->{array_wwpn},
                                    dbg_level => &MSG_LEVEL_ERROR,
                                    error_msg => $msg
                                 };
             goto report_err;
        }
        if (!exists $obj_hash->{$map}->{hba_wwpn} ||  $obj_hash->{$map}->{hba_wwpn} eq &UNDEF_FIELD)
        {
             $msg = "Missing HBA Port WWPN. Ignoring the record";
             $bad_record->{hba_wwpn} = {
                                    fld_name => $def->{hba_wwpn},
                                    value => $obj_hash->{$map}->{hba_wwpn},
                                    dbg_level => &MSG_LEVEL_ERROR,
                                    error_msg => $msg
                                 };
             goto report_err;
        }

        if (!exists $obj_hash->{$map}->{scsi_lunid} ||  $obj_hash->{$map}->{scsi_lunid} eq &UNDEF_FIELD)
        {
             $msg = "Missing SCSI LunId. Ignoring the record";
             $bad_record->{scsi_lunid} = {
                                    fld_name => $def->{scsi_lunid},
                                    value => $obj_hash->{$map}->{scsi_lunid},
                                    dbg_level => &MSG_LEVEL_ERROR,
                                    error_msg => $msg
                                 };
             goto report_err;
        }
        elsif ($obj_hash->{$map}->{scsi_lunid} !~ /^[0-9A-Fa-f]+$/)
        {
             $msg = "SCSI LunId should be a hex number(e.g. 3C).Ignoring the record";
             $bad_record->{scsi_lunid} = {
                                    fld_name => $def->{scsi_lunid},
                                    value => $obj_hash->{$map}->{scsi_lunid},
                                    dbg_level => &MSG_LEVEL_ERROR,
                                    error_msg => $msg
                                 };
             goto report_err;
        }

        $row[&VRTS::Constants::P_LUN_MAPPING::HBA_PORT] = $obj_hash->{$map}->{hba_wwpn};
        $row[&VRTS::Constants::P_LUN_MAPPING::SCSI_LUN_ID] = uc ($obj_hash->{$map}->{scsi_lunid});
        $row[&VRTS::Constants::P_LUN_MAPPING::LDEV_ID] = $obj_hash->{$map}->{id};
        $row[&VRTS::Constants::P_LUN_MAPPING::TARGET_ID] = $obj_hash->{$map}->{array_wwpn};
        $row[&VRTS::Constants::P_LUN_MAPPING::SUBSYS_ID] = $vdid;
        $lunmap_tb->addrow(\@row);
        $num_objects++;

report_err:
        my @errors = keys %$bad_record;
        if (scalar(@errors))
        {
            $bad_record->{id} = {
                                    fld_name => $def->{id},
                                    value => $obj_hash->{$map}->{id},
                                    dbg_level  => &MSG_LEVEL_DBG
                                };
            print_record ($bad_record);
        }
    }
    log_msg("Total Records discovered: $num_objects");
    $lunmap_tb->print_table() if ($debug_level == &MSG_LEVEL_DBG_3);
    return (undef, 0);
}

__DISCOVER_ARRAY_GROUPS__
# Description : method discovers raidgroups/thinpools
# using a common method to discover them as most of the fields overlap
# Input :
#       agent - Reference to the Agentlet framework
#       vdid - VDID of the array
# Ouput:
#       msg - Error message on failure to discover
#       retval - 0 on success. non-zero on failure.
sub discover_arraygrps
{
    my $agent = shift;
    my $vdid = shift;
    my $devgrp_type = shift;
    my $msg;
    my $object_type;
    my $object_def_name;
    my $def;
    my $cmd;
    my $total_sz = undef;
    my $free_sz = undef;
    my $subscribed_sz = undef;
    my $tp_tuneables;
    my $num_objects = 0;

    if ($devgrp_type eq "RG")
    {
        $cmd = $discovery_cmdline." --list raidgroups --encl $vdid";
        log_msg ("Discovering RaidGroups...", &MSG_LEVEL_DBG,1 );
        log_msg ("(cmd: $display_cmd_line --list raidgroups --encl $vdid)...", &MSG_LEVEL_DBG);
        $object_type = "Raid group";
        $object_def_name = "RG";
    }
    else
    {
        $cmd = $discovery_cmdline." --list thinpools --encl $vdid";
        log_msg ("Discovering Thinpools...", &MSG_LEVEL_DBG,1 );
        log_msg ("(cmd: $display_cmd_line --list thinpools --encl $vdid)...", &MSG_LEVEL_DBG);
        $object_type = "Thin Pool";
        $object_def_name = "TP";
        $tp_tuneables = VRTS::Enclosure::Utils::read_tp_tuneables($method, $vdid);
    }
    $discovery_ctxt->{CMD_LINE} = $cmd;

    my $out = VRTS::Enclosure::SIPlugin::Util::exec_cmd($ch, $discovery_ctxt);
    if ($out->[0] != 0)
    {
        $msg = "Failed to push/exec CLI: $out->[1]. $out->[0]";
        $msg = VRTS::Enclosure::SIPlugin::Util::get_error($msg, $out->[0]);
        log_msg ($msg, &MSG_LEVEL_ERROR);
        return undef, $out->[0];
    }
    my $fh = $out->[1];
    my ($obj_hash, $ret) = $parser->get_object($object_def_name, $fh, "id");
    if ($ret)
    {
        $msg = "Vendor CLI failed to execute:$obj_hash:retcode $ret";
        $msg = VRTS::Enclosure::SIPlugin::Util::get_error($msg, $ret);
        log_msg ($msg, &MSG_LEVEL_ERROR);
        return (undef, $ret);
    }
    if (!defined $obj_hash )
    {
         $msg = VRTS::Enclosure::SIPlugin::Util::get_error($msg,
                &VRTS::Enclosure::SIPlugin::Catalog::NO_OBJECTS_DISCOVERED);
         log_msg ($msg);
         return undef, &VRTS::Enclosure::SIPlugin::Catalog::NO_OBJECTS_DISCOVERED;
    }

    my $def = $objects_def->{$object_def_name};
    my $val, $ret;
    my $msg;

    foreach my $devgrp (keys%$obj_hash)
    {
        my $bad_record = ();
        my @row;
        $row[&VRTS::Constants::P_ARRAY_DEVICE_GROUP::ID] = $obj_hash->{$devgrp}->{id};
        $row[&VRTS::Constants::P_ARRAY_DEVICE_GROUP::SUBSYS_ID] = $vdid;
        if (exists $obj_hash->{$devgrp}->{name} && $obj_hash->{$devgrp}->{name} ne "")
        { 
           $row[&VRTS::Constants::P_ARRAY_DEVICE_GROUP::NAME] = $obj_hash->{$devgrp}->{name};
        }
        else
        {
           $row[&VRTS::Constants::P_ARRAY_DEVICE_GROUP::NAME] = $obj_hash->{$devgrp}->{id};
        } 

        $row[&VRTS::Constants::P_ARRAY_DEVICE_GROUP::STATE] = &RG_STATUS_FAILED;
        $val = $obj_hash->{$devgrp}->{state};
        if (defined $val && $val ne &UNDEF_FIELD)
        {
            if (!exists $rg_states{$val})
            {
                $msg = "Invalid value specified. Possible values are [".&RG_STATUS_HEALTHY.",".
                       &RG_STATUS_FAILED."]";
                $bad_record->{state} = {
                                           fld_name => $def->{state},
                                           value => $val,
                                           dbg_level => &MSG_LEVEL_ERROR,
                                           error_msg => $msg
                                       };
            }
            else
            {
                # VOM DB stores healthy rg/tp state as 1 and non healthy as 0.
                # but the discovery script specifies this as 0 for healthy and 1 for failed
                # to be consistent with state fields for other objects.
                $row[&VRTS::Constants::P_ARRAY_DEVICE_GROUP::STATE] = $rg_states{$val};
            }
        }

        $val = $obj_hash->{$devgrp}->{dev_cnt};
        if (defined $val && $val ne &UNDEF_FIELD)
        {
            if ($val !~ /\d+/)
            {
                $msg = "Invalid value specified. Should be an integer value";
                $bad_record->{dev_cnt} = {
                                           fld_name => $def->{dev_cnt},
                                           value => $val,
                                           dbg_level => &MSG_LEVEL_ERROR,
                                           error_msg => $msg
                                       };
            }
            else
            {
                $row[&VRTS::Constants::P_ARRAY_DEVICE_GROUP::NDEVICES] = $val;
            }
        }

        ($ret, $val) = get_capacity($obj_hash->{$devgrp}, "raw_sz", "MB");
        if ($ret) {
            $bad_record->{raw_sz} = {
                                        fld_name => $def->{raw_sz},
                                        value => $val,
                                        dbg_level => &MSG_LEVEL_ERROR
                                    };
        }
        else
        {
            $row[&VRTS::Constants::P_ARRAY_DEVICE_GROUP::RAW_SIZE] = $val
                if (defined $val);
        }

        ($ret, $val) = get_capacity($obj_hash->{$devgrp}, "logical_sz", "MB");
        if ($ret) {
            $bad_record->{logical_sz} = {
                                        fld_name => $def->{logical_sz},
                                        value => $val,
                                        dbg_level => &MSG_LEVEL_ERROR
                                    };
        }
        else
        {
            if (defined $val)
            {
               $total_sz = $val;
               $row[&VRTS::Constants::P_ARRAY_DEVICE_GROUP::TOTAL_SIZE] = $val;
            }
        }

        ($ret, $val) = get_capacity($obj_hash->{$devgrp}, "overhead_sz", "MB");
        if ($ret) {
            $bad_record->{overhead_sz} = {
                                        fld_name => $def->{overhead_sz},
                                        value => $val,
                                        dbg_level => &MSG_LEVEL_ERROR
                                    };
        }
        else
        {
            $row[&VRTS::Constants::P_ARRAY_DEVICE_GROUP::OVERHEAD_SIZE] = $val
                if (defined $val);
        }

        ($ret, $val) = get_capacity($obj_hash->{$devgrp}, "logical_used_sz", "MB");
        if ($ret) {
            $bad_record->{logical_used_sz} = {
                                        fld_name => $def->{logical_used_sz},
                                        value => $val,
                                        dbg_level => &MSG_LEVEL_ERROR
                                    };
        }
        else
        {
            $row[&VRTS::Constants::P_ARRAY_DEVICE_GROUP::DEVICE_SIZE] = $val
                if (defined $val);
        }

        ($ret, $val) = get_capacity($obj_hash->{$devgrp}, "logical_free_sz", "MB");
        if ($ret) {
            $bad_record->{logical_free_sz} = {
                                        fld_name => $def->{logical_free_sz},
                                        value => $val,
                                        dbg_level => &MSG_LEVEL_ERROR
                                    };
        }
        else
        {
            if (defined $val)
            {
               $free_sz = $val;
               $row[&VRTS::Constants::P_ARRAY_DEVICE_GROUP::FREE_SIZE] = $val;
            }
        }

        if (exists $obj_hash->{$devgrp}->{subscribed_sz})
        {
            ($ret, $val) = get_capacity($obj_hash->{$devgrp}, "subscribed_sz", "MB");
            if ($ret) {
                $bad_record->{subscribed_sz} = {
                                        fld_name => $def->{subscribed_sz},
                                        value => $val,
                                        dbg_level => &MSG_LEVEL_ERROR
                                    };
            }
            else
            {
                if (defined $val)
                {
                    $subscribed_sz = $val;
                    $row[&VRTS::Constants::P_ARRAY_DEVICE_GROUP::VVOL_SIZE] = $val;
                }
            }
        }


        $val = $obj_hash->{$devgrp}->{vol_cnt};
        if (defined $val && $val ne &UNDEF_FIELD)
        {
            if ($val !~ /\d+/)
            {
                $msg = "Invalid value specified. Should be an integer value";
                $bad_record->{vol_cnt} = {
                                           fld_name => $def->{vol_cnt},
                                           value => $val,
                                           dbg_level => &MSG_LEVEL_ERROR,
                                           error_msg => $msg
                                       };
            }
            else
            {
                $row[&VRTS::Constants::P_ARRAY_DEVICE_GROUP::NVVOLS] = $val;
            }
        }

        # TODO Allow RGs with no raid levels? GUI does not show such records.
        $val = $obj_hash->{$devgrp}->{raid_level};
        if (defined $val && $val ne &UNDEF_FIELD)
        {
            $row[&VRTS::Constants::P_ARRAY_DEVICE_GROUP::RAID_LEVEL] = $val;
        }

        $row[&VRTS::Constants::P_ARRAY_DEVICE_GROUP::TYPE] = $devgrp_type;
        if ($devgrp_type eq "TP" && defined $total_sz && defined $free_sz && defined $subscribed_sz)
        {
            my $pct_consumed = (($total_sz - $free_sz) * 100 / $total_sz);
            my $pct_subs = (($subscribed_sz * 100 ) / $total_sz);
            VRTS::Enclosure::Faults::compute_tp_faults($agent, $row[&VRTS::Constants::P_ARRAY_DEVICE_GROUP::NAME],
                  $row[&VRTS::Constants::P_ARRAY_DEVICE_GROUP::ID],
                  $pct_consumed, $pct_subs, $vdid, $vdid, $tp_tuneables);
        }

        $devgrp_tb->addrow(\@row);
        $num_objects++;
        my @errors = keys %$bad_record;
        if (scalar(@errors))
        {
            $bad_record->{id} = {
                                    fld_name => $def->{id},
                                    value => $row[&VRTS::Constants::P_ARRAY_DEVICE_GROUP::ID],
                                    dbg_level  => &MSG_LEVEL_DBG
                                };
            print_record ($bad_record);
        }
    }
    log_msg("Total Records discovered: $num_objects");
    $devgrp_tb->print_table() if ($debug_level == &MSG_LEVEL_DBG_3);
    return (undef, 0);
}

__DISCOVER_RGPDEV_MAP__
# Description : method discovers raidgroup to physical disk mapping
# Input :
#       agent - Reference to the Agentlet framework
#       vdid - VDID of the array
# Ouput:
#       msg - Error message on failure to discover
#       retval - 0 on success. non-zero on failure.
sub discover_rgpdev_map
{
    my $agent = shift;
    my $vdid = shift;
    my $msg;
    my $num_objects = 0;
    my $cmd = $discovery_cmdline." --list rgpdevmap --encl $vdid";

    log_msg ("Discovering raidgroup to physical disk mapping...", &MSG_LEVEL_DBG,1 );
    log_msg ("(cmd: $display_cmd_line --list rgpdevmap --encl $vdid)...", &MSG_LEVEL_DBG);
    $discovery_ctxt->{CMD_LINE} = $cmd;
    my $out = VRTS::Enclosure::SIPlugin::Util::exec_cmd($ch, $discovery_ctxt);
    if ($out->[0] != 0)
    {
        $msg = "Failed to push/exec CLI: $out->[1]. $out->[0]";
        $msg = VRTS::Enclosure::SIPlugin::Util::get_error($msg, $out->[0]);
        log_msg ($msg, &MSG_LEVEL_ERROR);
        return undef, $out->[0];
    }
    my $fh = $out->[1];
    my ($obj_hash, $ret) = $parser->get_object("RG_PDEV_MAP", $fh, "dev_grp");
    if ($ret)
    {
        $msg = "Vendor CLI failed to execute:$obj_hash:retcode $ret";
        $msg = VRTS::Enclosure::SIPlugin::Util::get_error($msg, $ret);
        log_msg ($msg, &MSG_LEVEL_ERROR);
        return (undef, $ret);
    }
    if (!defined $obj_hash )
    {
         $msg = VRTS::Enclosure::SIPlugin::Util::get_error($msg,
                &VRTS::Enclosure::SIPlugin::Catalog::NO_OBJECTS_DISCOVERED);
         log_msg ($msg);
         return undef, &VRTS::Enclosure::SIPlugin::Catalog::NO_OBJECTS_DISCOVERED;
    }

    my $def = $objects_def->{RG_PDEV_MAP};
    my $val, $ret;
    my $msg;

    foreach my $map (keys%$obj_hash)
    {
        my @row;
        my $bad_record = {};
        $bad_record->{dev_grp} = {
                               fld_name => $def->{dev_grp},
                               value => $obj_hash->{$map}->{dev_grp}
                            };

        my $idrow = $devgrp_tb->getrowwithvalue(&VRTS::Constants::P_ARRAY_DEVICE_GROUP::ID, 
                  $obj_hash->{$map}->{dev_grp});
        if (!defined $idrow)
        {
            $msg = "No such raid group discovered";
            $bad_record->{dev_grp}->{error_msg} = $msg;
            $bad_record->{dev_grp}->{dbg_level} = &MSG_LEVEL_ERROR;
            print_record ($bad_record);
            next;
        }

        $idrow = undef;
        $idrow = $pdev_tb->getrowwithvalue(&VRTS::Constants::P_ARRAY_PDEV::ID,
                  $obj_hash->{$map}->{pdev});
        if (!defined $idrow)
        {
            $msg = "No such physical disk discovered";
            $bad_record->{pdev} = {
                                    fld_name => $def->{pdev},
                                    value => $obj_hash->{$map}->{pdev},
                                    dbg_level  => &MSG_LEVEL_ERROR,
                                    error_msg => $msg
                                };
            print_record ($bad_record);
            next;
        }

        $row[&VRTS::Constants::P_ARRAY_DEVICEGROUP_MAP::DEVICE_GROUP_ID] = $obj_hash->{$map}->{dev_grp};
        $row[&VRTS::Constants::P_ARRAY_DEVICEGROUP_MAP::DEVICE_GROUP_TYPE] = "RG";
        $row[&VRTS::Constants::P_ARRAY_DEVICEGROUP_MAP::DEVICE_ID] = $obj_hash->{$map}->{pdev};
        $row[&VRTS::Constants::P_ARRAY_DEVICEGROUP_MAP::DEVICE_TYPE] = "PDEV";
        $row[&VRTS::Constants::P_ARRAY_DEVICEGROUP_MAP::SUBSYS_ID] = $vdid;
        $row[&VRTS::Constants::P_ARRAY_DEVICEGROUP_MAP::DEVICE_GROUP_NAME] = $obj_hash->{$map}->{dev_grp};
        $devgrp_map_tb->addrow(\@row);
        $num_objects++;
    }
    log_msg("Total Records discovered: $num_objects");
    $devgrp_map_tb->print_table() if ($debug_level == &MSG_LEVEL_DBG_3);
    return (undef, 0);
}

__DISCOVER_DEVGRP_LDEV_MAP__
# Description : method discovers raidgroup/thinpools to logical disk mapping
# Input :
#       agent - Reference to the Agentlet framework
#       vdid - VDID of the array
#       devgrp_type - type of device group - RG or TP
#       device_type - type of LUN - ldev, source ldev for TP
# Ouput:
#       msg - Error message on failure to discover
#       retval - 0 on success. non-zero on failure.
sub discover_devgrp_ldev_map
{
    my $agent = shift;
    my $vdid = shift;
    my $devgrp_type = shift;
    my $device_type = shift;
    my $msg;
    my $object_type;
    my $object_def_name;
    my $def;
    my $cmd;
    my $display_cmd;
    my $num_objects = 0;

    if ($devgrp_type eq "RG")
    {
        $cmd = $discovery_cmdline." --list rgldevmap --encl $vdid";
        log_msg ("Discovering raidgroup to logical disk mapping...", &MSG_LEVEL_DBG,1 );
        $object_type = "Raid group";
        $object_def_name = "RG_LDEV_MAP";
        $display_cmd = $display_cmd_line." --list rgldevmap --encl $vdid";
    }
    else
    {
        if ($device_type eq "LDEV")
        {
            $cmd = $discovery_cmdline." --list tpldevmap --encl $vdid";
            $display_cmd = $display_cmd_line." --list tpldevmap --encl $vdid";
        }
        elsif ($device_type eq "SOURCE_LDEV")
        {
            $cmd = $discovery_cmdline." --list tpsrcldevmap --encl $vdid";
            $display_cmd = $display_cmd_line." --list tpsrcldevmap --encl $vdid";
        }
      
        log_msg ("Discovering thinpool to logical disk mapping...", &MSG_LEVEL_DBG,1 );
        $object_type = "Thin pool";
        $object_def_name = "TP_LDEV_MAP";
    }

    log_msg ("(cmd: $display_cmd)...", &MSG_LEVEL_DBG);
    $discovery_ctxt->{CMD_LINE} = $cmd;

    my $out = VRTS::Enclosure::SIPlugin::Util::exec_cmd($ch, $discovery_ctxt);
    if ($out->[0] != 0)
    {
        $msg = "Failed to push/exec CLI: $out->[1]. $out->[0]";
        $msg = VRTS::Enclosure::SIPlugin::Util::get_error($msg, $out->[0]);
        log_msg ($msg, &MSG_LEVEL_ERROR);
        return undef, $out->[0];
    }

    my $fh = $out->[1];
    my ($obj_hash, $ret) = $parser->get_object($object_def_name, $fh, "dev_grp");
    if ($ret)
    {
        $msg = "Vendor CLI failed to execute:$obj_hash:retcode $ret";
        $msg = VRTS::Enclosure::SIPlugin::Util::get_error($msg, $ret);
        log_msg ($msg, &MSG_LEVEL_ERROR);
        return (undef, $ret);
    }
    if (!defined $obj_hash )
    {
         $msg = VRTS::Enclosure::SIPlugin::Util::get_error($msg,
                &VRTS::Enclosure::SIPlugin::Catalog::NO_OBJECTS_DISCOVERED);
         log_msg ($msg);
         return undef, &VRTS::Enclosure::SIPlugin::Catalog::NO_OBJECTS_DISCOVERED;
    }

    $def = $objects_def->{$object_def_name};
    my $val, $ret;
    my $msg;

    foreach my $map (keys%$obj_hash)
    {
        my @row;
        my $bad_record = {};
        $bad_record->{dev_grp} = {
                                fld_name => $def->{dev_grp},
                                value => $obj_hash->{$map}->{dev_grp}
                            };

        my $idrow = $devgrp_tb->getrowwithvalue(&VRTS::Constants::P_ARRAY_DEVICE_GROUP::ID,
                  $obj_hash->{$map}->{dev_grp});
        if (!defined $idrow)
        {
            $msg = "Invalid $object_type specified. Ignoring the record";
            $bad_record->{dev_grp}->{error_msg} = $msg;
            $bad_record->{dev_grp}->{dbg_level} = &MSG_LEVEL_ERROR;
            print_record ($bad_record);
            next;
        }

        my $ldevrow = $ldev_tb->getrowwithvalue(&VRTS::Constants::P_ARRAY_LDEV::ID,
                  $obj_hash->{$map}->{ldev});
        if (!defined $ldevrow)
        {
            $msg = "Invalid Logical disk specified. Ignoring the record";
            $bad_record->{pdev} = {
                                    fld_name => $def->{ldev},
                                    value => $obj_hash->{$map}->{ldev},
                                    dbg_level  => &MSG_LEVEL_ERROR,
                                    error_msg => $msg
                                };
            print_record ($bad_record);
            next;
        }
        
        $row[&VRTS::Constants::P_ARRAY_DEVICEGROUP_MAP::DEVICE_GROUP_ID] = $obj_hash->{$map}->{dev_grp};
        $row[&VRTS::Constants::P_ARRAY_DEVICEGROUP_MAP::DEVICE_GROUP_TYPE] = $devgrp_type;
        $row[&VRTS::Constants::P_ARRAY_DEVICEGROUP_MAP::DEVICE_ID] = $obj_hash->{$map}->{ldev};
        $row[&VRTS::Constants::P_ARRAY_DEVICEGROUP_MAP::DEVICE_TYPE] = "LDEV";
        $row[&VRTS::Constants::P_ARRAY_DEVICEGROUP_MAP::SUBSYS_ID] = $vdid;
        $row[&VRTS::Constants::P_ARRAY_DEVICEGROUP_MAP::DEVICE_GROUP_NAME] = $obj_hash->{$map}->{dev_grp};
        $devgrp_map_tb->addrow(\@row);
        $num_objects++;

        if ($device_type eq "SOURCE_LDEV")
        {
            # change type on LDEV object to designate it as ldev backing up a thin pool.
            # this ensures its capacity is not accounted for as logical storage and there is no
            # double counting of capacity.
            $ldevrow->[&VRTS::Constants::P_ARRAY_LDEV::TYPE] = "SOURCELUN";
            $ldev_tb->addrow($ldevrow);
        }
    }
    log_msg("Total Records discovered: $num_objects");
    $devgrp_map_tb->print_table() if ($debug_level == &MSG_LEVEL_DBG_3);
    return (undef, 0);
}

__DISCOVER_REPLICATIONS__
# Description : method discovers replications
# Input :
#       agent - Reference to the Agentlet framework
#       vdid - VDID of the array
# Ouput:
#       msg - Error message on failure to discover
#       retval - 0 on success. non-zero on failure.
sub discover_replications
{
    my $agent = shift;
    my $vdid = shift;
    my $msg;
    my $num_objects = 0;
 
    my $cmd = $discovery_cmdline." --list replications --encl $vdid";
    log_msg ("\nDiscovering Replications...", &MSG_LEVEL_DBG, 1);
    log_msg ("(cmd: $cmd)...", &MSG_LEVEL_DBG);
    $discovery_ctxt->{CMD_LINE} = $cmd;

    my $out = VRTS::Enclosure::SIPlugin::Util::exec_cmd($ch, $discovery_ctxt);
    if ($out->[0] != 0)
    {
        $msg = "Failed to push/exec CLI: $out->[1]. $out->[0]";
        $msg = VRTS::Enclosure::SIPlugin::Util::get_error($msg, $out->[0]);
        log_msg ($msg, &MSG_LEVEL_ERROR);
        return undef, $out->[0];
    }
    my $fh = $out->[1];
    my ($obj_hash, $ret) = $parser->get_object("REPLICATION", $fh, "id");
    if ($ret)
    {
        $msg = "Vendor CLI failed to execute:$obj_hash:retcode $ret";
        $msg = VRTS::Enclosure::SIPlugin::Util::get_error($msg, $ret);
        log_msg ($msg, &MSG_LEVEL_ERROR);
        return (undef, $ret);
    }
	
    if (!defined $obj_hash )
    {
         $msg = VRTS::Enclosure::SIPlugin::Util::get_error($msg,
                &VRTS::Enclosure::SIPlugin::Catalog::NO_OBJECTS_DISCOVERED);
         log_msg ($msg);
         return undef, &VRTS::Enclosure::SIPlugin::Catalog::NO_OBJECTS_DISCOVERED;
    }

    my $def = $objects_def->{REPLICATION};
    my $val;
    my %enc_hash;

    foreach my $map (keys%$obj_hash)
    {
        my @row;
        my $bad_record = {};
        my $partner_row;
        my $remote_encl_id = "Unknown";
        my $partner_id;
        my $partner_name;
        my $partner_alt_id;

        my $ldev_row = $ldev_tb->getrowwithvalue(&VRTS::Constants::P_ARRAY_LDEV::ID, $obj_hash->{$map}->{id});
        if (!defined $ldev_row)
        {
             $msg = "Invalid Logical disk specified. Ignoring the record";
             $bad_record->{id} = {
                                    fld_name => $def->{id},
                                    value => $obj_hash->{$map}->{id},
                                    dbg_level => &MSG_LEVEL_ERROR,
                                    error_msg => $msg
                                 };
             print_record ($bad_record);
             next;
        }
        $partner_name = $obj_hash->{$map}->{partner_name};
        $partner_id = $obj_hash->{$map}->{partner_id};
        if (!defined $partner_id || $partner_id eq &UNDEF_FIELD)
        {
            $partner_id = "Unknown";
        }

        $partner_alt_id = $obj_hash->{$map}->{partner_alt_id};

        $val = $obj_hash->{$map}->{designation};
        if (!defined $val || $val eq &UNDEF_FIELD)
        {
             $msg = "Mandatory attribute missing. Ignoring the record\n";
             $bad_record->{designation} = {
                                              fld_name => $def->{designation},
                                              dbg_level => &MSG_LEVEL_ERROR,
                                              error_msg => $msg
                                          };
             goto report_err;
        }

        if (!exists $rep_designations{$val})
        {
             $msg = "Invalid value specified. Possible values are [".&REP_DESIG_SOURCE.",".
                    &REP_DESIG_TARGET."].Ignoring the record";
             $bad_record->{designation} = {
                                              fld_name => $def->{designation},
                                              value => $val,
                                              dbg_level => &MSG_LEVEL_ERROR,
                                              error_msg => $msg
                                          };
             goto report_err;
        }

        $val = $obj_hash->{$map}->{across_encls};
        if (!defined $val || $val eq &UNDEF_FIELD)
        {
             $msg = "Mandatory attribute missing. Ignoring the record\n";
             $bad_record->{across_encls} = {
                                              fld_name => $def->{across_encls},
                                              dbg_level => &MSG_LEVEL_ERROR,
                                              error_msg => $msg
                                          };
             goto report_err;
        }
        if (!exists $rep_across_encls{$val})
        {
             $msg = "Invalid value specified. Possible values are [".&REP_LOCAL.",".
                    &REP_REMOTE."].Ignoring the record";
             $bad_record->{across_encls} = {
                                              fld_name => $def->{across_encls},
                                              value => $val,
                                              dbg_level => &MSG_LEVEL_ERROR,
                                              error_msg => $msg
                                          };
             goto report_err;
        }
        
        if ($val == &REP_LOCAL)
        {
            $ldev_row->[&VRTS::Constants::P_ARRAY_LDEV::REPLICATION] = 
                &VRTS::Enclosure::Discover::LOCAL_REPLICATION;
            $val = $obj_hash->{$map}->{partner_id};
            if (defined $val && $val ne &UNDEF_FIELD)
            {
                $partner_id = $val;
                $partner_row = $ldev_tb->getrowwithvalue(&VRTS::Constants::P_ARRAY_LDEV::ID,
                               $obj_hash->{$map}->{partner_id});
                if (!defined $partner_row)
                {
                    $msg = "Invalid Logical disk specified. Ignoring the record";
                    $bad_record->{partner_id} = {
                                        fld_name => $def->{partner_id},
                                        value => $obj_hash->{$map}->{partner_id},
                                        dbg_level => &MSG_LEVEL_ERROR,
                                        error_msg => $msg
                                    };
                    goto report_err;
                }
                $partner_row->[&VRTS::Constants::P_ARRAY_LDEV::REPLICATION] =
                    &VRTS::Enclosure::Discover::LOCAL_REPLICATION;
                $partner_name = $partner_row->[&VRTS::Constants::P_ARRAY_LDEV::NAME];
            }
        }
        else
        {
            if ($obj_hash->{$map}->{designation} == &REP_DESIG_SOURCE) 
            {
                $ldev_row->[&VRTS::Constants::P_ARRAY_LDEV::REPLICATION] =
                    &VRTS::Enclosure::Discover::REMOTE_REPLICATION_PRIMARY;
            }
            else
            {
                $ldev_row->[&VRTS::Constants::P_ARRAY_LDEV::REPLICATION] =
                    &VRTS::Enclosure::Discover::REMOTE_REPLICATION_SECONDARY;
            }

            if (!defined $obj_hash->{$map}->{remote_encl_id} ||
             $obj_hash->{$map}->{remote_encl_id} eq &UNDEF_FIELD)
            {
                # If remote encl id is not specified but we have the enclosure name,
                # vendor and model then we can try to fetch the encl id from the database
                # assuming the remote enclosure is also configured in VOM
                my $remote_encl_name = $obj_hash->{$map}->{remote_encl_name};
                my $remote_encl_vendor = $obj_hash->{$map}->{remote_encl_vendor};
                my $remote_encl_model = $obj_hash->{$map}->{remote_encl_model};
                my $remote_encl_alt_id = $obj_hash->{$map}->{remote_encl_alt_id};

                if (!defined $remote_encl_vendor || $remote_encl_vendor eq &UNDEF_FIELD)
                {
                    $remote_encl_vendor = $vendor;
                }

                if (defined $remote_encl_name && $remote_encl_name ne &UNDEF_FIELD)
                {
                    if (!defined $remote_encl_model || $remote_encl_model eq &UNDEF_FIELD)
                    {
                        $remote_encl_model = $model;
                    }
                    if (exists $enc_hash{'$remote_encl_vendor'}->{'$remote_encl_model'}->{'$remote_encl_name'})
                    {
                        $remote_encl_id =
                            $enc_hash{'$remote_encl_vendor'}->{'$remote_encl_model'}->{'$remote_encl_name'};
                    }
                    else
                    {
                        my $rec = VRTS::Enclosure::SIPlugin::Util::get_enclid_from_name
                                   ($remote_encl_name, $remote_encl_vendor, $remote_encl_model);
                        if (defined $rec and scalar(@{$rec}) == 1)
                        {
                            $remote_encl_id =  $rec->[0]->{ID};
                            $enc_hash{'$remote_encl_vendor'}->{'$remote_encl_model'}->{'$remote_encl_name'} =
                            $remote_encl_id;
                        }
                    }
                }
                elsif (defined $remote_encl_alt_id && $remote_encl_alt_id ne &UNDEF_FIELD)
                {
                     if (exists $enc_hash{'$remote_encl_vendor'}->{'alt_id'}->{'$remote_encl_alt_id'})
                     {
                         $remote_encl_id = $enc_hash{'$remote_encl_vendor'}->{'alt_id'}->{'$remote_encl_alt_id'}; 
                     }
                     else
                     {
                         my $rec = VRTS::Enclosure::SIPlugin::Util::get_encl_from_alt_id ($remote_encl_alt_id);
                         if (defined $rec and scalar(@{$rec}) == 1)
                         {
                             $remote_encl_id = $rec->[0]->{ID};
                             $enc_hash{'$remote_encl_vendor'}->{'alt_id'}->{'$remote_encl_alt_id'} =
                                $remote_encl_id;
                         }
                     }
     
                 }
            }
            else
            {
                $remote_encl_id = $obj_hash->{$map}->{remote_encl_id};
            }

            if ($partner_id eq "Unknown" && $remote_encl_id ne "Unknown")
            {
                if (defined $partner_name && $partner_name ne &UNDEF_FIELD)
                {
                    # if we have the remote encl id then we can get the partner_id
                    # from remote ldev name if partner_id is not specified
                    my $rec = VRTS::Enclosure::SIPlugin::Util::get_ldevid_from_name
                             ($partner_name, $remote_encl_id);
                    if (defined $rec and scalar(@{$rec}) == 1)
                    {
                       $partner_id = $rec->[0]->{ID};
                    }
                }
                elsif (defined $partner_alt_id && $partner_alt_id ne &UNDEF_FIELD)
                {
                    my $rec = VRTS::Enclosure::SIPlugin::Util::get_ldevid_from_alt_id
                             ($partner_alt_id, $remote_encl_id);
                    if (defined $rec and scalar(@{$rec}) == 1)
                    {
                       $partner_id = $rec->[0]->{ID};
                    }
                }
            }
        }

        if (defined $obj_hash->{$map}->{rep_type})
        {
            $row[&VRTS::Constants::P_ARRAY_LDEV_MAP::SUBTYPE] = $obj_hash->{$map}->{rep_type};
        }

        $val = $obj_hash->{$map}->{rep_state};
        if (defined $val && $val ne "&UNDEF_FIELD")
        {
            if (!exists $replication_state{$val})
            {
                $msg = "Invalid value specified. Allowed values are: @rep_states";
	        $bad_record->{rep_state} = {
					      fld_name => $def->{rep_state},
					      value => $val,
					      dbg_level => &MSG_LEVEL_ERROR,
                                              error_msg => $msg
					   };
	    }
	    else
	    {
		$row[&VRTS::Constants::P_ARRAY_LDEV_MAP::STATE] = $replication_state{$val};
	    }
        }

	$val = $obj_hash->{$map}->{rep_mode};
	if (defined $val && $val ne "&UNDEF_FIELD")
	{
	    if (!exists $rep_mode{$val})
	    {
                $msg = "Invalid value specified. Allowed values are: @$rep_modes_list";
		$bad_record->{rep_mode} = {
					      fld_name => $def->{rep_mode},
					      value => $val,
					      dbg_level => &MSG_LEVEL_ERROR,
                                              error_msg => $msg
					  };
	    }
	    else
	    {
		$row[&VRTS::Constants::P_ARRAY_LDEV_MAP::REP_MODE] = $rep_mode{$val};
	    }
	}

	$val = $obj_hash->{$map}->{access_mode};
	if (defined $val) 
	{
	    if (!exists $replication_access_mode{$val})
	    {
                $msg = "Invalid value specified. Allowed values are: @rep_access_modes";
		$bad_record->{access_mode} = {
					      fld_name => $def->{access_mode},
					      value => $val,
					      dbg_level => &MSG_LEVEL_ERROR,
                                              error_msg => $msg
					  };
	    }
	    else
	    {
		$row[&VRTS::Constants::P_ARRAY_LDEV_MAP::ACCESS_MODE] = $replication_access_mode{$val};
	    }
	}


	$row[&VRTS::Constants::P_ARRAY_LDEV_MAP::TYPE] = "REPLICATION";
        $row[&VRTS::Constants::P_ARRAY_LDEV_MAP::SOURCE_ID] = $obj_hash->{$map}->{id};

        $row[&VRTS::Constants::P_ARRAY_LDEV_MAP::DEST_ID] = $partner_id;

        if ($obj_hash->{$map}->{across_encls} == &REP_REMOTE)
        {
            $row[&VRTS::Constants::P_ARRAY_LDEV_MAP::SUBTYPE] = "Remote Copy";
            $row[&VRTS::Constants::P_ARRAY_LDEV_MAP::ACROSS_SUBSYS] = 1;
            if ($obj_hash->{$map}->{designation} == &REP_DESIG_SOURCE)
            {
                $row[&VRTS::Constants::P_ARRAY_LDEV_MAP::SOURCE_SUBSYSID] = $vdid;
                $row[&VRTS::Constants::P_ARRAY_LDEV_MAP::DEST_SUBSYSID] = $remote_encl_id;
            }
            else
            {
                $row[&VRTS::Constants::P_ARRAY_LDEV_MAP::SOURCE_SUBSYSID] = $remote_encl_id;
                $row[&VRTS::Constants::P_ARRAY_LDEV_MAP::DEST_SUBSYSID] = $vdid;
                $row[&VRTS::Constants::P_ARRAY_LDEV_MAP::DEST_ID] = $obj_hash->{$map}->{id};
                $row[&VRTS::Constants::P_ARRAY_LDEV_MAP::SOURCE_ID] = $partner_id;
            }
        }
        else
        {
            $row[&VRTS::Constants::P_ARRAY_LDEV_MAP::ACROSS_SUBSYS] = 0;
            $row[&VRTS::Constants::P_ARRAY_LDEV_MAP::DEST_SUBSYSID] = $vdid;
            $row[&VRTS::Constants::P_ARRAY_LDEV_MAP::SOURCE_SUBSYSID] = $vdid;
        }

	$ldev_tb->addrow($ldev_row);     
        if (defined $partner_row)
        {
            $ldev_tb->addrow($partner_row);
        }
        $ldevmap_tb->addrow(\@row);
        $num_objects++;

report_err:
        my @errors = keys %$bad_record;
        if (scalar(@errors))
        {
            $bad_record->{id} = {
                                    fld_name => $def->{id},
                                    value => $obj_hash->{$map}->{id},
                                    dbg_level  => &MSG_LEVEL_DBG
                                };
            print_record ($bad_record);
        }
    }
    log_msg("Total Records discovered: $num_objects");
    $ldevmap_tb->print_table() if ($debug_level == &MSG_LEVEL_DBG_3);
    return (undef, 0);
}

__DISCOVER_META_LDEVS__
# Description : method discovers association of meta ldevs with its segments
# Input :
#       agent - Reference to the Agentlet framework
#       vdid - VDID of the array
# Ouput:
#       msg - Error message on failure to discover
#       retval - 0 on success. non-zero on failure.
sub discover_meta_ldevs
{
    my $agent = shift;
    my $vdid = shift;
    my $msg;
    my $num_objects = 0;

    my $cmd = $discovery_cmdline." --list meta-ldevs --encl $vdid";
    log_msg ("\nDiscovering Meta ldevs and segments...", &MSG_LEVEL_DBG, 1);
    log_msg ("(cmd: $cmd)...", &MSG_LEVEL_DBG);
    $discovery_ctxt->{CMD_LINE} = $cmd;

    my $out = VRTS::Enclosure::SIPlugin::Util::exec_cmd($ch, $discovery_ctxt);
    if ($out->[0] != 0)
    {
        $msg = "Failed to push/exec CLI: $out->[1]. $out->[0]";
        $msg = VRTS::Enclosure::SIPlugin::Util::get_error($msg, $out->[0]);
        log_msg ($msg, &MSG_LEVEL_ERROR);
        return undef, $out->[0];
    }
    my $fh = $out->[1];
    my ($obj_hash, $ret) = $parser->get_object("META_LDEV", $fh, "id");
    if ($ret)
    {
        $msg = "Vendor CLI failed to execute:$obj_hash:retcode $ret";
        $msg = VRTS::Enclosure::SIPlugin::Util::get_error($msg, $ret);
        log_msg ($msg, &MSG_LEVEL_ERROR);
        return (undef, $ret);
    }

    if (!defined $obj_hash )
    {
         $msg = VRTS::Enclosure::SIPlugin::Util::get_error($msg,
                &VRTS::Enclosure::SIPlugin::Catalog::NO_OBJECTS_DISCOVERED);
         log_msg ($msg);
         return undef, &VRTS::Enclosure::SIPlugin::Catalog::NO_OBJECTS_DISCOVERED;
    }

    my $def = $objects_def->{META_LDEV};
    my $val;
    my %enc_hash;

    foreach my $map (keys%$obj_hash)
    {
        my @row;
        my $bad_record = {};

        my $ldev_row = $ldev_tb->getrowwithvalue(&VRTS::Constants::P_ARRAY_LDEV::ID, $obj_hash->{$map}->{id});
        if (!defined $ldev_row)
        {
             $msg = "Invalid Logical disk specified. Ignoring the record";
             $bad_record->{id} = {
                                    fld_name => $def->{id},
                                    value => $obj_hash->{$map}->{id},
                                    dbg_level => &MSG_LEVEL_ERROR,
                                    error_msg => $msg
                                 };
             print_record ($bad_record);
             next; 
        }

        my $segment_row = $ldev_tb->getrowwithvalue(&VRTS::Constants::P_ARRAY_LDEV::ID, 
                    $obj_hash->{$map}->{segment_id});
        if (!defined $segment_row)
        {
             $msg = "Invalid Logical disk specified. Ignoring the record";
             $bad_record->{segment_id} = {
                                    fld_name => $def->{segment_id},
                                    value => $obj_hash->{$map}->{segment_id},
                                    dbg_level => &MSG_LEVEL_ERROR,
                                    error_msg => $msg
                                 };
             goto report_err;
        }

        $row[&VRTS::Constants::P_ARRAY_LDEV_MAP::SOURCE_ID] = $obj_hash->{$map}->{id};
        $row[&VRTS::Constants::P_ARRAY_LDEV_MAP::DEST_ID] = $obj_hash->{$map}->{segment_id};
        $row[&VRTS::Constants::P_ARRAY_LDEV_MAP::SOURCE_NAME] =
             $ldev_row->[&VRTS::Constants::P_ARRAY_LDEV::NAME];
        $row[&VRTS::Constants::P_ARRAY_LDEV_MAP::DEST_NAME] =
             $segment_row->[&VRTS::Constants::P_ARRAY_LDEV::NAME];
        $row[&VRTS::Constants::P_ARRAY_LDEV_MAP::SOURCE_SUBSYSID] = $vdid;
        $row[&VRTS::Constants::P_ARRAY_LDEV_MAP::DEST_SUBSYSID] = $vdid;
        $row[&VRTS::Constants::P_ARRAY_LDEV_MAP::TYPE] =  "LUSE";

        $ldev_row->[&VRTS::Constants::P_ARRAY_LDEV::TYPE] = "METALUN";
        $segment_row->[&VRTS::Constants::P_ARRAY_LDEV::TYPE] = "SEGMENT";

        if ($segment_row->[&VRTS::Constants::P_ARRAY_LDEV::STATUS] == 1)
        {
            # If segments can be assigned out to hosts then set CONSIDER_SEGMENT_LUN capability
            my $capbs_row = $capabilities_tb->getrowwithvalue(&VRTS::Constants::P_ARRAY_CAPABILITY::ARRAY_PROP_ID,
                            $capbs->{'CONSIDER_SEGMENT_LUN'}->{index});
            if (defined $capbs_row)
            {
                $capbs_row->[&VRTS::Constants::P_ARRAY_CAPABILITY::VALUE] = 1;
                $capabilities_tb->addrow($capbs_row);
            }
        }
 
        $ldev_tb->addrow($ldev_row);
        $ldev_tb->addrow($segment_row);

        $ldevmap_tb->addrow(\@row);
        $num_objects++;


report_err:
        my @errors = keys %$bad_record;
        if (scalar(@errors))
        {
            $bad_record->{id} = {
                                    fld_name => $def->{id},
                                    value => $obj_hash->{$map}->{id},
                                    dbg_level  => &MSG_LEVEL_DBG
                                };
            print_record ($bad_record);
        }
    }
    log_msg("Total Records discovered: $num_objects");
    $ldevmap_tb->print_table() if ($debug_level == &MSG_LEVEL_DBG_3);
    return (undef, 0);
}
